package edu.northwestern.cs.aqualab.Cattle;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocketFactory;
import javax.xml.ws.ProtocolException;

import edu.northwestern.cs.aqualab.Cattle.json.JSONException;
import edu.northwestern.cs.aqualab.Cattle.json.JSONObject;



/**
 * A Report (informally, a Cow) represents a single JSON report. Populate it by
 * calling getJSONList() and adding your data. To send it on its way, call
 * send(). (This method is synchronous to simplify error handling. Deal with
 * threading yourself.)<br>
 * <br>
 * setDestinationHost() and setResource() must be called before send(), but
 * setDestinationPort() is optional (the default destination port is 80).<br>
 * <br>
 * If you have one Report, you're likely to want more with similar
 * configurations. The instance method clear() helps you reuse a Report by
 * wiping all key/value pairs stored in its backing JSONObject. Alternatively,
 * the instance method clone() creates new Reports copied from a prototype.
 */
public class Report implements Serializable {
    private static final long serialVersionUID = 1L;

    private static ArrayList<KeyStore> TRUSTED_KEYSTORES = new ArrayList<KeyStore>();
    private static SocketFactory SSL_FACTORY = SSLSocketFactory.getDefault();
    private static final String POST_FORMAT = "POST %s HTTP/1.0\r\n" + "Content-Length: %d\r\n"
                                              + "\r\n" + "%s\r\n";

    private long last_attempt = -1;
    private int num_attempts = 0;
    private String dest_host;
    private int dest_port = 80;
    private String resource;
    private List<JSONObject> beefs;


    /**
     * Attempt to add a KeyStore to the trusted list and reinitialize the
     * SSLSocketFactory. Restores the old factory on failure.
     *
     * @param ks
     *            The new KeyStore.
     * @return True if the new KeyStore was added successfully, false otherwise.
     */
    public static synchronized boolean addKeyStore(KeyStore ks) {
        TRUSTED_KEYSTORES.add(ks);
        SocketFactory old_factory = SSL_FACTORY;

        try {
            SSL_FACTORY = new SelfSignedSSLSocketFactory(TRUSTED_KEYSTORES);
        } catch (KeyManagementException kme) {
            SSL_FACTORY = old_factory;
            TRUSTED_KEYSTORES.remove(ks);
            return false;
        } catch (NoSuchAlgorithmException nsae) {
            SSL_FACTORY = old_factory;
            TRUSTED_KEYSTORES.remove(ks);
            return false;
        }

        return true;
    }

    /**
     * Represents the outcome of a call to send(), the possibilities for which
     * are myriad. If success is true, the server's replies are in responses in
     * the same order as the reports were. If it's false, check exception first,
     * then HTTP_code, then responses.
     */
    public class ReportResult {
        public boolean success = false;
        public int HTTP_code = -1;
        public JSONObject response = null;
        public List<JSONObject> responses = new ArrayList<JSONObject>();
        public Exception exception;


        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            sb.append("success: ");
            sb.append(this.success);

            if (this.exception != null) {
                sb.append(", exception: ");
                sb.append(exception.toString());
            } else {
                sb.append(", HTTP_CODE: ");
                sb.append(this.HTTP_code);
                if (this.HTTP_code == 200) {
                    sb.append(", response: ");
                    if (this.response == null) {
                        sb.append("null");
                    } else {
                        sb.append(response.toString());
                    }
                }
            }
            return sb.toString();
        }
    }


    public Report() {
        if (this.beefs == null) {
            this.beefs = new ArrayList<JSONObject>();
        }
    }


    public Report(JSONObject init) {
        this.beefs.add(init);
    }


    public Report(List<JSONObject> inits) {
        this.beefs = inits;
    }


    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Report facing ");
        sb.append(this.dest_host);
        sb.append(":");
        sb.append(String.valueOf(this.dest_port));
        sb.append(this.resource);
        sb.append(" carrying ");
        for (JSONObject beef : this.beefs) {
            sb.append(beef.toString());
        }
        return sb.toString();
    }


    /**
     * Must be called before send().
     */
    public void setDestinationHost(String dest) {
        this.dest_host = dest;
    }


    public void setDestinationPort(int port) {
        this.dest_port = port;
    }


    /**
     * Must be called before send().
     */
    public void setResource(String res) {
        this.resource = res;
    }


    /**
     * @return The timestamp, in millis, when the last call to send() completed.
     */
    public long getLastAttempt() {
        return this.last_attempt;
    }


    /**
     * @return The number of times send() has been called.
     */
    public int getNumAttempts() {
        return this.num_attempts;
    }


    /**
     * Connects to the server and transfers data. Returns null if
     * @param ssl Use SSL?
     * @return A {@link ReportResult}.
     */
    public ReportResult send(boolean ssl) {
        if (this.dest_host == null) {
            throw new IllegalStateException("No destination host set!");
        }
        if (this.resource == null) {
            throw new IllegalStateException("No HTTP resource set!");
        }

        ReportResult retval = new ReportResult();
        Socket s = null;
        PrintWriter out = null;
        BufferedReader in = null;

        try {
            if (ssl) {
                s = SSL_FACTORY.createSocket(dest_host, dest_port);
            } else {
                s = new Socket(dest_host, dest_port);
            }

            out = new PrintWriter(s.getOutputStream(), false);
            in = new BufferedReader(new InputStreamReader(s.getInputStream()));

            StringBuilder sb = new StringBuilder();
            for (JSONObject beef : this.beefs) {
                sb.append(beef.toString());
                sb.append("\n");
            }
            String json_out = sb.toString();

            String post = String.format(POST_FORMAT, resource, json_out.length(), json_out);
            out.write(post);
            out.flush();

            String reply = in.readLine();
            if (reply == null) {
                if (ssl) {
                    throw new ProtocolException("No response read - maybe you need to turn off SSL?");
                } else {
                    throw new ProtocolException("No response read - maybe you need to turn on SSL?");
                }
            }

            retval.HTTP_code = Integer.parseInt(reply.split("\\s")[1]);
            boolean HTTP_success = (retval.HTTP_code == 200);

            if (HTTP_success) {
                // Strip off HTTP headers and advance to the body.
                while (!(reply = in.readLine()).equals("")) {}

                while ((reply = in.readLine()) != null) {
                    retval.responses.add(new JSONObject(reply));
                }

                retval.response = retval.responses.get(0);

                boolean all_succeeded = true;
                for (JSONObject response : retval.responses) {
                    all_succeeded &= (response.has("error") == false);
                }
                retval.success = all_succeeded;
            }
        } catch (JSONException jsone) {
            retval.exception = jsone;
        } catch (UnknownHostException uhe) {
            retval.exception = uhe;
        } catch (IOException ioe) {
            retval.exception = ioe;
            //ioe.printStackTrace();
        } finally {
            try {
                if (out != null) out.close();
                if (in != null) in.close();
                if (s != null) s.close();
            } catch (IOException ioe) {            	
            }

            this.last_attempt = System.currentTimeMillis();
            ++this.num_attempts;
        }

        return retval;
    }


    /**
     * Clear the JSONObjects associated with this Report.
     */
    public void clear() {
        this.beefs = new ArrayList<JSONObject>();
    }


    /**
     * Create a new Report configured identically to this one, including a copy
     * of its JSON payload.
     */
    public Report clone() {
        List<JSONObject> beefs_copy = new ArrayList<JSONObject>(this.beefs.size());
        for (JSONObject beef : this.beefs) {
            beefs_copy.add(beef);
        }
        Report ret = new Report(beefs_copy);
        ret.setDestinationHost(this.dest_host);
        ret.setDestinationPort(this.dest_port);
        ret.setResource(this.resource);
        return ret;
    }


    /**
     * Provides direct access to a JSONObject backing this Report.
     *
     * @return This Report's internal JSONObject.
     * @deprecated Use {@link #getJSONList() getJSONList} instead.
     */
    @Deprecated
    public JSONObject getJSON() {
        if (this.beefs.size() == 0) {
            this.beefs.add(new JSONObject());
        }
        return this.beefs.get(0);
    }


    /**
     * Provides direct access to the list of JSONObjects backing this Report.
     *
     * @return This Report's internal {@literal List<JSONObject>}.
     */
    public List<JSONObject> getJSONList() {
        return this.beefs;
    }


    private void writeObject(ObjectOutputStream out) throws IOException {
        out.writeLong(serialVersionUID);
        out.writeLong(this.last_attempt);
        out.writeInt(this.num_attempts);
        out.writeObject(this.dest_host);
        out.writeInt(this.dest_port);
        out.writeObject(this.resource);
        out.writeInt(this.beefs.size());
        for (int i = 0; i < this.beefs.size(); ++i) {
            out.writeObject(this.beefs.get(i).toString());
        }
    }


    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        long serialVersionID = in.readLong();
        // I will make the assumption that we won't need more than INT_MAX
        // versions.
        switch ((int) serialVersionID) {
        case 1:
            this.last_attempt = in.readLong();
            this.num_attempts = in.readInt();
            this.dest_host = (String) in.readObject();
            this.dest_port = in.readInt();
            this.resource = (String) in.readObject();
            this.beefs = new ArrayList<JSONObject>();
            int num_beefs = in.readInt();
            try {
                // An exception here would indicate a bug in the JSON
                // implementation or in this serialization code.
                for (int i = 0; i < num_beefs; ++i) {
                    this.beefs.add(new JSONObject((String) in.readObject()));
                }
            } catch (JSONException jsone) {}
        }
    }
}
