/**
 * Dasu Project
 *
 * File:         DasuConfiguration.java
 * RCS:          $Id$
 * Description:  DasuConfiguration class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:23:49 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.KeyStore;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Properties;
import java.util.UUID;
import java.util.logging.Logger;
import org.gudy.azureus2.core3.util.Debug;

import com.azureus.plugins.aznetmon.main.RSTPacketStats;

import edu.northwestern.cs.aqualab.Cattle.json.JSONException;
import edu.northwestern.cs.aqualab.Cattle.json.JSONObject;
import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez / David Choffnes &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The DasuConfiguration class ...
 */

public class DasuConfiguration {
	
	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	
	private static DasuConfiguration self = null;
	
	
	public static final String DASU_SERVER_LIST_LOC = "http://haleakala.cs.northwestern.edu/dasu_v4/ws/servers-list.dat"; // mas939-imac.cs.northwestern.edu/dasu/ws/dasu-list.dat
	public static final String DASU_DEFAULT_SERVER = "http://haleakala.cs.northwestern.edu";	
	
	/** old PhpReporter variables **/	
	public static final String DASU_DEFAULT_REPORTER_SCRIPT = "/dasu_v4/ws/reportStats.php";
	
	/** active scripts */
	public static final String DASU_CONFIG = "/dasu/ws/dasuConfig.properties"; /**/
	public static final String DASU_MESSAGES = "/dasu/ws/message-list.dat";	 /**/
	
	/** new variables **/
	public static final String DASU_FILENAME_PREFIX = "DasuUiRulesModules_";
	public static final String DASU_FILENAME_SUFFIX = ".jar";
	public static final String DASU_AME_FILENAME_PREFIX = "AMExperiment";
	public static final String DASU_AME_FILENAME_SUFFIX = ".drl";
	public static final String DASU_AME_SIGNATURE_SUFFIX = ".signature";
	
	public static final String DASU_REPORT_SERVER_HOST = "reports.aqualab.cs.northwestern.edu";
	public static final Integer DASU_REPORT_SERVER_PORT = 8088;
	public static final String DASU_REPORT_RESOURCE = "/dasu";
	
	public static final String DASU_CONFIG_SERVER_HOST = "https://config.aqualab.cs.northwestern.edu";
	public static final Integer DASU_CONFIG_SERVER_PORT = 443;
	//public static final String DASU_CONFIG_SERVER_RESOURCE = "/config";
	public static final String DASU_CONFIG_SERVER_RESOURCE = "";
	
	public static final String DASU_CONFIG_DYNAMIC = "/data/dasu/dyn_version/";
	public static final String DASU_CONFIG_PROPERTIES = "/data/dasu/config_properties";
	public static final String DASU_MESSAGE_LIST = "/data/dasu/message_list";
	
	public static final String DASU_SCRIPT_NDT = "/data/dasu/dynamic/ndt.php";
	public static final String DASU_SCRIPT_AME = "/data/dasu/dynamic/AMExperiment.php";
	public static final String DASU_SCIPT_UPDATE_CONFIG = "/data/dasu/dynamic/UpdateConfig.php";
	public static final String DASU_SCRIPT_AME_ON_DEMAND = "/data/dasu/dynamic/communicateWithTaskManager.php"; 
	
	public static final String KEYSTORE_JKS = "dasu_keyStore.jks";
	public static final String KEYSTORE_PASS = "fadazani";
	
	protected static ArrayList<String> propertiesLocations = new ArrayList<String>();
	public static String propertiesFileName = "dasu.properties";
	protected static JSONObject propsJson = null;
	private static KeyStore dasuKeyStore = null;
	
	public static final String PROBE_MODULES_PROPERTIES = "probe.modules.properties";
	
	private String dasuConfigDynamicResource = "release.php";
//	private String dasuConfigDynamicResource = "beta.php";
	private boolean enablePhPReporter = true;
	private boolean enableAzureusSpeedTest = true;
	private int peerUpdateIntervalSec = 30; //interval (seconds) for subsequent updates for the peerManager et al.
	private boolean doTraceroute = true;
	private boolean doAnaLookup = true;
	private int tracerouteTimeout = 60; //seconds
	private int dnsTimeout = 1; //seconds
	private int dbUpdateFreqSec = 30; //seconds
	private int updateCheckFreqSec = 600; //seconds
	private int maxPeers = 100;
	private String uuid = null;
	private int UUIDid = -1;
	private int schedulerInterval = 302; //5*60 + 2 seconds
	private int AMCSleepInterval = 30; //seconds
	private int managerSleepInterval = 30; //seconds
	private int AMECheckupInterval = 5*60; //seconds
	private int messageManagerInterval = 60*60; //60*60 seconds = 1 hour
	private int dbStaticFlushInterval = 15*60; //seconds (15 min)
	private int DBManagerSleepInterval = 2*60; //seconds (2 min)

	private int maxLogEntries = 200;
	private int maxHistoricSignals = 20;	
	private float inferenceVersion = 2.0F;
	private int maxDataFileSize = 20*1024*1024; //bytes 
	
	private int maxNumEntriesPerHistoricSignal = 370;
	private final int maxHistoricSignalsHARDLIMIT = 40;
	private int maxPings = 10;
	private final int maxNumEntriesPerHistoricSignalHARDLIMIT = 740;	
	private final int maxPingsHARDLIMIT = 40;
	private int maxAMExperiments = 1;
	private String maxMemJVM = "64M";
	private final int maxAMExperimentsHARDLIMIT = 3;

	private int dbReconnectInterval = 5*60; // 5 minutes to retry
	private int reportingTimeout = 60;
	
	private boolean ignoreStateChangeRST=false;
	private int stateChangeRSTInterval = 2; // seconds

	/** the sample interval for news, in seconds */
	private int sampleInterval = 30;
	private int RSTimeWindowSize = 1; /*minutes*/	
	private int timeSpan = 5; /*minutes*/	
	private Properties probeModulesProperties = new Properties();

	/** Variables that can only be set locally (not overwritten by server) **/
	private boolean recordStats = true;
	private boolean sendToDatabase = true;
	private boolean enablePopups = true;
	private boolean runExperiments = true;
	
	private DasuConfiguration() {
		
	}
	
	public synchronized static DasuConfiguration getInstance(){
		
		if (self==null){
			self = new DasuConfiguration();
			try {
				loadKeyStore();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
		return self;
	}
	
	public String getResourceScriptRequest() {
		
		StringBuilder resource = new StringBuilder();
		
		//basic information
		resource.append("?ip=" + Main.getPublicIpAddress() +
			"&uuid=" + DasuConfiguration.getInstance().getUUID());
		
		//geographic location
		if ((NetworkAdminASNLookupImpl.getInstance() != null) 
				&& (NetworkAdminASNLookupImpl.getInstance().getASNInfo() != null)) {

			resource.append("&asn=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS() +
				"&cc=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getCountry());
		}

		//rules and client version
		resource.append( "&clientversion=" + Main.getPluginInterface().getPluginVersion() +
				"&interfversion=" + getInferenceVersion() );
		
		
		return resource.toString();
		
	}
	
	public int getDbUpdateFreqSec() {
		return dbUpdateFreqSec;
	}

	public KeyStore getKeyStore() {
		return dasuKeyStore;
	}

	public void setDbUpdateFreqSec(int dbUpdateFreqSec) {
		this.dbUpdateFreqSec = dbUpdateFreqSec;
	}
	
	public int getTracerouteTimeout() {
		return tracerouteTimeout;
	}

	public void setTracerouteTimeout(int tracerouteTimeout) {
		this.tracerouteTimeout = tracerouteTimeout;
	}
	
	public int getDnsTimeout() {
		return dnsTimeout;
	}

	public void setDnsTimeout(int dnsTimeout) {
		this.dnsTimeout = dnsTimeout;
	}

	public int getMaxPeers() {
		return maxPeers;
	}

	public void setMaxPeers(int maxPeers) {
		this.maxPeers = maxPeers;
	}

	public int getUpdateCheckFreqSec() {
		return updateCheckFreqSec;
	}

	public void setUpdateCheckFreqSec(int updateCheckFreqSec) {
		this.updateCheckFreqSec = updateCheckFreqSec;
	}

	public boolean isDoTraceroute() {
		return doTraceroute;
	}

	public boolean isDoAnaLookup() {
		return doAnaLookup;
	}
	
	public void setDoTraceroute(boolean doTraceroute) {
		this.doTraceroute = doTraceroute;
	}

	public void setDoAnaLookup(boolean doAnaLookup) {
		this.doAnaLookup = doAnaLookup;
	}
	
	public int getPeerUpdateIntervalSec() {
		return peerUpdateIntervalSec;
	}

	public void setPeerUpdateIntervalSec(int peerUpdateIntervalSec) {
		this.peerUpdateIntervalSec = peerUpdateIntervalSec;
	}

	public boolean isRecordStats() {
		return recordStats;
	}

	public void setRecordStats(boolean recordStats) {
		this.recordStats = recordStats;
	}
	
	public boolean isEnablePopups() {
		return enablePopups;
	}

	public void setEnablePopups(boolean enablePopups) {
		this.enablePopups = enablePopups;
	}

	public boolean isEnablePhPReporter() {
		return enablePhPReporter;
	}

	public void setEnablePhPReporter(boolean enablePhPReporter) {
		this.enablePhPReporter = enablePhPReporter;
	}

	public boolean isEnableAzureusSpeedTest() {
		return enableAzureusSpeedTest;
	}

	public void setEnableAzureusSpeedTest(boolean enableAzureusSpeedTest) {
		this.enableAzureusSpeedTest = enableAzureusSpeedTest;
	}

	public boolean isSendToDatabase() {
		return sendToDatabase;
	}

	public void setSendToDatabase(boolean sendToDatabase) {
		this.sendToDatabase = sendToDatabase;
	}

	public void setProbeModulesProperties(Properties prop) {
		probeModulesProperties = prop;
	}
	
	private static void loadKeyStore() throws Exception {
		
        dasuKeyStore = KeyStore.getInstance("JKS");
        InputStream is = null;
        
        if ( Main.getPluginInterface() != null)
        	is = Main.getPluginInterface().getPluginClassLoader().getResourceAsStream(KEYSTORE_JKS);
        else
        	is = ClassLoader.getSystemResourceAsStream(KEYSTORE_JKS);
        
        dasuKeyStore.load(is, KEYSTORE_PASS.toCharArray());

	}
	
	public int getDbReconnectInterval() {
		return dbReconnectInterval;
	}

	public void setDbReconnectInterval(int dbReconnectInterval) {
		this.dbReconnectInterval = dbReconnectInterval;
	}

	public void setReportingTimeout(int timeout) {
		this.reportingTimeout = timeout;
	}

	public int getReportingTimeout() {
		return reportingTimeout;
	}

	public void setMaxPings(int value) {
		
		if (value <= maxPingsHARDLIMIT)
			this.maxPings = value;
		
	}
	
	public int getMaxPings(){
		return maxPings;
	}

	public void setMaxAMExperiments(int value) {
		
		if (value <= maxAMExperimentsHARDLIMIT)
			this.maxAMExperiments = value;
		
	}
	
	public int getMaxAMExperiments(){
		return maxAMExperiments;
	}
	
	public void setMaxMemJVM(String value) {
		this.maxMemJVM = value;
	}
	
	public String getMaxMemJVM(){
		return maxMemJVM;
	}

	public void setInferenceVersion(float value) {
		this.inferenceVersion = value;
	}

	public float getInferenceVersion(){
		return inferenceVersion;
	}

	public String getUUID() {
		return uuid;
	}
	
	public int getUUIDid() {
		return UUIDid;
	}
	
	public void setUUID(String uuid){
		this.uuid = uuid;
	}

	public void setUUIDid(int uuidid){
		this.UUIDid = uuidid;
	}

	public boolean isRunExperiments() {
		return runExperiments;
	}
	
	public void setRunExperiments(boolean runExperiments){
		this.runExperiments = runExperiments;
	}

	public boolean isIgnoreStateChangeRST() {
		return ignoreStateChangeRST;
	}
	public int getSampleInterval() {
		return sampleInterval;
	}

	public void setSampleInterval(int sampleInterval) {
		this.sampleInterval = sampleInterval;
	}

	public int getSchedulerInterval() {
		return schedulerInterval ;
	}

	public void setSchedulerInterval(int interval){
		this.schedulerInterval = interval;
	}
	
	public String getDasuConfigDynamicResource() {
		return dasuConfigDynamicResource;
	}
	
	public void setDasuConfigDynamicResource(String resource) {
		this.dasuConfigDynamicResource = resource;
	}
	
	public int getAMECheckupInterval() {
		return AMECheckupInterval ;
	}
	
	public void setAMECheckupInterval(int interval){
		this.AMECheckupInterval = interval;
	}
	
	public int getMessageManagerInterval() {
		return messageManagerInterval ;
	}
	
	public void setMessageManagerInterval(int interval){
		this.messageManagerInterval = interval;
	}

	public int getStateChangeRSTInterval() {
		return stateChangeRSTInterval ;
	}

	public void setStateChangeRSTInterval(int interval){
		this.stateChangeRSTInterval = interval;
	}


	public int getAMCSleepInterval() {
		return AMCSleepInterval;
	}

	public void setAMCSleepInterval(int interval){
		this.AMCSleepInterval = interval;
	}
	
	public int getDBManagerSleepInterval() {
		return DBManagerSleepInterval;
	}

	public void setDBManagerSleepInterval(int interval){
		this.DBManagerSleepInterval = interval;
	}
	
	public int getDbStaticFlushInterval() {
		return DBManagerSleepInterval;
	}

	public void setDbStaticFlushInterval(int interval){
		this.dbStaticFlushInterval = interval;
	}
	
	public int getManagerSleepInterval() {
		return managerSleepInterval;
	}

	public void setManagerSleepInterval(int interval){
		this.managerSleepInterval = interval;
	}

	public int getMaxLogEntries() {
		return maxLogEntries ;
	}
	
	public void setMaxLogEntries(int entries){
		this.maxLogEntries = entries;
	}
	
	public int getMaxDataFileSize() {
		return maxDataFileSize ;
	}
	
	public void setMaxDataFileSize(int size){
		this.maxDataFileSize = size;
	}
	
	public int getMaxHistoricSignals() {
		return maxHistoricSignals ;
	}
	
	public void setMaxHistoricSignals(int entries){
		
		if (entries <= maxHistoricSignalsHARDLIMIT)
			this.maxHistoricSignals = entries;
	}

	public int getMaxNumEntriesPerHistoricSignal() {
		return maxNumEntriesPerHistoricSignal ;
	}
	
	public void setMaxNumEntriesPerHistoricSignal(int entries){

		if (entries <= maxNumEntriesPerHistoricSignalHARDLIMIT)
			this.maxNumEntriesPerHistoricSignal = entries;
	}

	public boolean getIgnoreStateChangeRST() {
		return ignoreStateChangeRST;
	}
	
	public void setIgnoreStateChangeRST(boolean ignoreStateChangeRST) {
		this.ignoreStateChangeRST = ignoreStateChangeRST;
	}

	public int getRSTimeWindowSize() {
		return RSTimeWindowSize;
	}

	public int getTimeSpan() {
		return timeSpan;
	}

	public void setRSTimeWindowSize(int timeWindowSize) {
		RSTimeWindowSize = timeWindowSize;
	}

	public void setTimeSpan(int timeSpan) {
		this.timeSpan = timeSpan;
	}
	
	public Properties getProbeModulesProperties() {
		return probeModulesProperties;
	}

	public void updatePropertiesJson(JSONObject json ) {		
		updatePropertiesJson(json, false);		
	}
	
	private void updatePropertiesJson(JSONObject json, boolean override) {

		LOGGER.info("CONFIG: Updating properties...");
		try {
					
			Iterator<Object> iter = json.keys();
			while ( iter.hasNext() ) {
				String key = iter.next().toString();
				String value = json.get(key).toString();
	
				try {
										
					if (key.equals("dasu.enableStats")){
						if ( override )
							recordStats= value.equals("true");
					}else if (key.equals("dasu.sendToDatabase")){
						if ( override )
							sendToDatabase= value.equals("true");			
					}else if (key.equals("dasu.enablePopups")){
						if ( override )
						enablePopups= value.equals("true");
					}else if (key.equals("dasu.runExperiments")){
						if ( override )
							runExperiments= value.equals("true");
					} 									
					else if (key.equals("dasu.stats.dbUpdate")){
						setDbUpdateFreqSec(Integer.valueOf(value));
					} else if (key.equals("dasu.stats.tracerouteTimeout")){
						setTracerouteTimeout(Integer.valueOf(value));
					} else if (key.equals("dasu.stats.dnsTimeout")){
						setDnsTimeout(Integer.valueOf(value));
					} else if (key.equals("dasu.update.updateCheck")){
						setUpdateCheckFreqSec(Integer.valueOf(value));
					}else if (key.equals("dasu.doTraceroute")){
						doTraceroute= value.equals("true");
					}else if (key.equals("dasu.doAnaLookup")){
						doAnaLookup= value.equals("true");
					}else if (key.equals("dasu.enablePhPReporter")){
						enablePhPReporter= value.equals("true");
					}else if (key.equals("dasu.enableAzureusSpeedTest")){
						enableAzureusSpeedTest= value.equals("true");
					}else if (key.equals("dasu.stats.peerUpdate")){
						setPeerUpdateIntervalSec(Integer.valueOf(value));
					}else if (key.equals("dasu.reportingTimeout")){
						setReportingTimeout(Integer.valueOf(value));
					} else if (key.equals("dasu.maxPings")){
						setMaxPings(Integer.valueOf(value));
					} else if (key.equals("dasu.maxDataFileSize")){
						setMaxDataFileSize(Integer.valueOf(value));
					} else if (key.equals("dasu.maxAMExperiments")){
						setMaxAMExperiments(Integer.valueOf(value));
					} else if (key.equals("dasu.uuid")){
						
						//only override the uuid if it is null in the first place
						if ( uuid == null )
							uuid = value;
						
					} else if (key.equals("dasu.UUIDid")){
						
						//only override the uuid if it is -1 in the first place
						if ( UUIDid == -1 )
							setUUIDid(Integer.valueOf(value));
						
					} else if (key.equals("dasu.ignoreStateChangeRST")){
						ignoreStateChangeRST = value.equals("true");
					} else if (key.equals("dasu.sampleInterval")){
						sampleInterval = Integer.valueOf(value);
						RSTPacketStats.INTERVAL = sampleInterval*1000;
					}else if (key.equals("dasu.inference.version")){
						setInferenceVersion(Float.valueOf(value));
					}else if (key.equals("dasu.schedulerInterval")){
						setSchedulerInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.messageManagerInterval")){
						setMessageManagerInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.AMECheckupInterval")){
						setAMECheckupInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.stateChangeRSTInterval")){
						setStateChangeRSTInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.AMCSleepInterval")){
						setAMCSleepInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.db.managerSleepInterval")){
						setDBManagerSleepInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.db.staticFlushInterval")){
						setDbStaticFlushInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.managerSleepInterval")){
						setManagerSleepInterval(Integer.valueOf(value));
					}else if (key.equals("dasu.maxLogEntries")){
						setMaxLogEntries(Integer.valueOf(value));
					}else if (key.equals("dasu.maxHistoricSignals")){
						setMaxHistoricSignals(Integer.valueOf(value));
					}else if (key.equals("dasu.maxNumEntriesPerHistoricSignal")){
						setMaxNumEntriesPerHistoricSignal(Integer.valueOf(value));
					}else if (key.equals("dasu.RSTimeWindowSize")){
						setRSTimeWindowSize(Integer.valueOf(value));
					}else if (key.equals("dasu.RSTimeSpan")){
						setTimeSpan(Integer.valueOf(value));
					}else if (key.equals("dasu.maxMemJVM")){
						setMaxMemJVM(value);						
					}else if (key.equals("dasu.config.dynamic.resource")){
						setDasuConfigDynamicResource(value);
					}else if (key.equals("dasu.debug.filter")){
						
						String[] args = value.split(",");
						Main.setLogFilter(args);
						
					} else {
						LOGGER.warning("Unknown key: *"+key+"*");
					}
					
				} catch (Exception e){
					System.err.println("Error processing property...");
					e.printStackTrace();
				} catch (Error e){
					System.err.println("Error processing property...");
					e.printStackTrace();
				}
				
			}
			
			//boolean write = false;
			
			// if no uuid, generate it now
			if (uuid==null){
				uuid = UUID.randomUUID().toString();
				json.put("dasu.uuid", uuid);
				//write = true;				
			}
			
			if (UUIDid == -1) {
				
				UUIDid = Statistics.getInstance().makeUUIDid(uuid);
				if (UUIDid != -1) {					
					json.put("dasu.UUIDid", UUIDid);
				}
			}
			
			//if (write)
				writePropertiesJson(json);
		}
		catch (Exception e) {
			LOGGER.info("CONFIG: empty properties file" );
		}
		
		LOGGER.info("UUID " + uuid + " " + UUIDid);
		
	}
	
	private void writePropertiesJson(JSONObject props) {
		
		try {
			
			props.put("dasu.enableStats", recordStats);
			props.put("dasu.enablePopups", enablePopups);
			props.put("dasu.enablePhPReporter", enablePhPReporter);
			props.put("dasu.enableAzureusSpeedTest", enableAzureusSpeedTest);
			props.put("dasu.sendToDatabase", sendToDatabase);		
			props.put("dasu.stats.dbUpdate", dbUpdateFreqSec);
			props.put("dasu.stats.tracerouteTimeout", tracerouteTimeout);
			props.put("dasu.stats.dnsTimeout", dnsTimeout );
			props.put("dasu.update.updateCheck", updateCheckFreqSec);
			props.put("dasu.doTraceroute", doTraceroute);
			props.put("dasu.doAnaLookup", doAnaLookup);
			props.put("dasu.stats.peerUpdate", peerUpdateIntervalSec);
			props.put("dasu.reportingTimeout", reportingTimeout);
			props.put("dasu.maxPings", maxPings);
			props.put("dasu.maxAMExperiments", maxAMExperiments);
			props.put("dasu.maxMemJVM", maxMemJVM);
			props.put("dasu.uuid", uuid);
			props.put("dasu.UUIDid", UUIDid);
			props.put("dasu.runExperiments", runExperiments);
			props.put("dasu.ignoreStateChangeRST", ignoreStateChangeRST);
			props.put("dasu.sampleInterval", sampleInterval);
			props.put("dasu.schedulerInterval", schedulerInterval);
			props.put("dasu.messageManagerInterval", messageManagerInterval);
			props.put("dasu.AMECheckupInterval", AMECheckupInterval);
			props.put("dasu.config.dynamic.resource", dasuConfigDynamicResource);
			props.put("dasu.stateChangeRSTInterval", stateChangeRSTInterval);
			props.put("dasu.AMCSleepInterval", AMCSleepInterval);
			props.put("dasu.managerSleepInterval", managerSleepInterval);
			props.put("dasu.db.staticFlushInterval", dbStaticFlushInterval);
			props.put("dasu.db.managerSleepInterval", DBManagerSleepInterval);
			props.put("dasu.maxLogEntries", maxLogEntries);
			props.put("dasu.maxDataFileSize", maxDataFileSize);
			props.put("dasu.maxHistoricSignals", maxHistoricSignals);
			props.put("dasu.maxNumEntriesPerHistoricSignal", maxNumEntriesPerHistoricSignal);
			props.put("dasu.inference.version", inferenceVersion);
			props.put("dasu.RSTimeWindowSize", RSTimeWindowSize);
			props.put("dasu.RSTimeSpan", timeSpan);
	
		} catch (Exception e) {
			e.printStackTrace();
		}

		for (String filename : propertiesLocations){
			try {
				
				File f = new File(filename);
				FileWriter fow = new FileWriter(f);
				props.write(fow);
				fow.close();
				break;
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (JSONException e) {
				e.printStackTrace();
			}
		}
		
	}
	
	private boolean loadPropertiesJsonFromLocation() {
		
		LOGGER.info("CONFIG: loadPropertiesJsonFromLocation()");
		
		boolean found = false;
		try {
			for (String location : propertiesLocations) {
				File properties_file = new File(location);
				// if properties file exists on its own then override any
				// properties file potentially held within a jar
				if (properties_file.exists()) {
					FileInputStream fis = null;

					try {
						fis = new FileInputStream(properties_file);						
				    	BufferedReader bufferedreader = new BufferedReader(
				    			new InputStreamReader(fis));
						String string;
				    	StringBuilder sb = new StringBuilder();
				    	while ((string = bufferedreader.readLine()) != null) {	    	    
				    	    sb.append(string);
				    	}
				    	
						propsJson = new JSONObject(sb.toString());						
						found = true;
						LOGGER.info("CONFIG: Succesfully loaded file " + location );
						break;
					}
					catch (Exception e) {
						
						String msg = "Can't read '" + location + "' for plugin '"
								+ Main.getPluginInterface().getPluginName() + "': file may be missing";
				
						LOGGER.warning("CONFIG: " + msg);
						LOGGER.warning("CONFIG: " + e.getMessage());
						
					} finally {
						if (fis != null) {
							fis.close();
						}
					}
				}
			}
			
			if (!found) {
				// properties file wasn't there, so use the jar
				ClassLoader classLoader = Main.getPluginInterface().getPluginClassLoader();

				if (classLoader instanceof URLClassLoader) {
					URLClassLoader current = (URLClassLoader) classLoader;

					URL url = current.findResource("dasu.properties.default");

					if (url != null) {
						
				    	BufferedReader bufferedreader = new BufferedReader(
				    			new InputStreamReader(url.openStream()));
						String string;
				    	StringBuilder sb = new StringBuilder();
				    	while ((string = bufferedreader.readLine()) != null) {	    	    
				    	    sb.append(string);
				    	}
				    	propsJson = new JSONObject(sb.toString());
				    	LOGGER.info("CONFIG: Succesfully loaded file dasu.properties.default from jars" );
				    	
					} else {
						throw (new Exception(
								"failed to load dasu.properties.default from jars"));
					}
				}
			}
			
		
		} catch (Throwable e) {
			Debug.printStackTrace(e);
	
			String msg = "Can't read 'dasu.properties.default' for plugin '"
					+ Main.getPluginInterface().getPluginName() + "': file may be missing";
	
			LOGGER.warning("CONFIG: " + msg);
		}
		
		return found;
		
	}

	protected void loadPropertiesJson() {
		updatePropertiesJson(propsJson, true);
	}
	
	public void updatePropertiesJson() {
		writePropertiesJson(propsJson);
	}

	public void checkPropertiesPreviousVersionDasu(String domain) {
			
		try {
			
			String propsFileName = Main.getDir() + File.separator  + propertiesFileName.concat("." + domain);
			File f = new File(propsFileName);
			if ( f.exists() ) {
				Properties p = new Properties();
				FileInputStream fis = new FileInputStream(f);
				p.load(fis);
				//System.out.println(p);
				
				String uuid = null;
				Integer UUIDid = -1;
				
				if ( p.containsKey("dasu.uuid") ) {
					uuid = p.getProperty("dasu.uuid");
					DasuConfiguration.getInstance().setUUID(uuid);
				}
				if ( p.containsKey("dasu.UUIDid") ) {
					UUIDid = Integer.parseInt(p.getProperty("dasu.UUIDid"));
					DasuConfiguration.getInstance().setUUIDid(UUIDid);
				}
				
				ReportableFact rf = new ReportableFact("dasuUpdated");
				rf.getContentsMap().put("timestamp", Util.currentDasuTime());
				rf.getContentsMap().put("uuid", uuid);
				rf.getContentsMap().put("UUIDid", UUIDid);
				Statistics.getInstance().addReportableFact(rf);
				
				f.delete();				
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	public boolean initializeProperties(String domain) {
		
		checkPropertiesPreviousVersionDasu(domain);
		
		//propertiesFileName = propertiesFileName.concat("." + domain);
		String asn = NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS();
		if ( !Util.isNumeric(asn) ) 
			asn = "0";
			
		propertiesFileName = propertiesFileName.concat(".AS" + asn);		
		propertiesLocations.add(Main.getDir()
				+ java.io.File.separator + propertiesFileName);

		/* load properties */
		return loadPropertiesJsonFromLocation();
		
	}
	

}
