/**
 * Dasu Project
 *
 * File:         DasuManager.java
 * RCS:          $Id$
 * Description:  DasuManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:27:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.File; 
import java.security.KeyStore;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Random;
import java.util.TreeMap;
import java.util.Map.Entry;
import java.util.concurrent.TimeUnit;

import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.config.ParameterListener;
import org.gudy.azureus2.core3.logging.Logger;

import edu.northwestern.cs.aqualab.Cattle.HttpsChannel;
import edu.northwestern.dasu.DasuData.DasuDataListener;
import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.database.DBReporterManager;
import edu.northwestern.dasu.drools.FactDidEvent;
import edu.northwestern.dasu.drools.FactMeasurementLpsd;
import edu.northwestern.dasu.drools.FactMeasurementSystem;
import edu.northwestern.dasu.drools.FactMeasurementUPnP;
import edu.northwestern.dasu.drools.RuleEngine;
import edu.northwestern.dasu.measurement.*;
import edu.northwestern.dasu.measurement.ame.AMEManager;
import edu.northwestern.dasu.measurement.monitors.DownloadMonitor;
import edu.northwestern.dasu.measurement.monitors.ResetMonitor;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.measurement.types.TimeType;
import edu.northwestern.dasu.netstat.NetstatBanwidthPerformer;
import edu.northwestern.dasu.profilers.Profiler;
import edu.northwestern.dasu.profilers.ResourceProfiler;
import edu.northwestern.dasu.signal.statistics.SignalManipulation;
import edu.northwestern.dasu.simulator.InterferenceSimulatorV3;
import edu.northwestern.dasu.stats.*;
import edu.northwestern.dasu.timer.AzureusTimer;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.ui.BarPlot;
import edu.northwestern.dasu.upnp.UPnPBanwidthPerformer;
import edu.northwestern.dasu.util.DasuClassLoader;
import edu.northwestern.dasu.util.PluginInterface;
import edu.northwestern.dasu.util.UpdaterV2;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 * 
 *         The DasuManager class ...
 */
public class DasuManager implements Runnable {

	public final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
	
	private static boolean isActive = false;
	public Object halt = new Object();
	public static boolean isInitialized = false;
	private int sleepInterval = DasuConfiguration.getInstance().getManagerSleepInterval();
	public static boolean USE_SIMULATION_V2 = false;
	private static DasuManager self;

	/** Interference Simulator **/
	public static InterferenceSimulatorV3 interferenceSimulator = null;
	
	/** Monitor for added/removed downloads */
	public DownloadMonitor dlm;
	public NetstatBanwidthPerformer nbp;
	public UPnPBanwidthPerformer ubp;
	public ResetMonitor rm;
	
	/* DROOLS ENGINES */
	public static RuleEngine engine = null;
	private static Hashtable<String, ParameterListener> COConfigurationListeners = new Hashtable<String, ParameterListener>();
	
	private static Profiler profiler = null;
	
	/* FACTS */
	public static FactMeasurementSystem systemFact = null;
	public static FactMeasurementLpsd localPeerFact = null;
	public static FactMeasurementUPnP localNetworkFact = null;
	
	public static PersistentData persistentDataManager = new PersistentData();
	protected static PluginInterface pi = Main.getPluginInterface();
	
	//private static String persistentDataFileName = Main.getDir() + File.separatorChar + "dasu.data." + Main.getDomainName();
	private static String persistentDataFileName = Main.getDir() + File.separatorChar + "dasu.data.AS" + 
			NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS();

	private static final boolean DASU_SKIP_SERVER_SWITCH = false;
	private static TreeMap<Double, String> DasuUrls = new TreeMap<Double, String>();
	public static String currentDasuUrl = DasuConfiguration.DASU_DEFAULT_SERVER;

	/* Pertaining to updating the UI */
	public static UpdaterV2 updater = null;
	
	/* Pertaining to the simulator */
	private ITimer tickTimer;
	public static long startUpTime = Util.currentTimeMillis();
	
	/* Pertaining the Profiler*/
	protected volatile int yieldInterval = -1; //milliseconds
	
	private static DasuClassLoader dasuClassloader = null;

	/* Configuration service*/
	public static HttpsChannel configService = null;
	
	public static AMEManager ameManager = null;
		
	ITimer AMETJVMStartTimer = null;
	ITimer PersistentDataDumpTimer = null;
	
	public void initialize() {
		
		// Initialize
		initializeConfigService();	
		
		// Start updating the UI and rules
		updater = new UpdaterV2(LOGGER);
		
		if (!Main.SKIP_UPDATE_UI_RULES) {			
			updater.run(); // just inline this :) don't run as a thread.
		}
		
		if(!updater.isReady() || !updater.haveValidFile()) {
			LOGGER.severe("Updater failed to find a working set of rules/UI.");
		}
		
		dasuClassloader = new DasuClassLoader(this.getClass().getClassLoader(), 
				updater.getJarPath(DasuConfiguration.getInstance().getInferenceVersion()) );
		dasuClassloader.loadExternalClasses();
		
		getDasuServerList();
		LOGGER.info("DEBUG: loading persistent data");
		Object pd = Main.loadObject(persistentDataFileName);
		if (pd != null)  {
			persistentDataManager = (PersistentData) pd;

			//if file is too large, empty reportable data
			File f = new File(persistentDataFileName);
			if ( f.exists() && f.isFile() && f.length() > DasuConfiguration.getInstance().getMaxDataFileSize() ) {
				persistentDataManager.getDbIdMaps().clear();
				persistentDataManager.getDbQueue().clear();
				persistentDataManager.getReports().clear();
			 	persistentDataManager.getBaos().clear();
			 	persistentDataManager.getReportableFacts().clear();
			 	
			 	LOGGER.severe("Data file is too large, overriding...");
			}		
		}
		
	}
	
	public DasuManager() {
		
		self = this;
		isActive = true;

		initialize();
		StatManager.registerUsage();
		StatManager.start();
		
	}

	public static DasuManager getInstance() {
		return self;
	}
	
	private static void initializeConfigService() {
        
		KeyStore ks = DasuConfiguration.getInstance().getKeyStore();
        HttpsChannel.addKeyStore(ks);
        configService = new HttpsChannel(DasuConfiguration.getInstance().DASU_CONFIG_SERVER_HOST,
        		DasuConfiguration.getInstance().DASU_CONFIG_SERVER_PORT, 
        		DasuConfiguration.getInstance().DASU_CONFIG_SERVER_RESOURCE);
		
	}
	
	public HttpsChannel getConfigService() {
		return configService;
	}
	
	public static ClassLoader getDasuClassLoader() {		
		return dasuClassloader;
	}
	
	public static PersistentData getPersistentDataManager() {
		return persistentDataManager;
	}

	private void launchConfigurationChangeListener() {

		final HashMap<String, String> paramsEvent = new HashMap<String, String>();
		paramsEvent.put("Max Download Speed KBs", "MaxDownloadChanged");
		paramsEvent.put("Max Upload Speed KBs", "MaxUploadChanged");
		paramsEvent.put("Max.Peer.Connections.Total", "MaxConnectionsChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent", "MaxConnectionsPerTorrentChanged");
		paramsEvent.put("Max Seeds Per Torrent", "MaxSeedsPerTorrentChanged");
		paramsEvent.put("Use Lazy Bitfield", "LazyBitfieldChanged");
		paramsEvent.put("network.transport.encrypted.require", "EncryptionRequiredChanged");
		paramsEvent.put("network.transport.encrypted.min_level", "EncrytionLevelChanged");
		paramsEvent.put("network.transport.encrypted.fallback.outgoing", "EncryptionOutgoingChanged");
		paramsEvent.put("network.transport.encrypted.fallback.incoming", "EncryptionIncominChanged");
		paramsEvent.put("network.transport.encrypted.use.crypto.port", "EncryptionCryptoPortChanged");
		paramsEvent.put("TCP.Listen.Port", "TCPPortChanged");
		paramsEvent.put("UDP.Listen.Port", "UDPPortChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent.When.Seeding.Enable", "MaxConnectionsPerTorrentWhenSeedingChanged");
		paramsEvent.put("enable.seedingonly.maxuploads", "MaxUploadWhenSeedingChanged");
		
		for (String param : paramsEvent.keySet()) {
			
			ParameterListener pl = new ParameterListener() {
				public void parameterChanged(String parameterName) {					
					AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent(paramsEvent.get(parameterName), -1L));
				}
			};
			
			COConfigurationManager.addParameterListener(param, pl);
			COConfigurationListeners.put(param, pl);
			
		}		

	}
	
	private void printDebug() {
	
		if ((localPeerFact != null) && (systemFact != null)) {

			StringBuilder sb = new StringBuilder();
			
			sb.append("local lpsd: avgUploadRate: "
					+ (localPeerFact.getBytesSent()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localPeerFact.getBytesReceived()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localPeerFact.getShortDeltaTime()) + "\n");
			
			sb.append("local netstat: avgUploadRate: "
					+ (systemFact.getBytesSent()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (systemFact.getBytesReceived()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (systemFact.getShortDeltaTime()) + "\n");
			
			sb.append("local upnpstat: avgUploadRate: "
					+ (localNetworkFact.getTotalBytesSent()/(localNetworkFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localNetworkFact.getTotalBytesReceived()/(localNetworkFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localNetworkFact.getShortDeltaTime()) + "\n");


			sb.append("local lpsd: uploadRate: "
					+ localPeerFact.getUploadRate()
					+ " avgUpload: "
					+ localPeerFact.getAverageUpload()
					+ " stdevUpload: "
					+ localPeerFact.getStdevUpload()
					+ " downloadRate: "
					+ localPeerFact.getDownloadRate()
					+ " avgDownload: "
					+ localPeerFact.getAverageDownload()
					+ " stdevDownload: "
					+ localPeerFact.getStdevDownload()
					+ " deltaTime: "
					+ localPeerFact.getLongDeltaTime() + "\n");
			
			sb.append("local netstat: bytesSent: "
					+ systemFact.getBytesSent()
					+ " avgBytesSent: "
					+ systemFact.getAverageBytesSent()
					+ " stdevBytesSent: "
					+ systemFact.getStdevBytesSent()
					+ " bytesReceived: "
					+ systemFact.getBytesReceived()
					+ " avgBytesReceived: "
					+ systemFact.getAverageBytesReceived()
					+ " stdevBytesReceived: "
					+ systemFact.getStdevBytesReceived()
					+ " deltaTime: "
					+ systemFact.getLongDeltaTime() + "\n");
			
			sb.append("local lpsd: numTorrents: "
					+ localPeerFact.getNumTorrents()
					+ " numConnctedSeeds: "
					+ localPeerFact.getNumConnectedSeeds()
					+ " avgNumConnectedSeeds: "
					+ localPeerFact.getAverageNumConnectedSeeds()
					+ " numConnctedLeechers: "
					+ localPeerFact.getNumConnectedLeechers()
					+ " avgNumConnectedLeechers: "
					+ localPeerFact.getAverageNumConnectedLeechers()
					+ " maxPeerConnections: "
					+ localPeerFact.getMaxPeerConnectionsGlobal()
					+ " maxPeerConnectionsPerTorrent: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrent()
					+ " maxPeerConnectionsPerTorrentSeeding: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrentSeeding()
					+ " maxUploadRateSeeding: "
					+ localPeerFact.getMaxUploadRateSeeding()
					+ " maxSeedsPerTorrent: "
					+ localPeerFact.getMaxSeedsPerTorrent()
					+ " lazyBitfield: "
					+ localPeerFact.isLazyBitfield()
					+ " requireEncryption: "
					+ localPeerFact.isRequireEncryption()
					+ " encryptionMinLevel: "
					+ localPeerFact.getEncryptionMinLevel()
					+ " allowNonEncryptedOut: "
					+ localPeerFact.isAllowNonEncryptedOut()
					+ " allowNonEncryptedIn: "
					+ localPeerFact.isAllowNonEncryptedIn()
					+ " useCryptoPortExtension: "
					+ localPeerFact.isUseCryptoPortExtension()
					+ " tcpPortIn: " 
					+ localPeerFact.getTcpPortIn()
					+ " udpPortIn: "
					+ localPeerFact.getUdpPortIn() + "\n");
			
			LOGGER.info("DEBUG: " + sb.toString());
		}
		
		dlm.printDebug();		

		if (engine != null)
			engine.displayFacts();
		
	}

	
	public String twoDigit(Integer number) {
		
		if ( number < 10 && number >= 0) 
			return "0" + number;
		
		return number.toString();
	}
	
	public String dateToString(long date) {
		
		Calendar c = Calendar.getInstance();
		c.setTimeInMillis(date);
		
		String ret = "";
		
		ret = ret.concat(twoDigit(c.get(Calendar.MONTH)+1));
		ret = ret.concat(twoDigit(c.get(Calendar.DAY_OF_MONTH)));		
		ret = ret.concat(twoDigit(c.get(Calendar.YEAR)));
		ret = ret.concat(twoDigit(c.get(Calendar.HOUR_OF_DAY)));
		ret = ret.concat(twoDigit(c.get(Calendar.MINUTE)));
		ret = ret.concat(twoDigit(c.get(Calendar.SECOND)));
		ret = ret.concat(twoDigit(c.get(Calendar.MILLISECOND)));
		//ret = ret.concat(twoDigit(c.get(Calendar.ZONE_OFFSET)));
		
		return ret;

	}
 	
	public void testing() {
		
		//StressTestModules.testDns(null);
		//StressTestModules.testHttp(null);
		//StressTestModules.testNdt(null);
		//StressTestModules.testPing(null);
		//StressTestModules.testTrace(null);
		//if ( true )
		//	return;
		  
//		ITimer testTimer = Main.createTimer("testTimer");
//		testTimer.addEvent(System.currentTimeMillis() + 90*1000, 
//					new ITimerEventPerformer(){
//				
//				public void perform(ITimerEvent event) {				
//
//					System.out.println("\n\n\n\n\n\n\n################hola: " + 
//							new Date(Util.currentGMTTime()) + "##############\n\n\n\n\n\n\n\n");
//
//					Main.dasuUnload();
//					
//				}}
//			);
			
		System.out.println("currentDasuTime(): " + dateToString(Util.currentDasuTime()));
		System.out.println("currentGMTTime(): " + dateToString(Util.currentGMTTime()));
		
		//timer test
		
		ITimer testTimer = new AzureusTimer(Main.getPluginInterface().getUtilities().createTimer("testTimer"), "testTimer");
		Calendar testCal = Calendar.getInstance();
		testCal.setTimeInMillis(Util.currentGMTTime());
		testCal.set(Calendar.HOUR_OF_DAY, 20);
		testCal.set(Calendar.MINUTE, 18);
		testCal.set(Calendar.SECOND, 35);
		testCal.set(Calendar.DAY_OF_MONTH, 6);
		testCal.set(Calendar.MONTH, 1-1);
		testCal.set(Calendar.YEAR, 2012);
		
		Calendar gmtCal = Calendar.getInstance();
		gmtCal.setTimeInMillis(Util.currentGMTTime());	
		
		if ( testCal.after(gmtCal) )
			System.out.println("past!");
		
		
		/*
		System.out.println("**************************");
		System.out.println("Time Current: " + new Date(Util.currentTimeMillis()));
		System.out.println("Time Dasu: " + new Date(Util.currentDasuTime()));
		System.out.println("Time GMT: " + new Date(Util.currentGMTTime()));
		System.out.println("Time Sntp Offset: " + Util.getDasuSntpTimeOffset());
		System.out.println("Time testCal: " + new Date(testCal.getTimeInMillis()));
		
		long offset = Util.currentDasuTime() - Util.currentGMTTime() + Util.getDasuSntpTimeOffset();
		testCal.add(Calendar.MILLISECOND, (int) offset);
		System.out.println("Time testCal: " + new Date(testCal.getTimeInMillis()));
		
		System.out.println("**************************");
		
		testTimer.addEvent(testCal.getTimeInMillis(), 
				new ITimerEventPerformer(){
			
			public void perform(ITimerEvent event) {				

				System.out.println("\n\n\n\n\n\n\n################hola: " + 
						new Date(Util.currentGMTTime()) + "##############\n\n\n\n\n\n\n\n");
		        
				List<String> name = new ArrayList();
				name.add("testing.ana-aqualab.cs.northwestern.edu");
				
				ProbeTask task = new ProbeTask(-1, 0, -1l, -1l, name, ProbeType.DNS);
				AMCoordinatorV2.addProbeTask(task);
				
			}}
		);
		*/
		
		/***********************/

//		List<String> urls = Arrays.asList(new String[]{"http://google.com"});		
//		Map<String, String> headers = new HashMap();
//		headers.put("User-agent", "Mozzilla");		
//		Map<String, Object> params = new HashMap();		
//		params.put("headers", new HashMap());
//		params.put("headers", headers);		
//		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, urls, ProbeType.HTTP, TaskType.BULK, SourceType.RE, params);		
//		AMCoordinatorV2.addProbeTask(task);
		
//		AMCoordinatorV2.getInstance().addPriorityProbe("4.4.4.2", ProbeType.DNS);
		
//		Statistics.getInstance().addDnsTimeout(AMCoordinatorV2.getMyDnsServers().get(0).get(0), 
//				Main.getPublicIpAddress(), System.currentTimeMillis(), System.currentTimeMillis(), true);
		
		//AMCoordinatorV2.addPriorityProbe(null, ProbeType.NDT2);
		LOGGER.info("Azureus name: " + Main.getPluginInterface().getAzureusName());
		LOGGER.info(Main.getPluginInterface().getAzureusVersion());
		LOGGER.info(Main.getPluginInterface().getApplicationName());
		LOGGER.info(System.getProperty("os.name"));
		LOGGER.info(System.getProperty("os.arch"));
		LOGGER.info(System.getProperty("os.version"));
		LOGGER.info(System.getProperty("java.version"));
		LOGGER.info(System.getProperty("java.vendor"));
		LOGGER.info(Main.getInterfaceName());
		LOGGER.info(Main.getInterfaceMAC());
		
		//LOGGER.info("Disabling AME experiments due to testing mode...");
		DasuConfiguration.getInstance().setRunExperiments(true);
		
	//    	if (!AMCoordinatorV2.historicSampleExists("time_of_day_bw_limit_day_minute_5")) {
	//		AMCoordinatorV2.historicSampleCreate("time_of_day_bw_limit_day_minute_5", TimeType.DAY, TimeType.MINUTE, 5);
	//
	//    ArrayList maxDownloadRateLpsdPastTwoHours = (ArrayList) AMCoordinatorV2.historicSampleGetValuesSubsetPast(
	//            "time_of_day_bw_limit_day_minute_5", "maxDownloadRateLpsd", Util.currentDasuTime(), 60);
	//        System.out.println("\n\n\n\n\n\nENGINE: " + maxDownloadRateLpsdPastTwoHours);
		
	//	SignalSample ss = null;
	//	try {
	//		ss = new SignalSample("uploadRate", TimeType.DAY, TimeType.HOUR, 1, "BytesPerSecond");
	//	} catch (Exception e1) {
	//		e1.printStackTrace();
	//	}
	//	System.out.println(new Date(Util.currentDasuTime()));
	//	ss.set(Util.currentDasuTime(), 4532L);
	//	
	//	System.out.println(ss.getName());
	//	System.out.println(ss.getValuesHash());
	//	System.out.println(ss.getTimeStampsHash());
		
	//	Collection<Float> nt = AMCoordinatorV2.historicSampleGetValues("netstat_max_downloadRate_hour_minute_1");
	//	Collection<Float> bt  = AMCoordinatorV2.historicSampleGetValues("bt_max_downloadRate_hour_minute_1");
	//	System.out.println(SignalManipulation.evalAverage(new Float[]{1f, 2f, 3f}, 1));
		
	//	System.out.println(nt);
	//	System.out.println(bt);
	//	
		
//		Float[] maxBT = new Float[]{3.0f, 3.0f, 2.8f, 2.9f, 2.8f, 3.0f, 2.9f, 3.0f, 2.7f, 3.1f, 3.0f, 3.0f, 2.9f, 2.9f, -0f, -0f, 3.1f, 3.0f, 3.0f, 2.9f, 3.0f, 2.8f, 2.9f, 2.8f};
//		Float[] pctlBT = new Float[]{2.9f, 2.6f, 2.8f, 2.9f, 2.7f, 3.0f, 2.8f, 2.9f, 2.7f, 3.0f, 3.0f, 2.5f, 2.0f, 2.2f, -0f, -0f, 2.3f, 2.4f, 2.6f, 1.8f, 2.2f, 2.5f, 2.7f, 2.5f};
//		Float[] maxNetstat = new Float[]{3.0f, 3.9f, 3.0f, 3.0f, 2.9f, 2.8f, 2.9f, 3.0f, 2.9f, 2.9f, 3.2f, 3.4f, 3.0f, 2.9f, -0f, -0f, 3.0f, 2.9f, 2.8f, 3.2f, 2.7f, 2.9f, 2.8f, 3.0f};
		
		Float[] maxBT = new Float[]{3.0f, 3.0f, 2.8f, 2.9f, 2.8f, 3.0f, 2.9f, 3.0f, 2.7f, 3.1f, 3.0f, 3.0f, 2.9f, 2.9f, -0f, -0f, -0f, 3.0f, -0f, 2.9f, 3.0f, 2.8f, 2.9f, 2.8f};
		Float[] pctlBT = new Float[]{2.9f, 2.60f, 2.8f, 2.9f, 2.7f, 3.0f, 2.8f, 2.9f, 2.7f, 3.0f, 3.0f, 2.5f, 2.0f, 2.2f, -0f, -0f, -0f, 2.4f, -0f, 1.8f, 2.2f, 2.5f, 2.7f, 2.5f};
		Float[] maxNetstat = new Float[]{3.0f, 3.9f, 3.0f, 3.0f, 2.9f, 2.8f, 2.9f, 3.0f, 2.9f, 2.9f, 3.2f, 3.4f, 3.0f, 2.9f, -0f, -0f, -0f, 2.9f, -0f, 3.2f, 2.7f, 2.9f, 2.8f, 3.0f};

		
		for ( int i=0; i < maxBT.length; i++) {
			maxBT[i] = maxBT[i]*100000;
			pctlBT[i] = pctlBT[i]*100000;
			maxNetstat[i] = maxNetstat[i]*100000;
		}
		
		Calendar c = Calendar.getInstance();
		c.set(Calendar.HOUR_OF_DAY, 0);
		
		Long[] timeRecorded = new Long[maxBT.length];
		for (int i=0; i < maxBT.length; i++ )
			timeRecorded[i] = c.getTimeInMillis() + i*1000*60*60;
				
        if (!AMCoordinatorV2.historicSampleExists("time_of_day_bt_max_download_hour_minute_5")) {
	        AMCoordinatorV2.historicSampleCreate("time_of_day_bt_bw_limit_max_day_hour_1", TimeType.DAY, TimeType.HOUR, 1);
	        AMCoordinatorV2.historicSampleCreate("time_of_day_bt_bw_limit_percentile_day_hour_1", TimeType.DAY, TimeType.HOUR, 1);
	        AMCoordinatorV2.historicSampleCreate("time_of_day_netstat_limit_max_day_hour_1", TimeType.DAY, TimeType.HOUR, 1);
	        AMCoordinatorV2.historicSampleCreate("time_of_day_netstat_limit_percentile_day_hour_1", TimeType.DAY, TimeType.HOUR, 1);
        }
		
		for ( int i=0; i < maxBT.length; i++) {
            AMCoordinatorV2.historicSampleSetValue( "time_of_day_bt_bw_limit_max_day_hour_1", "maxDownloadRateLpsd", timeRecorded[i],                    
            		maxBT[i], "bytesPerSecond");
            AMCoordinatorV2.historicSampleSetValue( "time_of_day_bt_bw_limit_percentile_day_hour_1", "maxDownloadRateLpsd", timeRecorded[i],                    
            		pctlBT[i], "bytesPerSecond");
            AMCoordinatorV2.historicSampleSetValue( "time_of_day_netstat_limit_max_day_hour_1", "maxDownloadRateNetstat", timeRecorded[i],                    
            		maxNetstat[i], "bytesPerSecond");
		}

		BarPlot plot = new BarPlot("simulator.timeofday.maxDownloadRateLpsd.24", "maxDownloadRateLpsd", null, maxBT, "(hour)", "(MB/s)");
        plot.addOption("hoverable", true);        
		DasuData.getInstance().update(plot.getId(), plot.generatePlot());
		
		plot = new BarPlot("simulator.timeofday.pctlMaxDownloadRateLpsd.24", "90th-pctl maxDownloadRateLpsd", null, pctlBT, "(hour)", "(MB/s)");
		plot.addOption("hoverable", true);
		DasuData.getInstance().update(plot.getId(), plot.generatePlot());
		
		int intervalSize = 5;
		for ( int i=0; i < pctlBT.length; i+=intervalSize) {
			System.out.println("subsignal: " + SignalManipulation.subSignal(pctlBT, i, intervalSize));
			System.out.println("subsignal avg: " + SignalManipulation.getAverage(SignalManipulation.subSignal(pctlBT, i, intervalSize)) );
			//System.out.println("subsignal slope: " + SignalManipulation.getSlope(SignalManipulation.subSignal(pctlBT, i, intervalSize),  
			//		Arrays.asList(timeRecorded)) );
		}
		
		List<Integer> peakHour = Arrays.asList(new Integer[]{15,16,17,18,19,20,21,22,23}) ;
		
		List<Float> peak = new ArrayList<Float>();
		List<Float> offPeak = new ArrayList<Float>();

		for ( int i=0; i < timeRecorded.length; i++) {
			c.setTimeInMillis(timeRecorded[i]);
			int hour = c.get(Calendar.HOUR_OF_DAY);
			if ( peakHour.contains(hour)) {
				System.out.println("peak?: " + hour + " " + timeRecorded[i]);
				if (pctlBT[i] > -0)	
					peak.add( pctlBT[i] );
				
			}			
			else {
				if (pctlBT[i] > -0)	
					offPeak.add( pctlBT[i] );
			}
		}

		//compute avg, median and stdev for peak and off-peak for netstat, max and bt		
		System.out.println("peak size: " + peak.size() + " peak avg: "  + SignalManipulation.getAverage(peak) + " median: " + SignalManipulation.getMedian(peak) + " stdev: " + SignalManipulation.getStdDev(peak));
		System.out.println("offpeak size: " + offPeak.size() + " offpeak avg: " + SignalManipulation.getAverage(offPeak) + " median: " + SignalManipulation.getMedian(offPeak) + " stdev: " + SignalManipulation.getStdDev(offPeak));		
		
		
/*		
		Float[] nt = new Float[]{77.0f, 65.0f, 55.0f, 58.0f, 52.0f, 52.0f, 76.0f, 61.0f, 58.0f, 59.0f, 
				67.0f, 62.0f, 67.0f, 76.0f, 77.0f, 65.0f, 55.0f, 58.0f, 51.0f, 52.0f, 73.0f, 66.0f, 78.0f, 57.0f, 
				67.0f, 72.0f, 52.0f, 76.0f, 58.0f, 59.0f, 67.0f, 62.0f, 67.0f, 76.0f, 77.0f, 65.0f, 55.0f, 
				58.0f, 51.0f, 52.0f, 73.0f, 66.0f, 78.0f, 57.0f, 67.0f, 72.0f, 52.0f, 76.0f, 61.0f, 58.0f, 59.0f, 
				67.0f, 62.0f, 67.0f, 76f, 73.0f, 66.0f, 78.0f, 57.0f};
		
		Float[] bt = new Float[]{
				40.0f, 40.0f, 43.0f, 41.0f, 44.0f, 50.0f, 55.0f, 42.0f, 12.0f, 4.0f, 10.0f, 12.0f, 11.0f, 14.0f, 
				44.0f, 50.0f, 53.0f, 48.0f, 40.0f, 53.0f, 41.0f, 44.0f, 42.0f, 50.0f, 55.0f, 42.0f, 59.0f,
				44.0f, 50.0f, 66.0f, 51.0f, 40.0f, 43.0f, 41.0f, 44.0f, 42.0f, 50.0f, 55.0f, 42.0f, 59.0f, 44.0f, 50.0f, 
				66.0f, 59.0f, 66.0f, 44.0f, 50.0f, 53.0f, 48.0f, 40.0f, 66.0f, 44.0f, 50.0f, 53.0f, 48.0f, 53.0f,
				42.0f, 50.0f, 55.0f};
	
		Float[] signal = new Float[]{0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, .0f, 24.0f, 34f, 13f, 44f, 53f}; 
	
		Long[] time = new Long[bt.length];
		for (int i=0; i < bt.length; i++ )
			time[i] = Util.currentDasuTime()+i*1000*60;
	
		LOGGER.info("signal: " + Arrays.asList(signal));
		LOGGER.info("getMean: " + SignalManipulation.getAverage(signal));
		LOGGER.info("getStdev: " + SignalManipulation.getStdDev(signal));
		LOGGER.info("getPercentile: " + SignalManipulation.getPercentile(signal, 95));
		LOGGER.info("getPercentileAboveValue: " + SignalManipulation.getPercentileAboveValue(signal, 95, SignalManipulation.getAverage(signal)));
		LOGGER.info("getPercentileIndexPosition: " + SignalManipulation.getPercentileIndexPosition(signal, 95));
		
		MovingWindow mw =  new MovingWindow(bt, time, 5);
		LOGGER.info(mw.toString());	
		
		LOGGER.info("correlation 5: " + SignalManipulation.getCorrelationCollection(nt, bt, 5));
		LOGGER.info("correlation 5 cumulative: " + SignalManipulation.getCorrelationCollection(
				SignalManipulation.getCumulative(nt), SignalManipulation.getCumulative(bt), 5));
		
		LOGGER.info("time: (" + time.length + ") " + Arrays.asList(time));
		LOGGER.info("cumulative bt: (" + SignalManipulation.getCumulative(bt).size() + ") " + SignalManipulation.getCumulative(bt));
	
		LOGGER.info("eval Slope 1: " + SignalManipulation.getSlopeCollection( 
				SignalManipulation.getCumulative(bt), Arrays.asList(time), 2));
		
		LinePlot plot1 = new LinePlot("bt.lpsd.plot.test","test1",  mw.getTime(), mw.getMean(), "(time)", "(kbps)", LinePlot.FORMAT_NUMBER);
		DasuData.getInstance().update(plot1.getId(), plot1.generatePlot());		
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot1.generatePlot());
		
		LinePlot plot2 = new LinePlot("bt.lpsd.plot.test2", "test2", null, mw.getSignal(), "(time)", "(kbps)", LinePlot.FORMAT_NUMBER);
		DasuData.getInstance().update(plot2.getId(), plot2.generatePlot());
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot2.generatePlot());

		LinePlot plot3 = new LinePlot("bt.lpsd.plot.test3","test3", null, mw.getMean(), "(time)", "(kbps)", LinePlot.FORMAT_NUMBER);
		plot3.addSeries("title", mw.getStdDev());
		DasuData.getInstance().update(plot3.getId(), plot3.generatePlot());
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot3.generatePlot());
		
		BarPlot plot4 = new BarPlot("bt.lpsd.plot.test4", "test", new String[]{"Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat"}, new Float[]{6f, 5f, 4f, 3f, 2f, 1f, 0f}, "(test)", "(test)");
		plot4.addOption("horizontal", true);
		DasuData.getInstance().update(plot4.getId(), plot4.generatePlot());		
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot4.generatePlot());
		
		BarPlot plot5 = new BarPlot("bt.lpsd.plot.test5", "test", new String[]{"Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat"}, new Float[]{6f, 5f, 4f, 3f, 2f, 1f, 0f}, "(test)", "(test)");
		plot5.addSeries("test2", new Float[]{6f, 5f, 6f, 5f, 6f, 5f, 6f});
		DasuData.getInstance().update(plot5.getId(), plot5.generatePlot());		
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot5.generatePlot());
		
		*/
		
	}
	
	public void startUiUpdater() {
		
		tickTimer = Main.createTimer("NonSimulatorUiTimer");
		tickTimer.addPeriodicEvent( 100,
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {

						Long ret = Util.currentTimeMillis();
						long millis = ret - DasuManager.startUpTime; 
						String date = String.format("%d weeks, %d days, %d hour, %d min, %d sec",
								TimeUnit.MILLISECONDS.toDays(millis)/7,
								TimeUnit.MILLISECONDS.toDays(millis) - 7*(TimeUnit.MILLISECONDS.toDays(millis)/7),  		
								TimeUnit.MILLISECONDS.toHours(millis) - TimeUnit.DAYS.toHours(TimeUnit.MILLISECONDS.toDays(millis)),
								TimeUnit.MILLISECONDS.toMinutes(millis) - TimeUnit.HOURS.toMinutes(TimeUnit.MILLISECONDS.toHours(millis)),			    
							    TimeUnit.MILLISECONDS.toSeconds(millis) - TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(millis))
							);
						
						DasuData.getInstance().update("simulator.elapsed.time", date);
												
					}
		});		
	}
	
	public void startSimulation(Object simulationId, String baseDir) {
				
		Object seconds = DasuData.getInstance().get("simulator.second.equivalence");
		//seconds = 300;
		
		if (seconds == null)
			seconds = 300;
		
		if ( simulationId == null)
			simulationId = "65601790"; //default
			
		if (interferenceSimulator == null) {
			//interferenceSimulator = new InterferenceSimulatorV3("/files2/tmp/dasu-simulator/65601790/", 5*60);
			interferenceSimulator = new InterferenceSimulatorV3(baseDir + "/" + simulationId, Integer.parseInt(seconds.toString()));			
		    DasuData.getInstance().update("simulator.second.equivalence", seconds.toString());
		}

	}
	
	
	public void launchSimulatorShutdownListener() {
	
		DasuData.getInstance().registerListener("simulator.request.restart", new DasuDataListener() {

			public void updated(String key, Object val) {
				
					if ( ((Boolean)val) ) {
						
						try {
							
							LOGGER.info("SIMULATOR: DasuManager.class: saving simulation variables (do and id)");
							DasuData.getInstance().markPersistent("simulator.simulation.id");
							LOGGER.info("SIMULATOR: DasuManager.class: shutdown!!");
							//PluginManager.restartAzureus();
							//PluginManager.stopAzureus();
							
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
			});
	}
	
	protected void setYieldInterval(int timeInMillis) {
		yieldInterval = timeInMillis;
	}
	
	protected void clearYieldInterval() {
		yieldInterval = -1;
	}

	public void startAME() {
				
		if ( DasuConfiguration.getInstance().isRunExperiments() ) {
			try {
				
				if ( !AMEManager.getInstance().isJvmActive() ) {
					LOGGER.info("AME: Starting/Restarting AMEManager " + new Date());
					ameManager = AMEManager.getInstance();
					ameManager.initialize();
				}
			}
			catch (Exception e) {
				LOGGER.severe("AME: ERROR: couldn't initilize AMEManager!!! " + e.getLocalizedMessage() );
			}
		}
		else {
			LOGGER.info("Running experiments disabled by user...");
		}
		
	}
	
	public void stopAME() {

		if ( ameManager != null && ameManager.isActive() ) {
			ameManager.stop();
			ameManager = null;
		}

	}
	
	public void startMessageManager() {
		
		if ( DasuConfiguration.getInstance().isEnablePopups() )
			if ( !DasuMessageManager.getInstance().isActive() )
				pi.getUtilities().createThread("DasuMessageManager", DasuMessageManager.getInstance());
	}
	
	public void stopMessageManager() {
		DasuMessageManager.getInstance().stop();
	}

	/**
	 * Perform periodic data-collections
	 */
	  
	  public void run() {
		
		//testing();

		Main.createThread("DasuResourceProfiler", ResourceProfiler.getInstance());

		if (!Logger.isEnabled())
			COConfigurationManager.setParameter("Logger.Enabled", true);

		LOGGER.info("DYNAMIC: loading persistentData" );
		
		synchronized(persistentDataManager) {
			Iterator<String> iter = persistentDataManager.UIValues.keySet().iterator();
			synchronized(DasuData.values) {
				while (iter.hasNext()) {
					String key = iter.next();
					DasuData.values.put(key, persistentDataManager.UIValues.get(key));
					iter.remove();
				}
			}
		}
		
		/** SIMULATION **/

		//DasuData.values.put("simulator.simulation.id", "65601790");
		//DasuData.values.put("simulator.simulation.id", "370433691");
		//DasuData.getInstance().update("simulator.play.command", "pause");
		
		Object simulationId = null;
							
		if (USE_SIMULATION_V2 && (simulationId = DasuData.getInstance().get("simulator.simulation.id")) != null) {
			startSimulation(simulationId, "/files2/tmp/dasu-simulator/");
			launchSimulatorShutdownListener();
		}
		else
			startUiUpdater();		

		/*************/
		
		// Has to be before AMC does init
		nbp = new NetstatBanwidthPerformer(Main.getInterfaceName(), "Netstat BW");
		nbp.runOnce(1000);
		
		ubp = new UPnPBanwidthPerformer(Main.getInterfaceName(), "UPnP BW");
		ubp.runOnce(1000);
	
		//getDasuServerList();
				
		Main.createThread("DasuAMCoordinator", AMCoordinatorV2.getInstance());
				
		LOGGER.info("DEBUG: DasuManager.class: initializing facts");
		systemFact = new FactMeasurementSystem(new NetstatSystemStats()); /**/
		localPeerFact = new FactMeasurementLpsd(new DynamicLocalPeerStat(Util.currentGMTTime())); /**/
		localNetworkFact = new FactMeasurementUPnP(new UPnPSystemStats()); /**/


		LOGGER.info("DEBUG: DasuManager.class: starting rule engine");
		try {

			engine = new RuleEngine(SourceType.RE);
			engine.setDaemon(true);
			engine.initialize();
			
			//wait for AMC to start before starting the engine			
			int wait = 0;
			while ( !AMCoordinatorV2.isActive() ) {
				Thread.sleep(2*1000);			
				LOGGER.info("DYNAMIC: AMC hasn't started, waiting...");
				
				if ( (wait+=2) > 120 ) {
					LOGGER.severe("The AMC never started... waiting 2 minutes! unloading Dasu!!");					
					Main.dasuUnload();
					return;
				}
			}

			if (!engine.isAlive() && engine.isInitialized()) 
				engine.start();
			
			if(Main.DEBUG) {
				profiler = new Profiler();
				engine.profile(profiler);
				AMCoordinatorV2.profile(profiler);
			}
		}
		catch(Exception e) {
			e.printStackTrace();
			// TODO: If we get here, we couldn't load any rules. Make this stop the startup process...
		}
		LOGGER.info("DEBUG: started rule engine");
						
		pi.getUtilities().createThread("DasuCronScheduler", DasuCronScheduler.getInstance());
		startMessageManager();
				
		if (DasuConfiguration.getInstance().isRunExperiments()) {
			
			if ( !USE_SIMULATION_V2 ) {
				
				//Start JVM after a delay for AMECheckupInterval seconds	
				AMETJVMStartTimer = Main.createTimer("DasuAMEJVMTimer");
				Integer AMECheckupInterval = DasuConfiguration.getInstance().getAMECheckupInterval();
				if (Main.AME_DEBUG_SKIP_DOWNLOAD)
					AMECheckupInterval = 15;

				LOGGER.info("AME: Starting AMEManager after " + AMECheckupInterval + " seconds..." + new Date());
				AMETJVMStartTimer.addPeriodicEvent(AMECheckupInterval*1000, new ITimerEventPerformer(){ 

						public void perform(ITimerEvent event) {
				
							if ( !Main.isShuttingDown()) {								
								DasuManager.getInstance().startAME();
							}
						}
				});
				
			}
			else
				LOGGER.info("SIMULATOR: ActiveMeasurement Engine halted due to SIMULATION mode!!\n\n");
				
		}
		
		PersistentDataDumpTimer = Main.createTimer("DasuPersistentDataDumpTimer");		
		PersistentDataDumpTimer.addPeriodicEvent(DasuConfiguration.getInstance().getDbStaticFlushInterval()*1000, new ITimerEventPerformer() {
			
			public void perform(ITimerEvent event) {
				
				if ( !Main.isShuttingDown()) {		
					DasuManager.getInstance().writePersistentData();
				}
			}
		});
		
		isInitialized = true;
		
		Main.logPlugin("Inference engine enabled...");

		dlm = new DownloadMonitor(engine); /**/
		
		rm = new ResetMonitor(dlm); /**/
		rm.runOnce("Netstat Reset Monitor", 1000);
		
		launchConfigurationChangeListener();
					
		/*
		 * while is simulation, this while active loop should be performed 
		 */
		
		//AMCoordinatorV2.addPriorityProbe("38.102.0.97", ProbeType.NDT2);
		//ProbeTask pt = new ProbeTask(-1, -1, -1l, -1l, Arrays.asList(new String[]{"http://eecs.northwestern.edu"}), ProbeType.BROWSER);
		//AMCoordinatorV2.addProbeTask(pt);
		
		while (isActive) {

			if ( USE_SIMULATION_V2 ) {
				interferenceSimulator.increaseTick();
				interferenceSimulator.iterGetNextTickEvents();				
			}
			
			/*
			 * update measurement facts associated with the collected
			 * variables
			 */
					
			boolean again = false;

			do {
				
				if ( USE_SIMULATION_V2) {
					interferenceSimulator.iterNext();
				}
				
				rm.update(); /**/
				nbp.update(); /**/
				ubp.update(); /**/
	
				if (engine != null) {
	
					//Rpss facts are inserted by the PeerStats Manager 
					//asynchronously as peers disconnect, only for peers
					//that actually transmitted/received something from us
					
					dlm.updateFacts();
					
					DynamicLocalPeerStat dlps = new DynamicLocalPeerStat(Util.currentGMTTime());
					if ( dlps.uploadRate != -1 ) {						
						localPeerFact.addEntry(dlps); /**/
						engine.addFact(localPeerFact); /**/						
					}
					
					NetstatSystemStats nss = new NetstatSystemStats();
					if ( nss.timeRecorded != -1) {
						
						//System.out.println("total bytesReceived: " + nss.bytesReceived);
						
						if ((systemFact.entryList.size() ==0) || 
								(systemFact.entryList.size() >0) && ( systemFact.entryList.getLast().timeRecorded != nss.timeRecorded)) {
							
							systemFact.addEntry(nss); /**/
							engine.addFact(systemFact); /**/
							
						}
					}					
					//System.out.println("since last bytesReceived: " + systemFact.getBytesReceived() );					
					
					UPnPSystemStats uss = new UPnPSystemStats();
					if ( uss.timeRecorded != -1) {
						localNetworkFact.addEntry(new UPnPSystemStats()); /**/
						engine.addFact(localNetworkFact); /**/
					}
	
				}
		
				LOGGER.info("DEBUG: DasuManager: sleeping for " + sleepInterval + " seconds...");
	
				if ( USE_SIMULATION_V2 )
					LOGGER.info("SIMULATOR: DasuManager.class: BEFOREHALT: " + new Date(Util.currentTimeMillis())  + " eventQueue.size(): " + interferenceSimulator.eventQueue.size());
				
				synchronized (halt) {
	
					try {
						if (isActive && !again) {

							if (Main.DEBUG) {
								
								LOGGER.info("STOP: halting...");
								halt.wait(5 * 1000);
								LOGGER.info("STOP: halting done...");
								
							}
							else
								halt.wait(sleepInterval * 1000);
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}	
				}
				
				if ( USE_SIMULATION_V2 ) {
					LOGGER.info("SIMULATOR: DasuManager.class: AFTERHALT: " + new Date(System.currentTimeMillis()) + " eventQueue.size(): " + interferenceSimulator.eventQueue.size());
				}
	
				if (USE_SIMULATION_V2) {					

					checkPausedSimulation();
					
					if (interferenceSimulator.iterHasNext())
						again = true;
					else
						again = false;
				}
				
				if ( yieldInterval != -1) {
					try {
						Thread.sleep(yieldInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				
				if ( Main.DEBUG ) {
					printDebug();
					LOGGER.info("#######\nPERSISTENT: " + DasuData.getInstance().values + "\n########");
				}
				
			} while (again);
			
		}// while
		
		LOGGER.info("STOP: Adios...");

	}	
	
	private void checkPausedSimulation() {
		
		String command = (String) DasuData.getInstance().get("simulator.play.command");										
		while ( (command != null) && command.equals("pause") ) {

			try {
				LOGGER.info("SIMULATOR: DasuManager.class: paused!");
				Thread.sleep(1*1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			command = (String) DasuData.getInstance().get("simulator.play.command");
		}
		
		if ( command != null && command.equals("step") )
			DasuData.getInstance().update("simulator.play.command", "pause");
		
	}
	
	public boolean isRunning() {
		return isActive;
	}

	public void stop() {

		LOGGER.info("STOP: initiating DasuManager unload...");
		
		isActive = false;
		synchronized (halt) {
			halt.notify();
			LOGGER.info("STOP: halting keeps going now!");
		}

		nbp.stop();
		ubp.stop();
				
		if(Main.DEBUG) {
			if ( profiler != null)
				profiler.stop();
		}
		
		writePersistentData();
		
		DasuCronScheduler.getInstance().stop();
		stopMessageManager();
		stopAME();
		
		if (AMCoordinatorV2.isActive()) {
			AMCoordinatorV2.stop();
		}
		
		if ( ResourceProfiler.getInstance().isActive()) {
			LOGGER.info("STOP: Stopping ResourceProfiler");
			ResourceProfiler.getInstance().stop();
		}

		LOGGER.info("STOP: Stopping rule engines");

		if (engine != null && engine.isAlive()) 
			engine.stopEngine();
		
		LOGGER.info("STOP: Destryoing timers...");
		if(dlm != null) dlm.remove();		
		if(rm != null) rm.destroy();
		if (AMETJVMStartTimer != null ) AMETJVMStartTimer.destroy();		
		if (tickTimer != null) tickTimer.destroy();
		if (PersistentDataDumpTimer != null) PersistentDataDumpTimer.destroy();
		
		
		for (String param : COConfigurationListeners.keySet()) {
			COConfigurationManager.removeParameterListener(param, COConfigurationListeners.get(param));
		}

		Main.logPlugin("Inference engine disabled...");		
		try {
			finalize();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		
		LOGGER.info("STOP: DasuManager: Buh-bye");

	}
	
	public void writePersistentData() {
		
		synchronized(persistentDataManager) {
					
			for ( String key : DasuData.persistentKeys ) {
				if ( DasuData.values.containsKey(key)) {
					persistentDataManager.UIValues.put(key, DasuData.values.get(key));
				}
			}		
							
			if ( Main.DEBUG ) {
				
				StringBuilder sb = new StringBuilder();
				
				sb.append("SERIALIZING: dbQueue " + persistentDataManager.getDbQueue().size() + "\n"); 
				sb.append("SERIALIZING: reportableFacts " + persistentDataManager.getReportableFacts().size() + "\n");
				
				for ( String s : persistentDataManager.getReports().keySet()) 
					sb.append("SERIALIZING: reports " + s + " " + persistentDataManager.getReports().get(s).getJSONList().size() + "\n");						
				for ( Long l : persistentDataManager.getDbIdMaps().keySet()) 
					sb.append("SERIALIZING: dbIdMaps " + new Date(l) + " " + persistentDataManager.getDbIdMaps().get(l) + "\n");						
				for ( String s : persistentDataManager.getBaos().keySet()) 
					sb.append("SERIALIZING: baos " + s + " " + persistentDataManager.getBaos().get(s).length + "\n");
				
				LOGGER.info(sb.toString());
			}
	
			
			LOGGER.info("SERIALIZING: serializing persistent data...");
			Main.saveObject(persistentDataManager, persistentDataFileName);
		}
		
	}

	/**
	 * 
	 */
	public void getDasuServerList() {
		
		DasuUrls.put(1.0, DasuConfiguration.DASU_DEFAULT_SERVER);
		Properties servers = Main.downloadFromURL(DasuConfiguration.DASU_SERVER_LIST_LOC, 30 * 1000);
		for (Entry<Object, Object> ent : servers.entrySet()) {
			try {
				String pair[] = ((String) ent.getValue()).split(";");
				DasuUrls.put(Double.parseDouble(pair[1]), pair[0]);
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
		// remove default if there's more than one server and its weight wasn't
		// changed
		if (DASU_SKIP_SERVER_SWITCH) 
			return;
		
		if (DasuUrls.size() > 1 && DasuUrls.get(1.0).equals(DasuConfiguration.DASU_DEFAULT_SERVER))
			DasuUrls.remove(1.0);

		double val = Main.getRandom().nextDouble();
		for (Entry<Double, String> ent : DasuUrls.entrySet()) {
			if (val <= ent.getKey()) {
				currentDasuUrl = ent.getValue();
				break;
			}
		}
		
		LOGGER.info("CONFIG: DasuUrls: " + DasuUrls);
				
	}
	
	public String getNewRandomDasuUrl() {

		if (DasuUrls.size() <= 1) {
			return currentDasuUrl;
		}

		Random random = new Random();
		double index = DasuUrls.keySet().toArray(new Double[DasuUrls.size()])[ random.nextInt(DasuUrls.size()) ];
		return DasuUrls.get(index);
		
	}

	public InterferenceSimulatorV3 getSimulator() {
		return interferenceSimulator;
	}
	

}
