/**
 * Dasu Project
 *
 * File:         DBReporterManager.java
 * RCS:          $Id$
 * Description:  DBManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Dec 15, 2009 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.database;

import java.io.ByteArrayOutputStream;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DBReporterManager class periodically reports data to the Main Database.
 **/

public class DBReporterManager implements Runnable {	
	
	public final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
	private static DBReporterManager self = null;
	private DatabaseReporter reporter;
	private boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	public static int POST_THRESHOLD = 40*1024;
	public static int DROP_THRESHOLD = 1024*1024; //if baos is 1MB is size, clear it to avoid filling the memory
	long lastStaticFlushTime = -1;

	private boolean DEBUG = false;
	
	private DBReporterManager() {
		isActive = true;
		reporter = Main.getReporter();
	}

	public static DBReporterManager getInstance() {
		
		if (self == null) {
			self = new DBReporterManager();
		}
		return self;
	}

	public DatabaseReporter getReporter() {
		return reporter;
	}
	
	private Object invokeMethod(String methodName, List<Object> params) {
		
		if (DEBUG) {
			LOGGER.info("REPORTE: DBManager.class: invokeMethod: " + methodName + " " + params);
		}
		
		try {

            Class<?> paramTypes[] = new Class[params.size()];
            for (int i=0; i < params.size(); i++) { 
            	if ( params.get(i) == null) {
            		paramTypes[i] = String.class;
            		continue;
            	}
            	paramTypes[i] = params.get(i).getClass();
            	if ( paramTypes[i].getName().contains("Integer"))
            		paramTypes[i] = Integer.TYPE;
            	if ( paramTypes[i].getName().contains("Long"))
            		paramTypes[i] = Long.TYPE;
            	if ( paramTypes[i].getName().contains("Float"))
            		paramTypes[i] = Float.TYPE;
            	if ( paramTypes[i].getName().contains("Boolean"))
            		paramTypes[i] = Boolean.TYPE;
            	if ( paramTypes[i].getName().contains("Double"))
            		paramTypes[i] = Double.TYPE;

            }
            
            if ( paramTypes.length == 1 && params.get(0) instanceof FactTemplate )
            	paramTypes[0] = FactTemplate.class;
                        
            if ( methodName.equals("pushGenericReport") ) {
                Method method = WranglerReporter.getInstance().getClass().getMethod(methodName, paramTypes);
                return method.invoke(WranglerReporter.getInstance(), params.toArray());
            }
            else {
            	
            	if ( DasuConfiguration.getInstance().isEnablePhPReporter() ) {            	 
		            Method method = PhpReporter.getInstance().getClass().getMethod(methodName, paramTypes);
		            return method.invoke(PhpReporter.getInstance(), params.toArray());
            	}
            }
            
        }
        catch (Exception e) {
        	LOGGER.info("REPORTER: ERROR: " +  e.getCause() + " " +  e.getMessage() + " " + e.getLocalizedMessage());
        	e.printStackTrace();        	
        }
      
        return false;
	}
	
	public void run() {
		
		while ( !DasuManager.isInitialized) {
			try {
	        	LOGGER.info("REPORTER: Waiting for DasuManager to start...");
				Thread.sleep(2*1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		LOGGER.info("REPORTER: DasuManager started, proceed...");
		
		while (isActive) {

			int byteCount = 0;
			if ( reporter instanceof CallbackReporter) {
				
				List<Long> entries = new ArrayList<Long>();
				synchronized( ((CallbackReporter) reporter).dbQueue.keySet() ) {					
						entries.addAll(((CallbackReporter) reporter).dbQueue.keySet());
				}
				Collections.sort(entries);				
				
				TreeMap<Long, List<Object>> putBack = new TreeMap();

				/** the largest post that will be sent, 20K for now */
				while (byteCount < POST_THRESHOLD && entries.size() > 0) {

					if (!isActive)
						break;
					
					printDebug();
					
					Long time = entries.remove(0);					

					//update all ids
					batchProcessIds();
					List<Object> toInvoke = null;
					synchronized( ((CallbackReporter) reporter).dbQueue ) {
						toInvoke = ((CallbackReporter) reporter).dbQueue.remove(time);
						if (DEBUG)
							LOGGER.info("REPORTE: map size: " + ((CallbackReporter) reporter).dbQueue.size() + " " + toInvoke);
					}
					
					if ( toInvoke != null ){

						//if (DEBUG)
						//	LOGGER.info("REPORTER: byteCount: " + byteCount);

						String methodName = (String) toInvoke.get(0);
						List<Object> params = toInvoke.subList(1, toInvoke.size());
						Object success = invokeMethod(methodName, params);
						
						if ( success instanceof Boolean) {
							
							if ( !((Boolean)success) ) {
							
								if (DEBUG) 
									LOGGER.info("###################REPORTE: Submission failed! put back in queue and try again later!!");
											
								putBack.put(time,  toInvoke);
								
							}
							else {

								//data submitted successfully
								byteCount += sizeof(toInvoke);							
								
							}
						}
					}
						
				}	

				if ( putBack.size() != 0) {
										
					synchronized( ((CallbackReporter) reporter).dbQueue ) {						
						((CallbackReporter) reporter).dbQueue.putAll(putBack);									
					}
					
					if (DEBUG) {
						LOGGER.info("REPORTER: some submissions failed, putting back " + 
								putBack.size() + " reports in the queue...");
					}
				}

			}
			
			if ( Main.DEBUG) {
				printDebug();
				flush();
			}
						
			if ( ((CallbackReporter) reporter).dbQueue.size() == 0 && Statistics.reportableFacts.size() == 0 && 
					(Statistics.getInstance().peerMap.size() > 500 || Statistics.getInstance().routerMap.size() > 500) )  {
				
				Statistics.getInstance().peerMap.clear();
				Statistics.getInstance().routerMap.clear();
				Statistics.getInstance().dnsMap.clear();
				Statistics.getInstance().peerClientMap.clear();
				Statistics.getInstance().urlMap.clear();
				Statistics.getInstance().peerIds.clear();
				Statistics.getInstance().routerIds.clear();
				Statistics.getInstance().urlIds.clear();
				Statistics.getInstance().peerClientIds.clear();
				Statistics.getInstance().dnsIds.clear();
								
				Statistics.getInstance().addPeerForLookup(Main.getPublicIpAddress());				
				
			}
				
			
			//perform static flush
			int staticFlushInterval = DasuConfiguration.getInstance().getDbStaticFlushInterval();
			
			if ( Util.currentGMTTime()-lastStaticFlushTime >= staticFlushInterval * 1000){
				
				if (DEBUG)
					LOGGER.info("REPORTER: DBManager.class: Static flush! Next one in " + staticFlushInterval + " seconds...");										
				
				flush();
				lastStaticFlushTime = Util.currentGMTTime();
				
			}
			
			synchronized (halt) {

				try {
					if (!terminate) {

						int sleepInterval = DasuConfiguration.getInstance().getDBManagerSleepInterval();
						if (DEBUG) {
							sleepInterval = 5;
							LOGGER.info("REPORTER: DBManager.class: sleeping for " + sleepInterval + " seconds...");
							halt.wait(sleepInterval * 1000);
							printDebug();
						}
						else {
							halt.wait(sleepInterval * 1000);
						}
						
					}
				} catch (InterruptedException e) {
					e.printStackTrace();
				}				
			}
		}
		

		if (DEBUG)
			LOGGER.info("REPORTER: DBManager.class: Hasta Luego...");
	}
	
	public void printDebug() {
					
		StringBuilder sb = new StringBuilder();
		sb.append("\nREPORTER:================================\n");
		sb.append("REPORTER: currentSessionTime: " + new Date(Statistics.getInstance().sessionRestartMonitor.getCurrentSessionStartTime()) + "\n");
		sb.append("REPORTER: loadedSessionTime: " + new Date(Statistics.getInstance().sessionRestartMonitor.getLoadedSessionStartTime()) + "\n");
		sb.append("REPORTER:================================\n");
		
		sb.append("REPORTER: Statistics reportableFacts size: " + Statistics.reportableFacts.size() + "\n");

		sb.append("REPORTER: CallbackReporter dbQueue size: " + 
			((CallbackReporter)CallbackReporter.getInstance()).dbQueue.size() + "\n");

		for ( String s : ((WranglerReporter)WranglerReporter.getInstance()).reports.keySet()) {
			sb.append("REPORTER: WranglerReporter report " + s + " size: " +
			((WranglerReporter)WranglerReporter.getInstance()).reports.get(s).getJSONList().size() + "\n");
		}
		
		sb.append("REPORTER: PhpReporter table: peerIds  size: " + Statistics.getInstance().peerIds.size() + "\n");
		sb.append("REPORTER: PhpReporter table: peerMap  size: " + Statistics.getInstance().peerMap.size() + "\n");
		sb.append("REPORTER: PhpReporter table: dnsIds  size: " + Statistics.getInstance().dnsIds.size() + "\n");
		sb.append("REPORTER: PhpReporter table: dnsMap  size: " + Statistics.getInstance().dnsMap.size() + "\n");
		sb.append("REPORTER: PhpReporter table: routerIds  size: " + Statistics.getInstance().routerIds.size() + "\n");
		sb.append("REPORTER: PhpReporter table: routerMap  size: " + Statistics.getInstance().routerMap.size() + "\n");

		for ( String s : Statistics.getInstance().baos ) {
			
			try {
				sb.append("REPORTER: PhpReporter report " + s + " size: " +
					((ByteArrayOutputStream)
					Statistics.getInstance().getClass().getField(s).get(Statistics.getInstance())).size() + "\n");									
			
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		sb.append("REPORTER:================================\n");						
		
		LOGGER.info(sb.toString());
	
	}
	
	public int batchProcessTable(String tableName, String columnName,
			Hashtable<String, Integer> localMap, Hashtable<Integer, Integer> dbMap) {

		int numRemaining = 0;
		synchronized (localMap) {
			LinkedList<String> keys = new LinkedList<String>();
			for (Entry<String, Integer> ent : localMap.entrySet()) {
				if (!dbMap.containsKey(ent.getValue())) {
					keys.add(ent.getKey());
				}
			}

			if (keys.size() == 0)
				return 0;
			
			if (DEBUG)
				LOGGER.info("REPORTE: " + keys);
			
			Map<String, Integer> dbKeys = PhpReporter.getInstance().makeIdBatch(tableName, columnName, keys);

			if ( dbKeys != null ) {
				
				int numFound = 0;
				for (Entry<String, Integer> ent : dbKeys.entrySet()) {
					dbMap.put(localMap.get(ent.getKey()), ent.getValue());
					numFound++;
				}
			
				//in case the db returned 0 mappings, set mappings to -1 to avoid infinite loop
				//huh?, what loop?
				//if (numFound == 0) {
				//	for (String ent : keys) {
				//		dbMap.put(localMap.get(ent), -1);
				//		numFound++;
				//	}				
				//}
				numRemaining = keys.size() - numFound;
			}
		}
		return numRemaining;
	}
	
	public void batchProcessIds() {
		
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase()) {			
			return;
		}

		int totalLeft;
		int numLoops = 0; // to ensure termination, try 3 times
		do {
			
			totalLeft = 0;
			totalLeft += batchProcessTable("peerips", "ip", PhpReporter.peerMap, PhpReporter.peerIds);
			totalLeft += batchProcessTable("peerclients", "name", PhpReporter.peerClientMap, PhpReporter.peerClientIds);
			totalLeft += batchProcessTable("routerips", "ip", PhpReporter.routerMap, PhpReporter.routerIds);
			totalLeft += batchProcessTable("urlids", "url", PhpReporter.urlMap, PhpReporter.urlIds);
			totalLeft += batchProcessTable("dnsips", "ip", PhpReporter.dnsMap, PhpReporter.dnsIds);
			numLoops++;
			
		}while (totalLeft > 0 && PhpReporter.getInstance().canConnect()
				&& numLoops < 3);

	}

	public void stop() {

		//do we want to awake the manager one last time to send flushed data		
		//before closing up?
		isActive = false;
		synchronized (halt) {
			terminate = true;
			halt.notifyAll();
		}

		stopReporting();
		
	}
	
    private static final int SZ_REF = 4;
    
    private static int size_inst(Class<?> c) {
            Field flds[] = c.getDeclaredFields();
            int sz = 0;

            for (int i = 0; i < flds.length; i++) {
                    Field f = flds[i];
                    if (!c.isInterface() &&
                        (f.getModifiers() & Modifier.STATIC) != 0)
                            continue;
                    sz += size_prim(f.getType());
            }

            if (c.getSuperclass() != null)
                    sz += size_inst(c.getSuperclass());

            Class<?> cv[] = c.getInterfaces();
            for (int i = 0; i < cv.length; i++)
                    sz += size_inst(cv[i]);

            return sz;
    }

    private static int size_prim(Class<?> t) { 
            if (t == Boolean.TYPE)
                    return 1;
            else if (t == Byte.TYPE)
                    return 1;
            else if (t == Character.TYPE)
                    return 2;
            else if (t == Short.TYPE)
                    return 2;
            else if (t == Integer.TYPE)
                    return 4;
            else if (t == Long.TYPE)
                    return 8;
            else if (t == Float.TYPE)
                    return 4;
            else if (t == Double.TYPE)
                    return 8;
            else if (t == Void.TYPE)
                    return 0;
            else
                    return SZ_REF;
    }

    private static int size_arr(Object obj, Class<?> c) {
            Class<?> ct = c.getComponentType();
            int len = Array.getLength(obj);

            if (ct.isPrimitive()) {
                    return len * size_prim(ct);
            }
            else {
                    int sz = 0;
                    for (int i = 0; i < len; i++) {
                            sz += SZ_REF;
                            Object obj2 = Array.get(obj, i);
                            if (obj2 == null)
                                    continue;
                            Class<?> c2 = obj2.getClass();
                            if (!c2.isArray())
                                    continue;
                            sz += size_arr(obj2, c2);
                    }
                    return sz;
            }
    }

    public static int sizeof(Object obj) {
            if (obj == null)
                    return 0;

            Class<?> c = obj.getClass();

            if (c.isArray())
            	return size_arr(obj, c);
            else
            	return size_inst(c);
    }
    
    public boolean canConnect() {
    	
    	if ( !DasuConfiguration.getInstance().isRecordStats() )
    		return false;
    	
    	if ( DasuConfiguration.getInstance().isEnablePhPReporter() )  {
    		return PhpReporter.getInstance().canConnect() &&
    				WranglerReporter.getInstance().canConnect();
    	}
    	
    	return false;
    	
    }
    
    public int makeUUIDid(String uuid) {
    	
    	if ( DasuConfiguration.getInstance().isSendToDatabase() ) {
    		
			ReportableFact rf = new ReportableFact("makeUUID");
			rf.getContentsMap().put("timestamp", Util.currentDasuTime());
			rf.getContentsMap().put("uuid", uuid);
			Statistics.getInstance().addReportableFact(rf);    		
    		
    		if ( DasuConfiguration.getInstance().isEnablePhPReporter() )
    			return PhpReporter.getInstance().makeId("uuids", "uuid", uuid);
    	}

    	return 0;
    }

    public void flush() {
    	
    	if ( DasuConfiguration.getInstance().isEnablePhPReporter() )
    		PhpReporter.getInstance().flush();
    	
    	WranglerReporter.getInstance().flush();
    	
    }
	
    public void stopReporting() {

    	if ( DasuConfiguration.getInstance().isEnablePhPReporter() )
    		PhpReporter.getInstance().stopReporting();    	
    	WranglerReporter.getInstance().stopReporting();
    	
    }
}
