/**
 * Dasu Project
 *
 * File:         NullReporter.java
 * RCS:          $Id$
 * Description:  NullReporter class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Jul 8, 2007 at 10:00:02 AM
 * Language:     Java
 * Package:      edu.northwestern.news.database
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2007, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.database;

import java.nio.charset.Charset;
import java.security.KeyStore;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;

import edu.northwestern.cs.aqualab.Cattle.Report;
import edu.northwestern.cs.aqualab.Cattle.Report.ReportResult;
import edu.northwestern.cs.aqualab.Cattle.json.JSONObject;
import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.measurement.TraceEntry;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchezs@eeccs.northwestern.edu&gt;
 *
 * The WranglerRporter class reports data to a web service based on Wrangler.
 *  
 */
public class WranglerReporter implements DatabaseReporter {

	private final static Logger LOGGER = DasuManager.LOGGER;

	private static WranglerReporter instance = new WranglerReporter(); 
	protected Hashtable<String, Report> reports;
	private static boolean DEBUG = true;
	
	private static final String DASU_REPORT_SERVER_HOST = DasuConfiguration.DASU_REPORT_SERVER_HOST;	
	private static final String DASU_REPORT_RESOURCES = DasuConfiguration.DASU_REPORT_RESOURCE;
	public static final Integer DASU_REPORT_SERVER_PORT = DasuConfiguration.DASU_REPORT_SERVER_PORT;
	
    private static final boolean USE_SSL = true;
    private Long lastFailTime = -1L;
    
	public synchronized static DatabaseReporter getInstance() {

		if (instance == null
				|| (!(instance instanceof WranglerReporter) && !Main
						.isShuttingDown())) {
			instance = new WranglerReporter();
		}

		return instance;
	}

	
	public WranglerReporter() {

        KeyStore ks = DasuConfiguration.getInstance().getKeyStore();
        Report.addKeyStore(ks);        
		reports = DasuManager.getPersistentDataManager().getReports();
		
	}
	
	public boolean pushGenericReport(FactTemplate fact) {	
		
		try {
			
			JSONObject report = fact.toJSON();			
			String clazz = fact.getClass().getSimpleName();			
			
			if (DEBUG)
				LOGGER.info("REPORTER: wrangler pushReport(): " + clazz + " " + report + "\n\n");
					
			synchronized(reports) {		
	
				if ( !reports.containsKey(clazz) ) 
					reports.put(clazz, new Report());				
				
				return reports.get(clazz).getJSONList().add(report);
			}
		}
		catch (Exception e) {
			LOGGER.info("REPORTER: ERROR: " + e.getMessage() + " " + e.getMessage() + " " + e.getLocalizedMessage());
		}
		
		return false;
		
	}
	
	public void flush() {	
		
		LOGGER.info("REPORTER: flushing");
		if ( !DasuConfiguration.getInstance().isRecordStats() ||
				!canConnect() ) {
			LOGGER.info("REPORTER: can't connect!! wait...");
			return;
		}
		
		synchronized(reports) {
			
			Iterator<Entry<String, Report>> iter = reports.entrySet().iterator();
			while ( iter.hasNext() ) {
				Entry<String, Report> entry = iter.next();
				String resource = entry.getKey();
				Report report = entry.getValue();
				
		        report.setDestinationHost(DASU_REPORT_SERVER_HOST);
		        report.setDestinationPort(DASU_REPORT_SERVER_PORT);
		        report.setResource(DASU_REPORT_RESOURCES + "/" + resource);
		        if ( DEBUG )
		        	LOGGER.info("REPORTER: " + report);
		        
		        ReportResult result = report.send(USE_SSL);
		        if ( DEBUG )
		        	LOGGER.info("REPORTER: result(): " + result);
		        if (result.success == false) {
		        	
		        	if ( result.HTTP_code == 200) {
		        		
		        		try {
		        			if (result.response.get("error").toString().equals("malformed request")) {
		        				LOGGER.info("REPORTER: malformed request, drop!!");
		        			}
		        			else {
		        				LOGGER.info("REPORTER: ERROR: Report failed: " + result.response.get("error").toString() );
					        	updateLastFailTime();
					            return;
		        			}
	        			}
		        		catch (Exception e) {
		        			if ( DEBUG )
		        				e.printStackTrace();
		        		}
		        		
		        	}
		        	else if ( result.HTTP_code != 404 ) { 
		        		LOGGER.info("REPORTER: ERROR: Report failed: " + result.response );
			        	updateLastFailTime();
			            return;			        	
		        	}
		        }
		        
		        iter.remove();		        
		        setLastFailTime(-1L); // reset failure time since we are successful
			}
			
		}
	}
	
	//send data right away instead of storing it locally 
	private boolean sendData(FactTemplate fact) {
		
		JSONObject jsonReport = fact.toJSON();			
		String resource = fact.getClass().getSimpleName();
		Report report = new Report();
		report.getJSONList().add(jsonReport);
		
        report.setDestinationHost(DASU_REPORT_SERVER_HOST);
        report.setDestinationPort(DASU_REPORT_SERVER_PORT);
        report.setResource(DASU_REPORT_RESOURCES + "/" + resource);
        if ( DEBUG )
        	LOGGER.info("REPORTER: " + report);
        
        ReportResult result = report.send(USE_SSL);
        if ( DEBUG )
        	LOGGER.info("REPORTER: result(): " + result);
        if (result.success == false) {
        			        
        	if ( result.HTTP_code != 404 ) { 
	        	updateLastFailTime();
	        	return false;
        	}
        }
    	        
        setLastFailTime(-1L); // reset failure time since we are successful
        return true;

	}
	
	public boolean canConnect() { 

		return lastFailTime < 0
				|| (Util.currentTimeMillis() - lastFailTime) > DasuConfiguration
						.getInstance().getDbReconnectInterval() * 1000;

	}
	
	/**
	 * Update lastFailTime in a way that prevents a thundering herd by using randomness
	 */
	public synchronized void updateLastFailTime() {
		
		long lastFailTime = Util.currentTimeMillis() + 
			(long)(DasuConfiguration.getInstance().getDbReconnectInterval()*
					1000*(0.5-Main.getRandom().nextDouble()));
	
		setLastFailTime(lastFailTime);
	}

	/**
	 * Update lastFailTime in a way that prevents a thundering herd by using randomness
	 */
	public synchronized void setLastFailTime(long val) {
		lastFailTime = val;
	}
	
	/********************************************************
	 *  Legacy support, all these methods are deprecated 
	 ********************************************************/

	public void setExtendedConnection(boolean extend) { }

	public boolean connect() { return false; }

	public int makeId(String tableName, String colName, String value) { return 0; }

	public void disconnect() { }

	public boolean insertPingResultDasu(int destId, int sourceId, float rtt,
			long time, int destType, int taskId) { return false; }

	public boolean insertHttpResultDasu(int sourceId, int url,
			double downloadDuration, double downloadSpeed, long time,
			int noBytes, int taskId) { return false; }

	public boolean insertSpeedTestResult(int sourceId, long testTime,
			int downloadSpeed, int uploadSpeed, int encryption, char type,
			int downloadSpeedNetstat, int uploadSpeedNetstat) { return false; }

	public boolean insertUpnpTestResult(int sourceId, long testTime, int port,
			String protocol, int success) {	return false; }

	public boolean insertNdtResultDasu(int myId, String osName, String osArch,
			String osVer, String interfaceName, String javaVer, String server,
			double rate, double rate2, double mylink, int mismatch,
			double avgrtt, int packetSize, int connectionStalledTimes,
			double connectionIdlePercentage, int packetQueuing,
			int packetQueuing2,
			double throughputPacketQueueingDetectedPercentage,
			double connectionReceiverLimitedPercentage,
			double connectionSenderLimitedPercentage,
			double connectionNetworkLimitedPercentage, int RFC2018, int RFC896,
			int RFC3168, int RFC1323TimeStamping, int RFC1323WindowScaling,
			int serverScalingFactor, int clientScalingFactor,
			int serverBehindFirewall, int clientBehindFirewall,
			double theoreticalNetworkLimit, double serverBuffer,
			double serverThroughputLimit, double clientBuffer,
			double clientThroughputLimit, long time, int failed,
			String errorMsg, double c2sRateNetstat, double s2cRateNetstat,
			double networkFlowControlThroughputLimit, int congestion,
			int congestionSignals, int fastRetran, int timeouts,
			double outOfOrderPercentage, int pktsRetrans, int dupAcksIn,
			int sACKsRcvd, double packetLoss) {	return false; }

	public boolean insertRSTStat(int sourceId, int destId, long downloadId,
			int numRst, long timeRecorded, int type, int port,
			String bitTorrentClient, String source, int isSeed, int amSeed,
			int numSeeds, int numPeers, String encryptionLevel) { return false; }

	public boolean insertRemotePeerConnectionStat(int sourceId, int destId,
			long timeRecorded, int type, int port, String bitTorrentClient,
			int isSeed, int amSeed, int numSeeds, int numPeers,
			String encryptionLevel) { return false; }

	public long insertTraceEntry(int sourceId, int destId, long time,
			int entries, int taskId) { return 0; }

	public boolean insertTraceEntryData(long entryId, int[] myRouterIds,
			TraceEntry te) { return false; }


	public long insertScrapeEntry(String type, int peerId, String url,
			String hash, int entries, int numSeeds, int numLeechers,
			int taskId, long time) { return 0; }

	public boolean insertScrapeEntryData(long entryId, int peerId, int tcp,
			int udp) { return false; }

	public boolean insertDnsResult(int myId, int dnsId, long requestTime,
			long responseTime, String query, String type, String Options,
			String response, boolean failed, long taskId) {	return false; }

	public boolean insertLocalPeerStatDynamic(int selfId, long currentTime,
			int uploadRate, int downloadRate, int maxDownloadRate,
			int maxUploadRate, boolean isAutoSpeed, int rst,
			int numConnections, int passiveOpens, int activeOpens,
			int failedConns, int numTorrents, int numSeedingTorrents,
			int numLeechingTorents, int numConnectedSeeds,
			int numConnectedLeechers, long bytesSentLpsd,
			long bytesReceivedLpsd, long bytesSentNetstat,
			long bytesReceivedNetstat, long swarmPeerSpeed,
			long numberOfActiveConnections, long totalBytesSent,
			long totalBytesReceived, long totalPacketsSent,
			long totalPacketsReceived, long uptime) { return false; }

	public boolean insertLocalPeerStatStatic(int selfId, long uptime,
			int maxUploadRate, int maxDownloadRate,
			int maxPeerConnectionsGlobal, int maxPeerConnectionsPerTorrent,
			int maxPeerConnectionsPerTorrentSeeding, int maxUploadRateSeeding,
			int maxSeedsPerTorrent, int lazyBitfield, int requireEncryption,
			String encryptionMinLevel, int allowNonEncryptedOUT,
			int allowNonEncryptedIn, int useCryptoPortExtension, int tcpPortIn,
			int udpPortIn) { return false; }

	public boolean insertRemotePeerStatStatic(int selfId, int peerId,
			int peerClient, long uptime, int tcpPort, int udpPort,
			boolean sameCluster, int type, String encryption, long pieceSize) {	return false; }

	public boolean insertRemotePeerStatDynamic(int selfId, int peerId,
			long currentTime, int bytesReceivedFromPeer, int bytesSentToPeer,
			int currentReceiveRate, int currentSendRate, boolean isSeed,
			double ping, boolean isInterested, boolean isInteresting) {	return false; }

	public boolean insertDownloadStatSummary(int selfId, int downloadId,
			long currentTime, long downloadingTime, long seedingTime,
			int maxDownloadRateLimit, int maxUploadLimit, long totalUploaded,
			long totalDownloaded, long size, long pieceSize) { return false; }

	public boolean insertDownloadStatDynamic(int selfId, int downloadId,
			long timeRecorded, int numPeers, float availability, int completed,
			int discarded, int downloadRate, int uploadRate, int shareRatio,
			int numKnownNonSeeds, int numKnownSeeds, int state, int priority,
			long downloaded, long uploaded, int position,
			int totalNumKnownNonSeeds, int totalNumKnownSeeds, long hashFails,
			int connectedSeeds, int connectedLeechers, long swarmPeerSpeed) { return false;	}

	public boolean registerUsage(int myId, long GMTTime, String uuid,
			long currentTime, String timeZone, boolean failed) { return false; }

	public void stopReporting() { }

	public Map<String, Integer> makeIdBatch(String tableName,
			String columnName, List<String> values) { return null; }

	public void updateConfig() { }

	public void reportIpChange(int oldIp, int newIp, int localIp,
			long currentGMTTime) { }

	public String getIspInfo(String myIp) {	return null; }

	public boolean insertUserReports(int sourceId, String downloadIds,
			String comment) { return false; }

	public boolean insertIspInfo(int sourceId, String prefix, String ispname,
			long timeRecorded) { return false; }

	public boolean insertInterference(int sourceId, String type, long timestamp) { return false; }

	public boolean insertStaticUPnPInfo(int myId, String UpstreamMaxBitRate,
			String DownstreamMaxBitRate, String NATEnabled,
			String Layer1UpstreamMaxBitRate, String Layer1DownstreamMaxBitRate,
			String PortMappingEnabled, String WANAccessType,
			String MaximumActiveConnections, String ExternalIPAddress) { return false; }

	public boolean insertInterferenceEvent(int sourceId, String type,
			long timestamp, long downloadId, long droppedTo, float mean) { return false; }

	public boolean insertTask(int sourceId, long timeStamp, int taskId) { return false;	}

	public boolean insertDnsMapping(int myIp, int publicDns, int privateDns,
			String type, int order, long timeRecorded) { return false; }

	public boolean insertDnsTimeout(int myDnsId, int myId, long startTime,
			long endTime, int giveUp) {	return false; }

	public boolean getCanSendData() { return false;	}

	public String getDefaultServer() { return null;	}

	public String getCurrentUrl() {	return null; }

	public Charset getDefaultCharSet() { return null; }

	public boolean isUseGzip() { return false; }

	public boolean insertConfigEventDasu(int sourceId, String event,
			long timestamp) { return false; }

	public boolean insertFeedback(String type, String name, String email,
			String message, String dump, long systime, long dasuGmtTime) { return false; }
		

}
