/**
 * Dasu Project
 *
 * File:         RuleBasedEngine.java
 * RCS:          $Id$
 * Description:  RuleBasedEngine class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:23:51 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.drools
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */


package edu.northwestern.dasu.drools;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Logger;

import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderConfiguration;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderErrors;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.event.rule.ActivationCancelledEvent;
import org.drools.event.rule.ActivationCreatedEvent;
import org.drools.event.rule.AfterActivationFiredEvent;
import org.drools.event.rule.AgendaEventListener;
import org.drools.event.rule.AgendaGroupPoppedEvent;
import org.drools.event.rule.AgendaGroupPushedEvent;
import org.drools.event.rule.BeforeActivationFiredEvent;
import org.drools.io.ResourceFactory;
import org.drools.logger.KnowledgeRuntimeLogger;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.rule.FactHandle;

import edu.northwestern.dasu.DasuData;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.profilers.Profiler;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The RuleBasedEngine class ...
 */
public abstract class RuleBasedEngine extends Thread implements RuleBasedEngineInterface {

	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	public static boolean DEBUG = false;
	
	private SourceType engineId = null;
	
	private boolean active = false;
	private boolean initialized = false;
	protected String filename = null;
	protected String sigFilename = null;
	protected int taskId = -1;
	protected volatile int yieldInterval = -1; //milliseconds
	
	protected KnowledgeBase kbase = null;
	protected StatefulKnowledgeSession ksession = null; 
	protected KnowledgeRuntimeLogger logger = null;
	protected ClassLoader classLoader = DasuManager.getDasuClassLoader(); //this.getClass().getClassLoader(); //DasuManager.getDasuClassLoader();
	
	//protected ClassLoader classLoader = new URLClassLoader(new URL[]{
	//		new URL("file:" + DasuManager.updater.getJarPath(DasuConfiguration.getInstance().getInferenceVersion()))});

	private Hashtable<Long, FactHandle> facts = new Hashtable<Long, FactHandle>();
	private Set<FactHandle> modifiedFacts = 
		Collections.synchronizedSet(new HashSet<FactHandle>());
	
	private LinkedHashMap<String, ArrayList<Object>> firedRules = new LinkedHashMap<String, ArrayList<Object>>();
	
	private boolean USE_SIMULATION_V2 = false;
	
	protected RuleBasedEngine(String name, SourceType engineId) throws Exception {
		super(name);
		this.engineId = engineId; 
		if ( System.getSecurityManager() != null && System.getSecurityManager().toString().contains("AzureusSecurityManager") )
			this.USE_SIMULATION_V2 = DasuManager.USE_SIMULATION_V2;
	}
	
	protected RuleBasedEngine(String name, String filename, String sigFilename, int taskId) throws Exception {
		super(name);
		this.filename = filename;
		this.sigFilename = sigFilename;
		this.taskId = taskId;
	}
	
	public void profile(Profiler p) {
		p.addEngine(ksession);
	}
	
	public SourceType getEngineId() {
		return engineId;
	}
	
	protected void setYieldInterval(int timeInMillis) {
		yieldInterval = timeInMillis;
	}
	
	protected void clearYieldInterval() {
		yieldInterval = -1;
	}
    // Add the object and its contents to working memory
	public void addFact(final FactTemplate factObject, SourceType engineId) {
		
		if ( USE_SIMULATION_V2) {
			//i'm inside Vuze
			if ( System.getSecurityManager() != null && System.getSecurityManager().toString().contains("AzureusSecurityManager") )
				DasuManager.getInstance().getSimulator().addToInputPlot(factObject);			
		}
		
		if (isActive()) {
		
			if (!facts.containsKey(factObject.getId())) {
				
				FactHandle fh = null;
				fh = ksession.insert(factObject);
				facts.put(factObject.getId(), fh);
			
				factObject.addPropertyChangeListener(
						new PropertyChangeListener() {
							public void propertyChange(PropertyChangeEvent e) {
						        updateFact(factObject);
							}
						});
			}

			if (DEBUG)
				displayFacts();
		}
	}
	
	public void addFact(final FactTemplate factObject) {		
		addFact(factObject, this.getEngineId());
	}
		
	/*
	 * XXX: modifying fact can quietly fail if it is being modified...
	 */
	private void updateFact(FactTemplate factObject) {
					
		synchronized(factObject) {
					
			if (isActive()) {
				FactHandle fh = ksession.getFactHandle(factObject);
				
				if (fh != null) {
					
					/* object is being modified in working memory,
					 * avoid concurrent modification by doing nothing 
					 */
	
					if (modifiedFacts.contains(fh)) {
						if ( DEBUG )
							LOGGER.info("RULEENGINE: updateFact(): Fact being modified, skip: " + factObject);
						return;
					}
					
					ksession.update(fh, factObject);
					
				}
				else
					facts.remove(fh);
			}

		}
				
	}
	
    // Remove the object and its contents from working memory
	public void retractFact(FactTemplate factObject) {
		if (isActive()) {
			FactHandle fh = ksession.getFactHandle(factObject);
			if (fh != null)
				ksession.retract(fh);
			else
				facts.remove(factObject.getId());
		}
			
	}


	public void initialize() throws Exception {
		
		if (loadRules()) {
			loadFacts();
			setInitialized(true);
		}
		else {
			throw new Exception("No rules to load... aborting engine startup...");
		}
			
		
	}

	protected boolean loadRules() throws Exception {
		return false;
	}
	
	public void enableSimulationListener() {
	
		ksession.addEventListener(new AgendaEventListener(){

			public void activationCancelled(ActivationCancelledEvent arg0) {}
			public void activationCreated(ActivationCreatedEvent arg0) {}
			public void agendaGroupPopped(AgendaGroupPoppedEvent arg0) {}
			public void agendaGroupPushed(AgendaGroupPushedEvent arg0) {}
			public void beforeActivationFired(BeforeActivationFiredEvent arg0) {}
			
			public void afterActivationFired(AfterActivationFiredEvent arg0) {				

				if ( arg0.getActivation() != null) {
					
					String name = arg0.getActivation().getRule().getName();
					Long timeRecorded = Util.currentTimeMillis();					

					ArrayList<Object> lst = new ArrayList<Object>();					
					lst.add(timeRecorded);
					lst.add(name);
					lst.add(arg0.getActivation().getDeclarationIDs());
					for (FactHandle fh : arg0.getActivation().getFactHandles()) {
						lst.add(fh.toString());					
					}
					
					//second try
					if (firedRules.containsKey(name))
						firedRules.remove(name);				
					firedRules.put(name, lst);
					
					ArrayList<ArrayList<Object>> lists = new ArrayList();
					lists.addAll(firedRules.values());
					DasuData.getInstance().update("simulator.fired.rules2", lists);
					
				}
				
			}

		});
	}

	@Override
	public void run() {
		
		if (isInitialized()) {
			setActive(true);		
			if (DEBUG)
				LOGGER.info(this.getName() + ": Starting execution Drools engine!");				
		}
				
       	if ( USE_SIMULATION_V2)
       		enableSimulationListener();
       	
		try {
			
			//while (!Main.isShuttingDown() && isActive()) {
			while ( isActive() ) {

				if ( System.getSecurityManager() != null && System.getSecurityManager().toString().contains("AzureusSecurityManager")) {
					if ( Main.isShuttingDown() )
						break;
				}

				ksession.fireAllRules();
				if ( ksession.getGlobal("halt") != null && ((FactHalt)ksession.getGlobal("halt")).isHalt())
					break;
				
				if ( USE_SIMULATION_V2)
					Thread.sleep(1);
				else
					Thread.sleep(100);
				
				if ( yieldInterval != -1) {
					try {
						sleep(yieldInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				
			} // end while active
		}catch (InterruptedException e) {			
			//e.printStackTrace();
			LOGGER.info("RULEENGINE: exception " + e.getMessage());
		}
		
		ksession.halt();
		LOGGER.info("RULEENGINE: " + this.getName() + ": Finishing execution of Drools engine\n\n");
		setActive(false);
		
		destroySession();

	}
	
	protected void destroySession() {

		if (ksession != null)
			ksession.dispose();
	}
	
	protected KnowledgeBase loadRules(Vector<ByteArrayOutputStream> data) throws Exception {
		Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, 
        		new ClassLoader[]{classLoader});
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
		
		for(ByteArrayOutputStream baos: data) {
			kbuilder.add(ResourceFactory.newByteArrayResource(baos.toByteArray()), ResourceType.DRL);		
			KnowledgeBuilderErrors errors = kbuilder.getErrors();
			if (errors.size() > 0) {
				for (KnowledgeBuilderError error: errors) {
					LOGGER.severe("RULEENGINE: " + error.toString());
				}
				throw new IllegalArgumentException("Could not parse knowledge.");
			}
		}
		
		//KnowledgeBaseConfiguration baseconfig = KnowledgeBaseFactory.newKnowledgeBaseConfiguration(null, new MyClassLoader(new URL[] { new URL("http://mas939-imac.cs.northwestern.edu/") }));
		//kbase = KnowledgeBaseFactory.newKnowledgeBase(baseconfig);
		kbase = KnowledgeBaseFactory.newKnowledgeBase();
		
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		return kbase;

	}
	
	protected KnowledgeBase loadRulesFromDir(String dir, String firstName) throws MalformedURLException {
		
		Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, 
        		new ClassLoader[]{classLoader});
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
		
		File dirFile = new File(dir);
		
		kbuilder.add(ResourceFactory.newFileResource(new File(dirFile, firstName)), ResourceType.DRL);		
		KnowledgeBuilderErrors errors = kbuilder.getErrors();
		if (errors.size() > 0) {
			for (KnowledgeBuilderError error: errors) {
				LOGGER.severe(error.toString());
			}
			throw new IllegalArgumentException("Could not parse knowledge.");
		}
		
		for(File f: dirFile.listFiles()) {
			if(!f.getName().equals(firstName) && f.getName().endsWith(".drl")) {
				LOGGER.info("RULEENGINE: loading "  + dir + "/" + f.getName());
				kbuilder.add(ResourceFactory.newFileResource(f), ResourceType.DRL);		
				errors = kbuilder.getErrors();
				if (errors.size() > 0) {
					for (KnowledgeBuilderError error: errors) {
						LOGGER.severe(error.toString());
					}
					throw new IllegalArgumentException("Could not parse knowledge.");
				}
			}
		}
		
		//KnowledgeBaseConfiguration baseconfig = KnowledgeBaseFactory.newKnowledgeBaseConfiguration(null, Main.getPluginInterface().getPluginClassLoader());
		//kbase = KnowledgeBaseFactory.newKnowledgeBase(baseconfig);
		kbase = KnowledgeBaseFactory.newKnowledgeBase();
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		return kbase;

	}
	
	protected KnowledgeBase loadRules(String fileName) {
		Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props,
        		classLoader);
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
		
		kbuilder.add(ResourceFactory.newFileResource(fileName), ResourceType.DRL);		
		KnowledgeBuilderErrors errors = kbuilder.getErrors();
		if (errors.size() > 0) {
			for (KnowledgeBuilderError error: errors) {
				LOGGER.severe(error.toString());
			}
			throw new IllegalArgumentException("Could not parse knowledge.");
		}
		//KnowledgeBaseConfiguration baseconfig = KnowledgeBaseFactory.newKnowledgeBaseConfiguration(null, Main.getPluginInterface().getPluginClassLoader());
		//kbase = KnowledgeBaseFactory.newKnowledgeBase(baseconfig);
		kbase = KnowledgeBaseFactory.newKnowledgeBase();
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		return kbase;

	}
	
	protected void loadFacts() {}

	public void displayFacts() {
				
		if (isActive()) {
			
			StringBuilder sb = new StringBuilder();
			Iterator<FactHandle> f = ksession.getFactHandles().iterator();
			while (f.hasNext()) {
				FactHandle fh = f.next();
				Object obj = ksession.getObject(fh);
				if(obj != null)
					sb.append("RULEENGINE: " + getName() + " => " + obj.toString() + "\n");
			}
			
			LOGGER.info("\n**********\n" + sb.toString() + "***********\n");
		}
		
	}
	
	
	public void stopEngine() {
		
		LOGGER.info("STOP:: stop " + this.engineId);
		
		if (isActive()) {
			ksession.halt();
			setActive(false);
			destroySession();
		}
		
		LOGGER.info("STOP:: stop " + this.engineId);
	}
	
	public boolean isActive() {
		return active;
	}
	
	public void setActive(boolean val) {
		this.active = val;
	}
	
	public void setInitialized(boolean state) {
		initialized = state;
	}
	
	public boolean isInitialized() {
		return initialized;
	}

	public void setModifiedFacts(Set<FactHandle> modifiedFacts) {
		this.modifiedFacts = modifiedFacts;
	}

	public Set<FactHandle> getModifiedFacts() {
		return modifiedFacts;
	}
	
}
