package edu.northwestern.dasu.dynamic.coordination;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.regex.Pattern;

import org.apache.commons.net.util.SubnetUtils;
import org.apache.commons.net.util.SubnetUtils.SubnetInfo;
import org.apache.zookeeper.CreateMode;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.WatchedEvent;
import org.apache.zookeeper.Watcher;
import org.apache.zookeeper.ZooKeeper;
import org.apache.zookeeper.ZooDefs.Ids;
import org.apache.zookeeper.data.Stat;

import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;

public class ZooKeeperConnection implements Watcher {

	@SuppressWarnings("unused")
	private class Lease {
		
		private HashMap<Long, Pair<Integer, Integer>> startToLease = new HashMap<Long, Pair<Integer, Integer>>();
		private HashMap<String, Pair<Integer, Integer>> prefixToLease = new HashMap<String, Pair<Integer, Integer>>();  	//subnet, <timespan, limit>
		private HashMap<Long, Long> startToEnd = new HashMap<Long, Long>();
		private ArrayList<Long> rangeStart = new ArrayList<Long>();		    
		
		public ProbeType probeType;
		public Stat leaseStat;
		
		public Lease(ProbeType type, Stat stat) {
			this.probeType = type;
			this.leaseStat = stat;
		}
		
		public void addSubnet( String prefix, String timespanS, String limitS) {
									
			Integer timespan = Integer.parseInt(timespanS);
			Integer limit = Integer.parseInt(limitS);			
			
			SubnetInfo sni = null;
			if ( prefix.contains("/") )
				sni = new SubnetUtils(prefix).getInfo();
			
			Long start = sni != null ? ipToInt(sni.getLowAddress()) : ipToInt(prefix);
			Long end = sni != null ? ipToInt(sni.getHighAddress()) : ipToInt(prefix);
			startToEnd.put(start, end);
			startToLease.put(start, new Pair<Integer, Integer>(timespan, limit));
			rangeStart.add(start);
			Collections.sort(rangeStart);

			if (!prefixToLease.containsKey(prefix))
				prefixToLease.put( prefix, new Pair<Integer, Integer>(timespan, limit));
			
		}
		
		private long ipToInt(final String addr) {
				 
			 final String[] addressBytes = addr.split("\\.");

			 long ip = 0;
			 for (int i = 0; i < 4; i++) {
				 ip <<= 8;
				 ip |= Integer.parseInt(addressBytes[i]);
			 }
			 return ip;
		 }
		
		
		public Pair<Integer, Integer> getLease(String ip) {
			
			//System.out.println(ip);
			
			if (ip.contains("/")) {
				String lastDigit = ip.substring(ip.lastIndexOf(".")+1, ip.indexOf("/"));
				Integer lastDigitInt = Integer.parseInt(lastDigit);
				lastDigitInt+=1;
				ip = ip.substring(0, ip.lastIndexOf(".")+1);
				ip = ip.concat(lastDigitInt.toString());
			}
			
			//System.out.println(ip);
			
			if (!isValidIp(ip))
				return new Pair<Integer, Integer>(-1, -1);
				
			Long ipVal = ipToInt(ip);			
			int index = binarySearch(rangeStart.toArray(new Long[startToEnd.size()-1]), ipVal);
			
			if (index < 0)
				return new Pair<Integer, Integer>(-1, -1);
			
			long start = rangeStart.get(index);
			if (ipVal >= start && ipVal <= startToEnd.get( start ))
				return startToLease.get( start );
			else
				return new Pair<Integer, Integer>(-1, -1);
			
		}
		
		
		@Override
		public String toString() {
			
			return this.prefixToLease.toString();
		}
		
		private boolean isValidIp(String str) {  
			  Pattern ipPattern = Pattern.compile("\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}");  
			  return ipPattern.matcher(str).matches();  
		}  
		
		private int binarySearch( Comparable [ ] a, Comparable x ) {

			int low = 0;
	        int high = a.length - 1;
	        int mid = 0;

	        while( low <= high )
	        {
	            mid = ( low + high ) / 2;

	            if( a[ mid ].compareTo( x ) < 0 )
	                low = mid + 1;
	            else if( a[ mid ].compareTo( x ) > 0 )
	                high = mid - 1;
	            else
	                return mid;
	        }

		    return high;
		}
	}
	
	private ZooKeeper zk;
	private String server;
	private String serverPort;
	public HashMap<ProbeType, Lease> leaseMap = new HashMap<ProbeType, Lease>();
	
	public ZooKeeperConnection(String hostPort){
		
		try {
			zk = new ZooKeeper(hostPort, 15000, this);
			Thread.sleep(1000);
			
			server = hostPort.split(":")[0];
			serverPort = hostPort.split(":")[0];
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		catch (InterruptedException e) {
			e.printStackTrace();
		}
		
//		leaseMap.put(ProbeType.PING, new Lease(ProbeType.PING, null));
//		leaseMap.get(ProbeType.PING).addSubnet("207.110.23.0/19", "100", "200");
//		leaseMap.get(ProbeType.PING).addSubnet("165.124.182.0/21", "200", "300");
//		leaseMap.get(ProbeType.PING).addSubnet("207.110.23.0/20", "300", "400");

	}

	public String getServerName() {		
		return server;
	}
	
	public String getServerPort() {		
		return serverPort;
	}
	
	public void process(WatchedEvent event) {
		// TODO Auto-generated method stub
	}
	
	private void updateLease(String line, Stat stat) {
		
		//PING_200.3.81.5:11,12;192.168.2.1:99,15|HTTP_1.2.3.4:9,8		
		try {
			
			for ( String s : line.split("\\|") ) {
				
				String type = s.split("_")[0];
				String ips = s.split("_")[1];
				for ( String ipString : ips.split(";") ) {
	
					String ip = ipString.split(":")[0];
					String timespan = ipString.split(":")[1].split(",")[0];
					String limit = ipString.split(":")[1].split(",")[1];
				
					ProbeType pt = ProbeType.valueOf(type);
					if ( !leaseMap.containsKey(pt) )
						leaseMap.put(pt, new Lease(pt, stat));
					
					leaseMap.get(pt).addSubnet(ip, timespan, limit);
					
				}
				
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
		System.out.println(leaseMap);
		
	}
	
	public String pullUpdatedLease(String TaskID){

		String result = null;
		Stat stat = new Stat();
		byte[] B;
		System.out.println(zk.getSessionId());
		
		try {
			
			B=zk.getData("/zookeeper/"+TaskID+"/Update_Lease", false, stat);			
			result = new String(B);
			updateLease(result, stat);
			
		} catch (KeeperException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		return result;
		//returns a string in format "url:lease version:last updated timestamp"
		
	}
	
	public boolean registerWithZookeeper(String UUID, String TaskID){
		
		byte buffer2[]=UUID.getBytes();
		try {
			Stat temp = zk.exists("/zookeeper/"+TaskID+"/Dasu_Clients/"+UUID, false);
			if (temp!=null)
			{
				System.out.println("node "+UUID+" already exists");
				return false;
			}
			zk.create("/zookeeper/"+TaskID+"/Dasu_Clients/"+UUID, buffer2, Ids.OPEN_ACL_UNSAFE, CreateMode.PERSISTENT);
		} catch (Exception e) {
			e.printStackTrace();
		} 
		
		return true;
		//returns true if the registration is successful false otherwise.
	}
	
	private void createIfNotExists(String path, CreateMode mode) throws KeeperException, InterruptedException {
		
		String leaf = path.split("/")[ path.split("/").length-1 ];
		Stat temp = zk.exists(path, false);
		if (temp == null)//node doesn't exists so create it
			zk.create(path, leaf.getBytes(), Ids.OPEN_ACL_UNSAFE, mode);

	}
	
	private void recursiveCreateIfNotExists(String fullPath) throws KeeperException, InterruptedException {

		String parts[] = fullPath.split("/");
		String path = parts[0];

		for ( int i=1; i < parts.length; i++) {			
			path = path.concat("/"+parts[i]);			
			createIfNotExists(path, CreateMode.PERSISTENT);
		}
		
	}
	
	public void reportProbedIps(String UUID_No, String TaskID, ArrayList<String> ListOfIPs, ProbeType probeType){
		//System.out.println(ListOfIPs);
		int size = ListOfIPs.size();
		
		for (int i=0;i<size;i++){
			String ip = ListOfIPs.get(i);			
			String octects[] = ip.split("\\.");
			String path = "/zookeeper/"+TaskID+"/Probes/"+probeType;
			
			if(octects.length != 4) {
				System.out.println("Invalid IP address"+ListOfIPs.get(i));
			}
			else {
				try {

					String data = new Long(Util.currentGMTTime()).toString();
					
					for ( String octect : octects) {
						path = path.concat("/"+octect);
						createIfNotExists(path, CreateMode.PERSISTENT);
					}
					
					if (path != null) {

						path = path.concat("/" + UUID_No);
						createIfNotExists(path, CreateMode.PERSISTENT);
						path = path.concat("/" + data);
						createIfNotExists(path, CreateMode.PERSISTENT);
						
						byte[] b = zk.getData(path, false, null);
						zk.setData(path, ip.getBytes(), -1);
					
						//System.out.println("\n\n\n" + ip + " " + leaseMap + " " + new Lease(ProbeType.PING, null).ipToInt(ip) + " " + this.leaseMap.get(probeType).getLease(ip) + "\n\n\n\n");
						
					}
					
				} catch (Exception e) {
					e.printStackTrace();
				} 
			}
		}
	}
	

}

 
