/**
 * Dasu Project
 *
 * File:         DasuHttpModule.java
 * RCS:          $Id$
 * Description:  DasuHttpModule class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.measurement;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Map;
import java.util.Map.Entry;

import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.dynamic.drools.FactHttpProbeResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;


/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuHttpModule class manages the set of 
 *         HTT gets performed by the peer.
 */

public class DasuHttpModule extends DasuProbeModule<FactHttpProbeResult> {
		
	public Hashtable<Long, FactHttpProbeResult> httpResultTable;
	private ArrayList<ProbeSubTask> toDoUrlList = new ArrayList<ProbeSubTask>();
	private static DasuHttpModule self;
	private final int chunkSize=1024;

	public DasuHttpModule() {
		super("DasuHttpModule", 1.0f);
		this.pi = Main.getPluginInterface();
		httpResultTable = new Hashtable<Long, FactHttpProbeResult>();
		
		active = true;
	}
	
	public synchronized static DasuHttpModule getInstance() {
		if (self == null) {
			self = new DasuHttpModule();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}

	@Override
	public void run() {
		
	  while (!Main.isShuttingDown() && isActive()) {
		 
		  	ProbeSubTask pst = null;
			synchronized (toDoUrlList) {
				while ((toDoUrlList.size()==0) && !Main.isShuttingDown()) {
					try {
						toDoUrlList.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						setActive(false);
						LOGGER.info("STOP: Adios...");
						return;
					}
				}
				
				if (toDoUrlList.size() > 0) {
					pst = toDoUrlList.remove(0);				
					LOGGER.info("Http server " + pst.ipAddress);
				}
			}
			
			if ( pst != null ) {

				Object ret = probeIp(pst);
				
				if ( ret != null) {
					FactHttpProbeResult result = (FactHttpProbeResult) ret;
					result.setPst(pst);
					updateHttpResultQueue(result);
		
					if ( yieldInterval != -1) {
						try {
							sleep(yieldInterval);
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
				}
			}

	  } // end while active
	  
	  LOGGER.info("STOP: Adios...");
	  
	}
		
	@Override
	public void probeAll(ProbeTask pt) {

		if ( pt != null) {
			synchronized (toDoUrlList) {
				for (ProbeSubTask pst : pt.getSubTasks())
					toDoUrlList.add(pst);
				toDoUrlList.notifyAll();
			}
		}
	}
	
	@Override
	public void probeWithPriority(ProbeSubTask pst) {
		
		if ( pst != null ) {
			synchronized (toDoUrlList) {
				toDoUrlList.add(pst);
				toDoUrlList.notifyAll();
			}
		}
	}

	@Override
	public void probe(ProbeSubTask pst) {
		
		
		LOGGER.info("HTTP: probe() " + pst);
		
		if ( pst != null) {
			synchronized (toDoUrlList) {
				toDoUrlList.add(pst);
				toDoUrlList.notifyAll();
			}
		}

	}
		
	public FactHttpProbeResult fileUrl(String fAddress, String localFileName, Integer taskId, String target, Map<String, String> headers) {

		FactHttpProbeResult ret = new FactHttpProbeResult(fAddress);
		ret.setTarget(target);
		URLConnection  uCon = null;
		Long startTime = 0L;
		Long endTime = 0L;
		
		InputStream is = null;
		try {
			URL Url;
			byte[] buf;
			int ByteRead,ByteWritten=0;
			Url= new URL(fAddress);

			uCon = Url.openConnection();
			
			if (headers != null) {
				
				for ( Entry<String, String> e : headers.entrySet() ) {
					uCon.setRequestProperty(e.getKey(), e.getValue());
				}
				
				uCon.connect();
			}
			
			startTime = Util.currentTimeMillis();
			is = uCon.getInputStream();
			buf = new byte[chunkSize];
			while ((ByteRead = is.read(buf)) != -1) {
				//outStream.write(buf, 0, ByteRead);
				ByteWritten += ByteRead;
			}
			endTime = Util.currentTimeMillis();

			float fileSize = ((float) ByteWritten)/1024F/1024F; /*MB*/
			float dldTime = (endTime - startTime)/1000F;
			float dldRate = (((float) ByteWritten)*8)/dldTime/1024;
			
			ret = new FactHttpProbeResult(fAddress, dldTime, dldRate, ByteWritten, taskId);
			ret.setTarget(target);
			
			reportProbeResult(ret);
			//Statistics.getInstance().addHttpResultDasu((FactHttpProbeResult) fact);

			LOGGER.fine("Downloaded Successfully.");
			LOGGER.fine("File name:\""+localFileName+ "\"\nNo ofbytes: " + ByteWritten);
			LOGGER.fine("File Size: " + ((float) ByteWritten)/1024F/1024F + " MB");
			LOGGER.fine("Dld Time: " + (endTime - startTime)/1000F + " seconds");
			Float dr = (((float) ByteWritten)*8)/((endTime - startTime)/1000);
			LOGGER.fine("Dld Speed: " +  dr/1024 + " kbps");
			
		} catch (Exception e) {
			LOGGER.info("Http Timeout: " + fAddress);
			return ret;
			
		} finally {
			try {
				
				if (is != null)
					is.close();
				//outStream.close();
				
			} catch (IOException e) {
				LOGGER.info("File not found!: " + localFileName);
				return ret;
			}
		}
		return ret;

	}

	/*default probe uses default headers server*/
	
	protected Object probeIp(ProbeSubTask pst) {
	
		if ( !isActive() )
			return null;
		
		Map<String, String> headers = null;
		if(pst.params != null && pst.params.containsKey("headers")) {
			headers = (Map<String, String>) pst.params.get("headers");
		}
		
		FactHttpProbeResult ret = new FactHttpProbeResult(pst.ipAddress);
		ret.setTarget(pst.target);
		int slashIndex =pst.ipAddress.lastIndexOf('/');
		int periodIndex =pst.ipAddress.lastIndexOf('.');
	
		String fileName=pst.ipAddress.substring(slashIndex + 1);
	
		if (periodIndex >=1 &&  slashIndex >= 0 && slashIndex < pst.ipAddress.length()-1) {
			ret = fileUrl(pst.ipAddress, fileName, pst.taskId, pst.target, headers);
		}
		else {
			LOGGER.warning("ERROR: path or file name");
		}

		return ret;
	}

	
	private void updateHttpResultQueue(FactHttpProbeResult result) {
		
		Long id = generateId();
		httpResultTable.put(id, result);
		updateResultQueue(result, result.getProbeType(), id);		
		
	}
	
	@Override
	public void stopModule() {
		
		LOGGER.info("STOP: stopping DasuHttpModule...");
		setActive(false);
		synchronized(toDoUrlList) {
			toDoUrlList.notify();
		}
	
	}
	
	public Pair<String, FactTemplate> proccessResult(long eventId) {
		
		FactHttpProbeResult httpProbeResult = httpResultTable.remove(eventId);		
		return new Pair<String, FactTemplate>(httpProbeResult.url, httpProbeResult);
		
	}
	
}
