/**
 * Dasu Project
 *
 * File:         DasuPingModuleV2.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.measurement;


import java.util.ArrayList;
import java.util.Hashtable;
import java.util.logging.Logger;

import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.dynamic.drools.FactPingProbeResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.PingManager;
import edu.northwestern.dasu.util.Util;
import edu.northwestern.dasu.util.Util.PingResponse;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuPingModuleV2 class manages the set of 
 *         Pings performed by the peer.
 */

public class DasuPingModuleV2 extends DasuProbeModule<FactPingProbeResult> {	
		
	public Hashtable<Long, FactPingProbeResult> pingResultTable;
	private ArrayList<ProbeSubTask> toDoIpList = new ArrayList<ProbeSubTask>();
	private static DasuPingModuleV2 self;

	public DasuPingModuleV2() {
		super("DasuPingModule", 1.0f);
		this.pi = Main.getPluginInterface();
		pingResultTable = new Hashtable<Long, FactPingProbeResult>();
		
		active = true;
	}
	
	public synchronized static DasuPingModuleV2 getInstance() {
		if (self == null) {
			self = new DasuPingModuleV2();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}

	@Override
	public void run() {
		
	  while (!Main.isShuttingDown() && isActive()) {
		 
		  	ProbeSubTask pst = null;
			synchronized (toDoIpList) {
				while ((toDoIpList.size()==0) && !Main.isShuttingDown()) {
					try {
						toDoIpList.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						setActive(false);
						LOGGER.info("STOP: Adios...");
						return;
					}
				}
				
				if ( toDoIpList.size() > 0 ) {
					pst = toDoIpList.remove(0);
					LOGGER.info("Ping server " + pst.ipAddress);
				}
			}
			
			if ( pst != null ) {
								
				if (Util.isValidIp(pst.ipAddress)) {					
					probeIp(pst);
				} 
				else {
					FactPingProbeResult fppr = new FactPingProbeResult(pst.ipAddress, -1D);
					fppr.setTarget(pst.target);
					fppr.setPst(pst);
					
					updatePingResultQueue(fppr);
				}
				
				if ( yieldInterval != -1) {
					try {
						sleep(yieldInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}


	  } // end while active
		
	  LOGGER.info("STOP: Adios...");
	}
		
	@Override
	public void probeAll(ProbeTask pt) {

		if ( pt != null ) {
			synchronized (toDoIpList) {
				for (ProbeSubTask pst: pt.getSubTasks())
					toDoIpList.add(pst);
				toDoIpList.notifyAll();
			}
		}
	}
	
	@Override
	public void probeWithPriority(ProbeSubTask pst) {
		
		if ( pst != null ) {
			synchronized (toDoIpList) {
				toDoIpList.add(pst);
				toDoIpList.notifyAll();
			}
		}
	}

	@Override
	public void probe(ProbeSubTask pst) {

		if ( pst != null ) {
			synchronized (toDoIpList) {
				toDoIpList.add(pst);
				toDoIpList.notifyAll();
			}
		}
	}
	
	@Override
	protected Object probeIp(final ProbeSubTask pst) {

		if ( isActive() ) {		
		
			PingManager.getInstance().doPing(pst.ipAddress, 
					pi, new PingResponse() {
	
						public void response(double rtt) {
							
							FactPingProbeResult fppr = new FactPingProbeResult(pst.ipAddress, rtt);
							fppr.setTarget(pst.target);
							fppr.setPst(pst);
	
							if (rtt >= 0) {
								
								reportProbeResult(fppr);
								//Statistics.getInstance() .addPingResultDasu( getIp(), pst.ipAddress, rtt, Util.currentGMTTime(), 1, pst.taskId);
							}
							
							updatePingResultQueue(fppr);
							Statistics.getInstance().addPeerForLookup(pst.ipAddress);
						}				
					}
			);
		}
		
		return null;

	}

	/*
	private void updatePingResultQueue(FactPingProbeResult fppr) {

		pingResultTable.put(fppr.getId(), fppr);
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
				new Pair<Long, EventType>(fppr.getId(), EventType.HPM_RESULT));
			AMCoordinatorV2.events.notifyAll();
		}
	}*/
	
	private void updatePingResultQueue(FactPingProbeResult fppr) {

		pingResultTable.put(fppr.getId(), fppr);
		updateResultQueue(fppr, fppr.getProbeType(), fppr.getId());
		
	}	
	
	@Override
	public void stopModule() {
		
		LOGGER.info("STOP: stopping DasuPingModule...");
		setActive(false);
		synchronized(toDoIpList) {
			toDoIpList.notify();
		}
	
		self = null;
	}

	public Pair<String, FactTemplate> proccessResult(long eventId) {
		
		FactPingProbeResult pingResult = pingResultTable.remove(eventId);		
		return new Pair<String, FactTemplate>(pingResult.getIp(), pingResult);
		
	}

	
}
