/**
 * Dasu Project
 *
 * File:         DasuTraceRouteModuleV2.java
 * RCS:          $Id$
 * Description:  DasuTraceRouteModuleV2 class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:24:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.dynamic.measurement;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.dynamic.drools.FactTraceResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.TraceEntry;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.HashSetCache;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuTraceRouteModuleV2 class manages the set of 
 *         traceroute measurements performed by the peer.
 */

public class DasuTraceRouteModuleV2 extends DasuProbeModule<FactTraceResult> {	
	
	public Hashtable<Long, FactTraceResult> trResultTable;
	private static DasuTraceRouteModuleV2 self;

    private HashSetCache<ProbeSubTask> pendingRuns;
    private HashSetCache<ProbeSubTask> pendingRunsLowPrio;
    private Object lossSync = new Object();
    private double lossSum = 0;
    private double lossCount = 0;
    //private Integer currentRuns;
    private Hashtable<String, TraceRouteRun> currentRuns;
	
	public DasuTraceRouteModuleV2() {
		super("DasuTraceRouteModule", 1.0f);
		this.pi = Main.getPluginInterface();
		self = this;
		active = true;
    	pendingRuns = new HashSetCache<ProbeSubTask>(200);
    	pendingRunsLowPrio = new HashSetCache<ProbeSubTask>(200);
    	trResultTable = new Hashtable<Long, FactTraceResult>();
        currentRuns = new Hashtable<String, TraceRouteRun>();

	}
	
	public synchronized static DasuTraceRouteModuleV2 getInstance() {
		if (self == null) {
			self = new DasuTraceRouteModuleV2();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}
		
	public void probeWithLowPriority(ProbeSubTask pst) {
		
		if ( pst != null) {
			synchronized(pendingRunsLowPrio){
				pendingRunsLowPrio.add(pst);
			}
		}

	}

	@Override
	public void probe(ProbeSubTask pst) {
		
		if ( pst != null) {
			synchronized(pendingRuns){
				pendingRuns.add(pst);
				pendingRuns.notifyAll();
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void run() {

		//TraceRouteRun trr = null;
		while (!Main.isShuttingDown() && isActive()){
			
			synchronized(pendingRuns){
				while (pendingRuns.size()==0 && pendingRunsLowPrio.size()==0 
						&& !Main.isShuttingDown()){
					try {
						pendingRuns.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						setActive(false);
						LOGGER.info("STOP: Adios...");
						return;
					}
				}
			}
				
			synchronized(currentRuns) {
				while (currentRuns.size() >= 5 && isActive()) {
					try {
						LOGGER.info("TR: too many concurrent runs (" + currentRuns.size() + "), " + "wait for 5 secnods... ");
						LOGGER.info("TR: pedingRuns: " + pendingRuns.size() + " pendingRunsLowPrio: " + pendingRunsLowPrio.size());
						
						currentRuns.wait(5*1000);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}

			ProbeSubTask pst = null;

			synchronized(pendingRuns) {
				
				Iterator<ProbeSubTask> it;
				if (pendingRuns.size()>0){
					
					it = pendingRuns.iterator();
					pst = it.next();					
					it.remove();
					
					LOGGER.info("TR: High priority traceroute to "+pst);
				}
				
			}
			
			if (pst == null) { 
				
				synchronized(pendingRunsLowPrio) {
						
					Iterator<ProbeSubTask> it;
					if (pendingRunsLowPrio.size()>0){
						
						it = pendingRunsLowPrio.iterator();
						pst = it.next();
						it.remove();
						
						LOGGER.info("TR: Low priority traceroute to "+pst);
					}
				}
			}
				
			if (pst != null) {

				TraceRouteRun trr = new TraceRouteRun(pst);
				Main.createThread("DasuTraceRoute-" + pst.ipAddress,  trr);
				currentRuns.put(pst.ipAddress, trr);
				
			}
        
			if ( yieldInterval != -1) {
				try {
					sleep(yieldInterval);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			
		} // end while active

		LOGGER.info("STOP: Adios...");
	}
	
	@Override
	protected Object probeIp(ProbeSubTask pst) {
		
		TraceRouteRun trr = new TraceRouteRun(pst);
		Main.createThread("traceRoute " + pst.ipAddress,  trr);
		return null;
	}
	

	@Override
	public void probeAll(ProbeTask pt) {
		
		if ( pt != null) {
			synchronized(pendingRuns){
				for(ProbeSubTask pst: pt.getSubTasks())
					pendingRuns.add(pst);
				pendingRuns.notifyAll();
			}
		}
	}

	@Override
	public void probeWithPriority(ProbeSubTask pst) {
		
		if ( pst != null) {
			synchronized(pendingRuns){
				pendingRuns.add(pst);
				pendingRuns.notifyAll();
			}
		}
	}
	

	private void recordInDb(TraceRouteRun trr, String dest, ArrayList<TraceEntry> data) {
		
    	LOGGER.fine("TR: Traceroute Result for ip " + trr.pst.ipAddress + "(" + data.size() +"), taskId: " + trr.pst.taskId);
    	
    	FactTraceResult tr = new FactTraceResult();
    	tr.setTarget(trr.pst.target);
    	
        tr.dest = dest;
        tr.source = Main.getPublicIpAddress();
        tr.entries = data;
        tr.taskId = trr.pst.taskId;
        
        tr.setPst(trr.pst);
        
        updateTrResultQueue(tr);
        reportProbeResult(tr);
		
        synchronized(lossSync){
        	double dropped = 0;
        	double count = 0;
        	for (TraceEntry te : data){
        		int myDropped = 0;
        		boolean valid = false;
        		for (int i = 0; i < te.rtt.length; i++){
        			if (te.rtt[i]>0) valid = true;
        			else myDropped++;
        		} // end each entry
        		if (valid) {
        			dropped+=myDropped;
        			count+=te.rtt.length;
        		}
        	} // end all tr data
        	if (count>0){
            	lossSum+=dropped/count;
            	lossCount++;
        	}	
        } // end sync
       
	        
	}

	/*
	private synchronized void updateTrResultQueue(FactTraceResult ftr) {

		Long id = System.nanoTime();
		trResultTable.put(id, ftr);
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
					new Pair<Long, EventType>(id, EventType.HTRM_RESULT));
			AMCoordinatorV2.events.notifyAll();
		}
		
		currentRuns.remove(ftr.dest);
	}
	*/
	
	private synchronized void updateTrResultQueue(FactTraceResult ftr) {

		Long id = generateId();
		trResultTable.put(id, ftr);
		updateResultQueue(ftr, ftr.getProbeType(), id);		
		currentRuns.remove(ftr.dest);
		
	}
	
	private class TraceRouteRun implements Runnable {
		public ProbeSubTask pst;

		Process p;
	    BufferedReader in;
		ArrayList<TraceEntry> data = new ArrayList<TraceEntry>();
		
		public TraceRouteRun(ProbeSubTask pst) {
			this.pst = pst;
		}
		
		@Override
		public boolean equals(Object obj) {

			try {
				return obj!=null && obj instanceof TraceRouteRun && pst.ipAddress.equals(((TraceRouteRun)obj).pst.ipAddress) && pst.target.equals(((TraceRouteRun)obj).pst.target);
			}
			catch (Exception e) { 
				return false;
			}
		}
		@Override
		public int hashCode() {

			return pst.ipAddress.hashCode();
		}

		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {						

			String line = null;
			
			try {
				
				long now = Util.currentTimeMillis(); 
				long timeoutInMillis = 1000L * DasuConfiguration.getInstance().getTracerouteTimeout(); 
				long finish = now + timeoutInMillis; 
	            
				if (pi.getUtilities().isWindows()) {

					p = Runtime.getRuntime().exec("tracert -d -w 3000 " + pst.ipAddress);
					
	                in = new BufferedReader(new InputStreamReader(
	                            p.getInputStream()));              
	                
	                while ((line = in.readLine()) != null) {                	
	                    // parse data
	                	String[] entries = line.split("[\\s]+");
	                	if (entries.length<=1) continue;
	                	try {
	                		Integer.parseInt(entries[1]);
	                	} catch (NumberFormatException e){
	                		continue;
	                	}
	                	TraceEntry te = new TraceEntry();
	                	int i = 2;
	                	int numEntries = 0;
	                	while (i < entries.length-1){
	                        try{
	                		te.rtt[numEntries] = getValue(entries[i]);
	                        } catch (NumberFormatException e){
	                            // failure during traceroute
	                            te.router[te.numRouters++] = getRouterIp(entries[i]);
	                            break;
	                        }
	                		if (te.rtt[numEntries++]<0) i++;
	                		else i+=2;
	                		if (numEntries==te.rtt.length) break;
	                		
	                	}
	                	
	                	te.router[te.numRouters++] = getRouterIp(entries[entries.length-1]);  
	                	data.add(te);
	                	if (data.size()>=30) 
	                		break;
	                	
					    if ( Util.currentTimeMillis() > finish ) {

					    	LOGGER.info("TR: Traceroute to " + pst.ipAddress +" look longer than " + DasuConfiguration.getInstance().getTracerouteTimeout() + " seconds, aborted...");
					    	break;
					    }              
					    	
	                }

	                // then destroy
	                if (p!=null) 
	                	p.destroy();         
	                
	                
	            } else if (pi.getUtilities().isLinux() || pi.getUtilities().isOSX() 
	            		|| pi.getUtilities().isUnix() || pi.getUtilities().isFreeBSD()) {
	                	            	
	            	p = Runtime.getRuntime().exec("traceroute -n -w 3 " + pst.ipAddress);
	            	if ( Main.DEBUG )
	            		p = Runtime.getRuntime().exec("traceroute -n -w 3 -P icmp " + pst.ipAddress);
	                
	                in = new BufferedReader(new InputStreamReader(
	                            p.getInputStream()));

	                while (in!=null && (line = in.readLine()) != null) {                	
	                    // parse data
	                	
	                	String[] entries = line.split("[\\s]+");
	                	if (entries.length<=1) continue;
	                	int index = 0;
	                	while (entries[index].equals("")) index++;
	                	try {
	                		Integer.parseInt(entries[index]);
	                	} catch (NumberFormatException e){
	                		// not an entry we care about
	                		continue;
	                	}
	                	TraceEntry te = new TraceEntry();
	                	int rttCount = 0;
	                	for (int i = index+1; i < entries.length; i++){
	                		if (entries[i].contains("*")){
	                			te.router[te.numRouters++] = getRouterIp(entries[i]);
	                			te.rtt[rttCount++] = -1f;
	                			//System.out.println("Found asterisk!");
	                		}
	                		else {
	                			if(entries[i].matches("([0-9]{1,3}\\.){3}[0-9]{1,3}")){
	                				// if code reaches this point, this is an ip
	                				te.router[te.numRouters++] = getRouterIp(entries[i]);
	                				//System.out.println("Found router!");
	                			}
	                			else{
	                				// not an ip, so it must be a value
	                				try {
		                				te.rtt[rttCount] = Float.parseFloat(entries[i++]);
		                				rttCount++;
	                				} catch (NumberFormatException e ){
	                					
	                					continue; // ignore the garbage
	                				}
	                					                				
	                			}
	                		}
	                	}
	                 
	                	data.add(te);
	                	if (data.size()>30) 
	                		break;
	                	
					    if ( Util.currentTimeMillis() > finish ) {
					    						    	
					    	LOGGER.info("TR: Traceroute to " + pst.ipAddress +" took longer than " + DasuConfiguration.getInstance().getTracerouteTimeout() + " seconds, aborted...");
						    break;
					    }
					    

	                }

	                // then destroy
	                if (p!=null) 
	                	p.destroy();
	                
	            }
				
	        } catch (IOException e) {
	        	
	        } catch (Exception e){

	        }	
	        
			recordInDb(this, pst.ipAddress, data);	        
		 	
		}
		
		public void stop() {
			
			if (p!=null) {
				p.destroy();
			}
			
			try {
				finalize();
			} catch (Throwable e) {
				e.printStackTrace();
			}
		}
	}
	
	private float getValue(String val) {
		if (val.contains("<"))return 0;
		else if (val.contains("*")) return -1;
    	else return Float.parseFloat(val);
	}
	
	public double getLossSample() {
		synchronized(lossSync){
			double toReturn = lossCount==0?0:lossSum/lossCount;
			lossSum = 0;
			lossCount = 0;
			return toReturn;
		}
	}

	private String getRouterIp(String maybeIp) {
		
		if(!Util.isValidIp(maybeIp))
			maybeIp = "unknown";
		
		Statistics.getInstance().addRouterForLookup(maybeIp);
		return maybeIp;
	}
		

	@Override
	public void stopModule(){
		
		LOGGER.info("STOP: TR: stopping DasuTraceRouteModule...");
		setActive(false);
		
		synchronized(pendingRuns) {
			pendingRuns.clear();
			pendingRuns.notify();
		}
		
		synchronized(currentRuns) {
			for (String s : currentRuns.keySet())
				currentRuns.get(s).stop();
			
			currentRuns.clear();
			currentRuns.notify();
		}

		synchronized(self.pendingRuns){
			self.pendingRuns.notifyAll();
		}
		
	}
	
	
	public synchronized Integer getCurrentRuns() {
		return currentRuns.size();
	}

	@Override
	public Pair<String, FactTemplate> proccessResult(long eventId) {
		
		FactTraceResult trResult = trResultTable.remove(eventId);			
		return new Pair<String, FactTemplate>(trResult.dest, trResult);
		
	}
	

}
