/**
 * Dasu Project
 *
 * File:         AMCoordinatorV2.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Dec 15, 2009 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.logging.Logger;

import org.gudy.azureus2.core3.util.AEThread;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;
import org.gudy.azureus2.plugins.download.DownloadManager;
import org.gudy.azureus2.plugins.torrent.Torrent;
import org.gudy.azureus2.plugins.torrent.TorrentDownloader;
import org.gudy.azureus2.plugins.torrent.TorrentException;
import org.gudy.azureus2.plugins.torrent.TorrentManager;

import edu.northwestern.cs.aqualab.Cattle.json.JSONObject;
import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactProbeResult;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.drools.FactTestResult;
import edu.northwestern.dasu.drools.FactUpnpTestResult;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.drools.RuleBasedEngineInterface;
import edu.northwestern.dasu.measurement.ame.AMEManager;
import edu.northwestern.dasu.measurement.scheduler.LimitedQuotaQueue;
import edu.northwestern.dasu.measurement.scheduler.Scheduler;
import edu.northwestern.dasu.measurement.scheduler.UnlimitedQueue;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.tasks.TestTask;
import edu.northwestern.dasu.measurement.tests.AzureusSpeedTest;
import edu.northwestern.dasu.measurement.tests.AzureusUpnpTest;
import edu.northwestern.dasu.measurement.tests.DasuPortThrottlingTest;
import edu.northwestern.dasu.measurement.types.AlternateType;
import edu.northwestern.dasu.measurement.types.CronjobType;
import edu.northwestern.dasu.measurement.types.DynamicEnumFactory;
import edu.northwestern.dasu.measurement.types.EventTypeInterface;
import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.measurement.types.TaskType;
import edu.northwestern.dasu.measurement.types.TestType;
import edu.northwestern.dasu.measurement.types.TimeType;
import edu.northwestern.dasu.profilers.Profiler;
import edu.northwestern.dasu.signal.holders.HistoricSignalHolder;
import edu.northwestern.dasu.signal.holders.SessionSignalHolder;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.timer.CronjobEntry;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.util.HashMapCache;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The AMCoordinatorV2 class coordinates measurements/tests between all
 *         measurement/test modules.
 **/

public class AMCoordinatorV2 implements Runnable {

	private final static Logger LOGGER = Logger
			.getLogger("edu.northwestern.dasu");

	private static boolean active = false;
	private static boolean pause = false;
	private static AMCoordinatorV2 self = null;

	/* sleepInterval in seconds */
	private static int sleepInterval = -1;

	/* Map of ProbeType to DasuProbeModule instances */
	private static Hashtable<ProbeType, DasuProbeModule> probeModules = null;

	/************* TESTS' RELATED STRUCTURES **************/

	/* Queue of Tests to be executed by the AMC */
	private static BlockingQueue<TestTask> testQueue = null;

	/*
	 * Holds the results from dispatched tests, written to by the test thread
	 * itself
	 */
	public static Hashtable<Long, FactTestResult> testResultTable = null;

	/* Holds the tests that are already running */
	private static Hashtable<TestType, SourceType> dispatchedTests = null;

	/*************** AMC CONTROL STRUCTURES & VARIABLES ************/

	private static int shortTermInterval = 1; // multiples of one minute
	private static ITimer hourlyHistoryUpdaterTimer;

	/* holds probe/test history for the past minute */
	private static LinkedHashMap<Pair<TestType, String>, Long> shortTermTestHistory = null;
	private static Hashtable<TestType, Integer> shortTermTestTally = null;

	private static LinkedHashMap<TestType, Pair<String, Long>> temporaryShortTermTestHistory = new LinkedHashMap<TestType, Pair<String, Long>>();

	/* holds the hourly tally (count) of probes */
	private static Hashtable<TestType, Integer> hourlyTestTally = null;

	/*
	 * To keep track of Inference-Engine launched priority tasks that will awake
	 * the AMC
	 */
	private static BlockingQueue<ProbeType> REProbeTaskQueue = null;

	/*
	 * Queue used by every DasuProbeModule (and Drools Engine) used to alert
	 * (awake) the AMC of probe results it needs to collect (or of probe tasks
	 * it needs to perform)
	 */
	public static PriorityBlockingQueue<Pair<Long, EventTypeInterface>> events = null;

	/*
	 * Used by the AMC to keep track of the probes already dispatched and to
	 * know which ProbeTask launched what ProbeType when the results are back
	 */
	private static Hashtable<Pair<ProbeType, String>, SourceType> dispatchedEvents = null;

	/* probeTask history */
	private static HashMapCache<Integer, Long> taskLogHistory = null;

	private static LinkedList<Download> activeTorrents = new LinkedList<Download>();

	private static Hashtable<String, HistoricSignalHolder> historicSamples = null;

	private static Hashtable<String, SessionSignalHolder> sessionSamples = new Hashtable<String, SessionSignalHolder>();

	private static HashMap<ProbeType, Scheduler> probeSchedulers = new HashMap<ProbeType, Scheduler>(); 

	private static Scheduler getScheduler(ProbeType pt) {

		if ( pt == null )
			return null;
		
		if (!probeSchedulers.containsKey(pt)) {
			Scheduler s = new Scheduler();
			s.addQueue(SourceType.RE.toString(), new UnlimitedQueue());
			s.addQueue(
					"not-RE",
					new LimitedQuotaQueue(pt.getMaxPerMin(), pt.getMaxPerHour()));
			probeSchedulers.put(pt, s);
		}
		return probeSchedulers.get(pt);
	}

	private static String getSchedulerQueue(ProbeTask pt) {
		if (pt.sourceType == SourceType.RE)
			return SourceType.RE.toString();
		return "not-RE";
	}

	private AMCoordinatorV2() {

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinator.class: startDasuAmModules()");

		sleepInterval = DasuConfiguration.getInstance().getAMCSleepInterval();
		shortTermTestHistory = DasuManager.getPersistentDataManager()
				.getShortTermTestHistory();
		shortTermTestTally = DasuManager.getPersistentDataManager()
				.getShortTermTestTally();

		hourlyTestTally = DasuManager.getPersistentDataManager()
				.getHourlyTestTally();

		taskLogHistory = DasuManager.getPersistentDataManager()
				.getTaskLogHistory();

		historicSamples = DasuManager.getPersistentDataManager()
				.getHistoricSamplesV2();

	}

	public static AMCoordinatorV2 getInstance() {
		if (self == null) {
			self = new AMCoordinatorV2();
		}
		return self;
	}

	public static Profiler profiler = null;

	public static void profile(Profiler p) {
		profiler = p;
	}

	/*
	 * private static void initializeProbeLimits() { // XXX: These should be in
	 * a configuration file (JSO) // XXX: Not really, configuration files make
	 * the values easy to change, // these thresholds should be hard-coded in
	 * the source (MAS)
	 * 
	 * maxProbePerHour.put(ProbeType.PING, 600);
	 * maxProbePerHour.put(ProbeType.TRACEROUTE, 600);
	 * maxProbePerHour.put(ProbeType.NDT, 1);
	 * maxProbePerHour.put(ProbeType.HTTP, 600);
	 * maxProbePerHour.put(ProbeType.NDT2, 5);
	 * maxProbePerHour.put(ProbeType.SCRAPE, 60);
	 * maxProbePerHour.put(ProbeType.DNS, 2000);
	 * 
	 * maxTestPerHour.put(TestType.AZUREUS_DL, 2);
	 * maxTestPerHour.put(TestType.AZUREUS_UL, 2);
	 * maxTestPerHour.put(TestType.AZUREUS_DL_ENCRYPTION, 2);
	 * maxTestPerHour.put(TestType.AZUREUS_UL_ENCRYPTION, 2);
	 * maxTestPerHour.put(TestType.AZUREUS_UPNP, 100);
	 * 
	 * maxProbePerMin.put(ProbeType.PING, 20);
	 * maxProbePerMin.put(ProbeType.TRACEROUTE, 20);
	 * maxProbePerMin.put(ProbeType.HTTP, 20); maxProbePerMin.put(ProbeType.NDT,
	 * 1); maxProbePerMin.put(ProbeType.NDT2, 1);
	 * maxProbePerMin.put(ProbeType.SCRAPE, 1);
	 * maxProbePerMin.put(ProbeType.DNS, 20);
	 * 
	 * maxTestPerMin.put(TestType.AZUREUS_UPNP, 10);
	 * maxTestPerMin.put(TestType.AZUREUS_DL, 2);
	 * maxTestPerMin.put(TestType.AZUREUS_UL, 2);
	 * maxTestPerMin.put(TestType.AZUREUS_DL_ENCRYPTION, 2);
	 * maxTestPerMin.put(TestType.AZUREUS_UL_ENCRYPTION, 2);
	 * 
	 * for (TestType t : TestType.values()) {
	 * 
	 * if (!shortTermTestTally.contains(t)) { shortTermTestTally.put(t, 0);
	 * hourlyTestTally.put(t, 0); } }
	 * 
	 * launchHourlyProbeHistoryUpdater();
	 * 
	 * }
	 */

	private static void initializeProbeLimits() {

		for (TestType t : TestType.values()) {

			if (!shortTermTestTally.contains(t)) {
				shortTermTestTally.put(t, 0);
				hourlyTestTally.put(t, 0);
			}
		}

		launchHourlyProbeHistoryUpdater();

	}

	private static void initialize() {
		probeModules = new Hashtable<ProbeType, DasuProbeModule>();
		testQueue = new LinkedBlockingQueue<TestTask>();
		testResultTable = new Hashtable<Long, FactTestResult>();
		dispatchedTests = new Hashtable<TestType, SourceType>();
		REProbeTaskQueue = new LinkedBlockingQueue<ProbeType>();
		events = new PriorityBlockingQueue<Pair<Long, EventTypeInterface>>();
		dispatchedEvents = new Hashtable<Pair<ProbeType, String>, SourceType>();

		try {
			startDasuAmModules();
			initializeProbeLimits();
		}
		catch (Exception e) {
			LOGGER.severe("DYNAMIC: Probe Modules could no be loaded.. abort loading dasu...");
			Main.dasuUnload();
		}
		
		setActive(true);
	}

	/*
	 * Initialize each DasuModule available
	 */

	private static void startDasuAmModules() throws IOException {

		Properties map = DasuConfiguration.getInstance()
				.getProbeModulesProperties();
		
		//use default
		if ( map.size() == 0 ) {	        
	         map.load(DasuManager.getDasuClassLoader().getResourceAsStream(
	        		 DasuConfiguration.PROBE_MODULES_PROPERTIES));
		}
		
		LOGGER.info("DYNAMIC: loading modules: " + map);
		
		try {
			
			//add new ProbeTypes if they don't already exit
			
			for ( Object probeType : map.keySet()) {
				
				try {
					
					if ( !map.getProperty(probeType.toString()).contains("," ))
						continue;
					
					String maxVals[] = map.getProperty(probeType.toString()).split(":")[1].split(","); 
					int maxProbesPerHour = Integer.parseInt(maxVals[0]);
					int maxProbesPerMin = Integer.parseInt(maxVals[1]);					
										
					if (!Util.probeTypeExists(probeType.toString())) {
						
						//doesnt exit, add it												 
						DynamicEnumFactory.addEnum(ProbeType.class, probeType.toString(), 
								new Class[]{int.class, int.class}, new Object[]{maxProbesPerHour,maxProbesPerMin});
						LOGGER.info("DYNAMIC: adding a new probeType: "	+ probeType + " " 
								+ maxProbesPerHour + " " + maxProbesPerMin);
						
					}
					
				}
				catch (Exception e) {
					//malformed new probeType, skip
					e.printStackTrace();
					LOGGER.info("DYNAMIC: malformed new probeType, skipping: "	+ probeType + " " +
							map.getProperty(probeType.toString()));

				}
				
			}


			//ClassLoader cl = DasuManager.getDasuClassLoader();
			ClassLoader cl = Main.getPluginInterface().getPluginClassLoader();

			for (ProbeType pt : ProbeType.values()) {

				try {
					if (map.containsKey(pt.toString())) {
						LOGGER.info("DYNAMIC: adding: "	+ map.getProperty(pt.toString()));
						String probeType = map.getProperty(pt.toString()).split(":")[0];
						probeModules.put(pt, (DasuProbeModule) cl.loadClass(probeType).newInstance());
						LOGGER.info("DYNAMIC: Probe Type: " + pt);
						LOGGER.info("DYNAMIC: Probe Path: " + probeType);
						if (!probeModules.get(pt).isAlive()) {
							probeModules.get(pt).setDaemon(true);
							probeModules.get(pt).start();
							LOGGER.info("DYNAMIC: starting: "+ probeType + 
									" version: " + probeModules.get(pt).getVersion());
						}
					}
				} catch (Exception e1) {
					e1.printStackTrace();
				}

			}

		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	private static void stopDasuAmModules() {

		for (ProbeType pt : ProbeType.values()) {
			if (probeModules != null && probeModules.get(pt) != null) {
				if (probeModules.get(pt).isAlive()) {

					if (pt.equals(ProbeType.NDT)
							&& !probeModules.get(pt).getState()
									.equals(Thread.State.TIMED_WAITING))
						probeModules.get(pt).interrupt();

					LOGGER.info("STOP: " + pt + " "  + probeModules.get(pt).getName());
					probeModules.get(pt).stopModule();
				}
			}
		}
	}

	/* Add to the *task* queue */

	public static boolean addTestTask(TestType testType) {
		return addTestTask(testType, SourceType.RE, null);
	}

	public static boolean addTestTask(TestType testType, Object[] params) {
		return addTestTask(testType, SourceType.RE, params);
	}

	public static boolean addTestTask(TestType testType, Object param) {
		return addTestTask(testType, SourceType.RE, new Object[] { param });
	}

	public static boolean addTestTask(TestType testType, SourceType sourceType,
			Object[] params) {

		updateTestHistory();

		if (!testType.equals(TestType.AZUREUS_UPNP)
				&& !testType.equals(TestType.FORCE_SEEDING_THROUGH_PORT)
				&& !DasuConfiguration.getInstance().isEnableAzureusSpeedTest()) {

			LOGGER.info("DROP: Azureus Speed test disabled at this time...");

			return false;
		}

		if ((sourceType == SourceType.RE)
				|| (shortTermTestTally.get(testType) < testType.getMaxPerMin()
						* shortTermInterval)
				&& (hourlyTestTally.get(testType) < testType.getMaxPerHour())) {
			// || (shortTermTestTally.get(testType) <
			// maxTestPerMin.get(testType) * shortTermInterval)
			// && (hourlyTestTally.get(testType) <
			// maxTestPerHour.get(testType))) {

			if (params != null)
				return testQueue
						.add(new TestTask(testType, params, sourceType));
			else
				return testQueue.add(new TestTask(testType, sourceType));

		} else {
			LOGGER.info("DROP: Too many tests of type " + testType
					+ ", try again later... " + "ShortTerm: "
					+ shortTermTestTally.get(testType) + " hourly: "
					+ hourlyTestTally.get(testType));
		}

		return false;
	}

	/* Add to the *task* queue */
	public static boolean addProbeTask(ProbeTask pt) {

		if ( pt == null)
			return false;
		
		Scheduler s = getScheduler(pt.probeType);

		if ( s == null ) {
			LOGGER.severe("Queue does not exist " + pt.probeType);
			return false;
		}
		
		if (Main.DEBUG) {

			StringBuilder sb = new StringBuilder();
			for (ProbeType p : probeSchedulers.keySet())
				sb.append("AMC: " + p + " " + getScheduler(p).size() + " " + probeModules.get(p).getState() + "\n");

			LOGGER.info(sb.toString());
		}

		/* if task is already in the log, don't add it */
		//if (taskLogHistory.containsKey((pt.taskId))) {
		//	LOGGER.info("AMC: task already in log, don't add it");
		//	return false;
		//}

		/* if task is in the past, don't add to queue */
		if (pt.endTime >= Util.currentDasuTime())
			return false;

		// try to schedule
		boolean success = s.push(getSchedulerQueue(pt), pt);

		if (profiler != null) {
			if (success) {
				profiler.probeQueued(pt);
			} else {
				profiler.probeDenied(pt);
			}
		}

		/*
		 * If and ONLY if the task is of highest priority, add it to the
		 * *events* queue in order to awake the AMC, otherwise let the AMC deal
		 * with it on next scheduled awake
		 */

		if (pt.priority == -1) {

			synchronized (events) {
				Long id = System.nanoTime();
				if (events.size() != 0)
					id = events.peek().getKey() - 1000000;

				events.add(new Pair<Long, EventTypeInterface>(id,
						AlternateType.RE_REQUEST));
				events.notifyAll();
			}
		}

		return success;
	}

	private static boolean validateProbe(String ip, ProbeType pt) {
		
		if (pt.equals(ProbeType.NDT) || pt.equals(ProbeType.NDT2)) {
			
			LinkedList<String> servers = new LinkedList<String>();
			try {
				servers = (LinkedList<String>) probeModules.get(pt).getClass().getMethod("getServers", new Class<?>[]{})
						.invoke(probeModules.get(pt), new Object[]{});
			}
			catch (Exception e) {
				e.printStackTrace();
			}
			
			if ( servers.size() == 0 && ip == null) {
				LOGGER.info("DYNAMIC: Invalid server name " + ip);
				return false;
			}

		}
		
		return true;
	}

	public static boolean addLowPriorityProbe(String ip, ProbeType pt,
			SourceType st) {

		if (!validateProbe(ip, pt))
			return false;

		ArrayList<String> ipList = new ArrayList<String>();
		ipList.add(ip);

		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, ipList, pt,
				TaskType.BULK, st);

		task.sourceType = st;

		return addProbeTask(task);

	}

	/* To be used only by the Detection Rule Engine */
	public static boolean addPriorityProbe(String ip, ProbeType pt) {
		return addPriorityProbe(ip, pt, null);
	}

	/* To be used only by the Detection Rule Engine */
	public static boolean addPriorityProbe(String ip, ProbeType pt,
			String target) {
		
		LOGGER.info("AMC: probe " + ip + " " + pt + " " + target);		
		if (!validateProbe(ip, pt)) {
			LOGGER.info("DYNAMIC: probe could not be validated " + ip + " " + pt );
			return false;			
		}
		
		ArrayList<String> ipList = new ArrayList<String>();
		ipList.add(ip);

		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, ipList, pt,
				TaskType.BULK, SourceType.RE);
		task.setTarget(target);

		return addPriorityProbe(task);
	}

	/* To be used only by the Detection Rule Engine */
	public static boolean addPriorityProbe(ProbeTask task) {

		boolean ret = false;

		// force the source of the task to be the RE, in case the user didn't
		task.sourceType = SourceType.RE;

		if (REProbeTaskQueue != null) {

			REProbeTaskQueue.add(task.probeType);
			ret = addProbeTask(task);
			if (!ret)
				REProbeTaskQueue.remove(task.probeType);

		}

		return ret;

	}

	/*
	 * Returns (removes) ONE measurement from List member of the top-most
	 * ProbeTask in the Queue. Does not remove the ProbeTask from Queue unless
	 * the ProbeTask List is empty
	 * 
	 * XXX: JSO: this code is no longer called, but still has some logic that
	 * should be moved elsewhere.
	 */
	private static ProbeSubTask popProbeSubTask(ProbeType pt) {

		PriorityBlockingQueue<Pair<Long, ProbeTask>> taskQueue = null; // probeTaskQueue.get(pt);

		ProbeSubTask ret = null;
		Pair<Long, ProbeTask> firstTask = null;
		firstTask = taskQueue.peek();

		while (firstTask != null) {

			/* it's priority task or endTime = whenever possible */
			if ((firstTask.getValue().taskId == -1)
					|| (firstTask.getKey() == -1))
				break;

			/*
			 * endTime already passed, remove from queue and try again if there
			 * are more tasks in the queue
			 */
			else if (firstTask.getValue().endTime >= Util.currentDasuTime()) {

				taskQueue.poll();
				firstTask = null;
				if (taskQueue.size() == 0)
					break;
			}

			/* it's not time to start the task yet */
			else if (firstTask.getValue().startTime < Util.currentDasuTime()) {
				firstTask = null;
				break;
			}
		}

		if (firstTask != null) {
			ProbeTask task = firstTask.getValue();

			if (task.ipsToProbe.size() == 0) {
				taskQueue.poll();
			}

			/* if next probe is already dispatched (repeated), do nothing */

			else if (!dispatchedEvents.keySet().contains(
					new Pair<ProbeType, String>(firstTask.getValue().probeType,
							firstTask.getValue().ipsToProbe.get(0)))) {

				ret = new ProbeSubTask(task.ipsToProbe.remove(0),
						task.sourceType, task.taskId, task.params, task.target);

				ret.setParent(task);

				if (task.ipsToProbe.size() == 0)
					taskQueue.poll();

			}
		}
		return ret;
	}

	private static void launchNDTProbe(ProbeType pt, ProbeSubTask pst) {
		
		LOGGER.info("NDT: launching ndt...");
		
		if (profiler != null) {
			profiler.probeScheduled(pst);
		}

		/* pause all other probes */
		for ( ProbeType probetype : probeModules.keySet() ) {
			if ( !probetype.equals(ProbeType.NDT) &&  !probetype.equals(ProbeType.NDT2))
				probeModules.get(probetype).setPause(true);
		}
		
		int i = 0;

		boolean allPaused = true;
		while ( i++ < 10 ) {
			
			allPaused = true;
			for ( ProbeType probetype : probeModules.keySet() ) {
				if ( !probetype.equals(ProbeType.NDT) &&  !probetype.equals(ProbeType.NDT2))
					allPaused &= probeModules.get(probetype).getState() == Thread.State.TIMED_WAITING;
			}
		
			if ( !allPaused ) {
		
				if (Main.DEBUG)
					LOGGER.info("NDT: AMCoordinatorV2.class: Waiting for ProbeModules to pause...");

				try {
					Thread.sleep(2 * 1000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				
			}
			else
				break;
			
		}
		
		if ( !allPaused ) {
			LOGGER.info("NDT: Not every other probe modules paused within 20 seconds, aborting NDT...");
			for ( ProbeType probetype : probeModules.keySet() ) {
				if ( !probetype.equals(ProbeType.NDT) &&  !probetype.equals(ProbeType.NDT2))
					probeModules.get(probetype).setPause(false);
			}
		}
		else {
		
			/* pause all torrents */
			/* NDT2 does not stop torrents */
			if (pt.equals(ProbeType.NDT)) {

				if (Main.DEBUG)
					LOGGER.info("NDT: AMCoordinatorV2.class: Temporarily stopping all ("
							+ activeTorrents.size()
							+ ") active torrents");

				stopActiveTorrents();
			}

			probeModules.get(pt).probe(pst);
			
			dispatchedEvents.put(new Pair<ProbeType, String>(pt,
					(pst.ipAddress == null) ? "" : pst.ipAddress),
					pst.sourceType);

		}

	}

	/*
	 * Retrieves a probe of the specified *ProbeType* from the taskQueue and, if
	 * one is available, passes the target *ip* to the corresponding DasuModule
	 * for execution
	 */
	private static void launchProbe(ProbeType pt) {

		if (pt == null)
			return;

		Scheduler s = getScheduler(pt);		
		ProbeSubTask pst = null;
		
		if ( s == null )
			return;
		
		Integer skipped = 0;
		while (skipped < s.size() ) {
				
			pst = s.peek();	
			if ( pst != null ) {
				if (AMEManager.AMEngines.containsKey(pst.getSourceType())
						&& !AMEManager.AMEngines.get(pst.getSourceType())
								.schedulable(pst)) {
					s.pollAndPushToBack();
					LOGGER.info("AMC: skip!!!! " + pt + "\n\n\n\n\n");
					
				}
				else {
					break;
				}
			}
			else
				break;
			
			pst = null;
			skipped++;
			
		}
		
		if (pst != null) {
			
			LOGGER.info("AMC: launch " + pt + " " + pst);

			/* TODO: add simultaneous probe support */
			if (probeModules.get(pt).getState() == Thread.State.TIMED_WAITING) {

				pst = s.poll();

				if (pst != null) {
					if (profiler != null) {
						profiler.probeScheduled(pst);
					}

					// give low priority to traceroutes from the internal rule engine
					if ((pt.equals(ProbeType.TRACEROUTE))
							&& (pst.sourceType.equals(SourceType.NULL))) {
						try {
							probeModules
									.get(pt)
									.getClass()
									.getMethod("probeWithLowPriority",
											ProbeTask.ProbeSubTask.class)
									.invoke(probeModules.get(pt), pst);
						} catch (Exception e) {
							e.printStackTrace();
						}

					} else {
						
						if (pt.equals(ProbeType.NDT) || pt.equals(ProbeType.NDT2)) {
							launchNDTProbe(pt, pst);
							return;
						}
						
						probeModules.get(pt).probe(pst);
					}

					dispatchedEvents.put(new Pair<ProbeType, String>(pt,
						(pst.ipAddress == null) ? "" : pst.ipAddress),
						pst.sourceType);

				}
			}
			else {
				LOGGER.info("AMC: delay, probeModule not idle " + pt + " " + pst);
			}
		}
	}

	private void launchTest() {

		updateTestHistory();

		if (dispatchedTests.size() == 0) {

			TestTask test = testQueue.poll();

			if ((shortTermTestTally.get(test.testType) < test.testType
					.getMaxPerMin() * shortTermInterval)
					&& (hourlyTestTally.get(test.testType) < test.testType
							.getMaxPerHour())) {

				if (test.params != null)
					appendToTemporaryTestHistory(test.testType,
							Integer.toString((Integer) test.params.peek()),
							Util.currentTimeMillis());
				else
					appendToTemporaryTestHistory(test.testType, null,
							Util.currentTimeMillis());

				switch (test.testType) {

				case AZUREUS_UPNP:
					if ((test.params != null))
						doUpnpTest((Integer) test.params.pop());
					else
						doUpnpTest();
					break;

				case FORCE_SEEDING_THROUGH_PORT:
					if ((test.params != null) && (test.params.size() > 1))
						doForceSeedingThroughPort((Integer) test.params.pop(),
								(Integer) test.params.pop());
					else
						doForceSeedingThroughPort((Integer) test.params.pop());
					break;

				default:
					doSpeedTest(test.testType);
					break;

				}

				dispatchedTests.put(test.testType, test.sourceType);
			}

			/* return to queue */
			else {
				testQueue.add(test);

				LOGGER.info("RESOURCEPROFILER: WAIT: Too many tests of type "
						+ test.testType
						+ ", wait for some time... "
						+ "ShortTerm: "
						+ shortTermTestTally.get(test.testType)
						+ " hourly: " + hourlyTestTally.get(test.testType));

			}

		}

	}

	public LinkedList<String> getProbeModules() {
		// Push list of probe module names to DasuData
		LinkedList<String> l = new LinkedList<String>();
		for (DasuProbeModule pm : probeModules.values()) {
			l.add(pm.getName());
		}
		return l;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Runnable#run()
	 */

	public void run() {

		initialize();

		try {
			Thread.sleep(5 * 1000);
		} catch (InterruptedException e1) {
			e1.printStackTrace();
		}

		if (DasuManager.USE_SIMULATION_V2)
			LOGGER.info("SIMULATOR: AMCoordinatorV2.class: All AMC requests are being skipped dut to SIMULATION mode!!");

		while (isActive()) {

			try {
				LOGGER.fine("RESOURCEPROFILER: AMCoordinatorV2: hourlyTestTally: "
						+ hourlyTestTally);
				LOGGER.fine("RESOURCEPROFILER: AMCoordinatorV2: shortTermTestTally: "
						+ shortTermTestTally);
				LOGGER.fine("RESOURCEPROFILER: AMCoordinatorV2: shortTermTestHistory: "
						+ shortTermTestHistory);
			} catch (Exception e) {

			}

			for (ProbeType p : probeModules.keySet())
				LOGGER.fine("RESOURCEPROFILER: AMCoordinatorV2: " + p + " "
						+ probeModules.get(p).getState());

			if (!dispatchedEvents.contains(ProbeType.NDT)) {

				if (testQueue.size() != 0)
					launchTest();

				if (!dispatchedSpeedTest()) {

					if (!dispatchedEvents.contains(ProbeType.NDT)) {

						for (ProbeType pt : probeModules.keySet()) {
							launchProbe(pt);
						}
					}
				}

				try {

					if (Main.DEBUG)
						LOGGER.info("DEBUG: AMCoordinatorV2: sleeping for "
								+ sleepInterval + " seconds...");

					Pair<Long, EventTypeInterface> event = null;
					synchronized (events) {

						if (isActive() && events.isEmpty() || DasuManager.USE_SIMULATION_V2)
							events.wait(sleepInterval * 1000);
						else
							event = events.poll();
					}

					if (event != null) {
						if (event.getValue().equals(AlternateType.HALT))
							setActive(false);
						else {

							try {
								processEvent(event);
							} catch (Exception e) {
								e.printStackTrace();
							}
						}
					}

				} catch (InterruptedException e) {
					e.printStackTrace();
				}

			}
		}
		
		LOGGER.info("STOP: Adios...");

	}

	private void processEvent(Pair<Long, EventTypeInterface> event) {

		/*
		 * if (event.getValue() == EventType.UPNP_RESULT)
		 * processUpnpTest(event.getKey()); else if (event.getValue() ==
		 * EventType.SPEEDTEST_RESULT) processSpeedTest(event.getKey()); else if
		 * (event.getValue() == EventType.FORCED_SEEDING_THROUGH_PORT_RESULT)
		 * processForcedSeedingThroughPortTest(event.getKey()); else if
		 * (event.getValue() == EventType.HNDT_RESULT)
		 * processNdt(event.getKey()); else if (event.getValue() ==
		 * EventType.RE_REQUEST) { if (!dispatchedSpeedTest())
		 * processAMRequest(event.getKey()); } else
		 * processProbeModuleResult(event.getKey(), (ProbeType)
		 * event.getValue().getActionType());
		 */

		if (event.getValue() instanceof ProbeType)
			processProbeModuleResult(event.getKey(),
					(ProbeType) event.getValue());
		else if (event.getValue() instanceof TestType)
			processTestResult(event.getKey(), (TestType) event.getValue());
		else if (event.getValue() instanceof AlternateType) {
			if (!dispatchedSpeedTest())
				processAMRequest(event.getKey());
		}

	}

	/*
	 * private void processEvent(Pair<Long, EventType> event) {
	 * 
	 * switch (event.getValue()) { case HPM_RESULT: processPing(event.getKey());
	 * break; case HTRM_RESULT: processTraceRoute(event.getKey()); break; case
	 * HHM_RESULT: processHttp(event.getKey()); break; case HNDT_RESULT:
	 * processNdt(event.getKey()); break; case UPNP_RESULT:
	 * processUpnpTest(event.getKey()); break; case SPEEDTEST_RESULT:
	 * processSpeedTest(event.getKey()); break; case
	 * FORCED_SEEDING_THROUGH_PORT_RESULT:
	 * processForcedSeedingThroughPortTest(event.getKey()); break; case
	 * RE_REQUEST: if (!dispatchedSpeedTest()) processAMRequest(event.getKey());
	 * break; case SCRAPE_RESULT: processScrape(event.getKey()); break; case
	 * DNS_RESULT: processDns(event.getKey()); break;
	 * 
	 * }
	 * 
	 * }
	 */

	/*
	 * Only used to process measurements submitted by the Detection Engine. The
	 * parameter eventId is not necessary in this context, but receive it anyway
	 * in order to keep interface (API) consistent
	 */
	private void processAMRequest(long eventId) {

		ProbeType pt = REProbeTaskQueue.poll();
		launchProbe(pt);

	}

	private static RuleBasedEngineInterface getEngine(SourceType source) {


		RuleBasedEngineInterface eng = null;
		if ( source != null ) {
			if (source.equals(SourceType.RE))
				eng = edu.northwestern.dasu.DasuManager.engine;
			else if (!source.equals(SourceType.NULL))
				// eng = DasuManager.RmiServer;
				// eng = RmiServerAMCoordinatorV2.getInstance();
				eng = AMEManager.rmiAMCoordinator;
		}

		return eng;

	}

	public static void stop() {
		
		LOGGER.info("STOP: stopping AMCoordinaorV2...");
		synchronized (events) {
			setActive(false);
			events.add(new Pair<Long, EventTypeInterface>(0L,
					AlternateType.HALT));
			events.notify();
		}

		if (Main.DEBUG)
			LOGGER.info("STOP: Stopping dasu probe modules...");

		stopDasuAmModules();
		hourlyHistoryUpdaterTimer.destroy();

		self = null;
		
	}

	/********* SPECIAL OPERATIONS ********/

	// XXX: Main.DEBUGGING CODE -- should be deleted?
	public static void addTorrentDownload() {

		try {

			DownloadManager dm = Main.getPluginInterface().getDownloadManager();
			// URL url = new
			// URL("http://torrent.unix-ag.uni-kl.de/torrents/KNOPPIX_V6.2.1CD-2010-01-31-EN.torrent");
			URL url = new URL(
					"http://dl.btjunkie.org/torrent/Redhat-Linux-Enterprise-5-Exclusive-DVD-ISO/40323d496af676894143463d5517b4d91f184964f068/download.torrent");
			TorrentManager tm = Main.getPluginInterface().getTorrentManager();
			TorrentDownloader tdlnd = tm.getURLDownloader(url);
			Torrent torrent = tdlnd.download();

			// Torrent torrent = tm.createFromBEncodedFile(torrent_file);
			dm.addDownload(torrent);

			Download dlds[] = dm.getDownloads();
			if (Main.DEBUG)
				for (Download d : dlds) {
					LOGGER.info("DEBUG: AMCoordinatorV2: " + d.getName());
				}

		} catch (DownloadException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (TorrentException e) {
			e.printStackTrace();
		}
	}

	// XXX: Main.DEBUGGING CODE -- should be deleted?
	@SuppressWarnings("deprecation")
	public static void doSpeedTest(final TestType testType) {

		try {
			Thread.sleep(10000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

		System.setProperty("azureus.speed.test.challenge.jar.path",
				"/Users/Mario/Desktop/Azureus2.jar");
		String explicit_path = System.getProperty(
				"azureus.speed.test.challenge.jar.path", null);

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinatorV2: " + explicit_path);

		Thread t = new AEThread("SpeedTest Performer") {
			public void runSupport() {
				AzureusSpeedTest hst = new AzureusSpeedTest();
				hst.runTest(testType);
			}
		};

		t.setPriority(Thread.MIN_PRIORITY);
		t.setDaemon(true);
		t.start();

	}

	public static void doUpnpTest() {
		doUpnpTest(12345);
	}

	public static void doUpnpTest(int portNumber) {

		try {
			Thread.sleep(10000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		AzureusUpnpTest ant = new AzureusUpnpTest();
		ant.runTest(portNumber);

	}

	public static void doForceSeedingThroughPort(long port) {
		DasuPortThrottlingTest hptt = new DasuPortThrottlingTest(port);
		hptt.start();
	}

	public static void doForceSeedingThroughPort(long port, int sleepInterval) {
		DasuPortThrottlingTest hptt = new DasuPortThrottlingTest(port,
				sleepInterval);
		hptt.start();
	}

	public static void assertToEngine(SourceType sourceType, FactTemplate fact) {

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinatorV2: Asserting to engine "
					+ sourceType + " fact " + fact);

		RuleBasedEngineInterface eng = getEngine(sourceType);
		if (eng != null)
			eng.addFact(fact, sourceType);

	}

	/************ PROCESS RESULTS FROM MEASUREMENT/TEST MODULES ***********/

	private static void processTestResult(long eventId, TestType ttype) {

		FactTestResult testResult = testResultTable.remove(eventId);

		SourceType sourceEngine = dispatchedTests.remove(testResult
				.getTestType());
		assertToEngine(sourceEngine, testResult);

		Statistics.getInstance().addReportableFact(testResult);

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinatorV2: " + testResult);

		/* if test was successful, commit to history, else rollback */
		if (testResult.getTestType().equals(TestType.AZUREUS_UPNP)) {

			if (((FactUpnpTestResult) testResult).error)
				rollbackFromTestHistory(testResult.testType);
			else
				commitToTestHistory(testResult.testType);

		}

		else if (!testResult.getTestType().equals(
				TestType.FORCE_SEEDING_THROUGH_PORT)) {

			if (!testResult.isSuccess())
				rollbackFromTestHistory(testResult.getTestType());
			else
				commitToTestHistory(testResult.getTestType());
		}

	}

	/*
	 * NDT requires special treatment because requires other modules to be
	 * stopped!
	 */

	private static void processNdt(ProbeType pt) {

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinatorV2: processNdt");
		
		LOGGER.info("NDT: Restarting probe modules...");
		for ( ProbeType probetype : probeModules.keySet() ) {
			if ( !probetype.equals(ProbeType.NDT) &&  !probetype.equals(ProbeType.NDT2))
				probeModules.get(probetype).setPause(true);
		}		
		
		/* restart all torrents */
		if (pt.equals(ProbeType.NDT)) {
			restartActiveTorrents();
		}

//
//		DasuNdtModule dasuNdtModule = (DasuNdtModule) probeModules
//				.get(ProbeType.NDT);
//		FactNdtProbeResult ndtResult = dasuNdtModule.ndtResultTable
//				.remove(eventId);
//
//		if (profiler != null) {
//			profiler.probeCompleted(ndtResult.pst);
//		}
//
//		String server = ndtResult.server;
//
//		if ((!dispatchedEvents.containsKey(new Pair<ProbeType, String>(
//				ProbeType.NDT, server)))
//				&& (!dispatchedEvents.containsKey(new Pair<ProbeType, String>(
//						ProbeType.NDT2, server)))) {
//			server = "";
//		}
//
//		ProbeType pt = ProbeType.NDT;
//		if (!dispatchedEvents.containsKey(new Pair<ProbeType, String>(pt,
//				server)))
//			pt = ProbeType.NDT2;
//
//		SourceType sourceEngine = dispatchedEvents
//				.remove(new Pair<ProbeType, String>(pt, server));
//
//		assertToEngine(sourceEngine, ndtResult);
//
//		/* restart all probes */
//		probeModules.get(ProbeType.HTTP).setPause(false);
//		probeModules.get(ProbeType.PING).setPause(false);
//		probeModules.get(ProbeType.TRACEROUTE).setPause(false);
//		probeModules.get(ProbeType.SCRAPE).setPause(false);
//		probeModules.get(ProbeType.DNS).setPause(false);
//
//		/* restart all torrents */
//		if (pt.equals(ProbeType.NDT)) {
//			restartActiveTorrents();
//		}
	}

	private static void processProbeModuleResult(long eventId, ProbeType ptype) {

		@SuppressWarnings("rawtypes")
		DasuProbeModule probeModule = probeModules.get(ptype);
		
		Pair<String, FactProbeResult> resultPair = probeModule
				.proccessResult(eventId);

		String eventLabel = resultPair.getKey();
		FactProbeResult factResult = resultPair.getValue();

		if (profiler != null) {
			profiler.probeCompleted(factResult.pst);
		}

		if ( ptype.equals(ProbeType.NDT)  || ptype.equals(ProbeType.NDT2)) {
			
			//if (!dispatchedEvents.containsKey(new Pair<ProbeType, String>(
			//		ptype, eventLabel))) {
			//	eventLabel = "";				
			//}		
			processNdt(ptype);
		}
		
		SourceType sourceEngine = dispatchedEvents
				.remove(new Pair<ProbeType, String>(ptype, eventLabel));
		assertToEngine(sourceEngine, factResult);
	}

	public static String[] getPropertiesFromServer(String url) {		
		HashMap<String, String> map = Main.getPropertiesFromServer(url);
		return (String[]) map.values().toArray(new String[map.size()]);
		
	}
	
	public static HashMap<String, String> getPropertiesFromServer(Integer taskId, String resource) {
		
		HashMap<String, String> ret = new HashMap();
		
		JSONObject json = DasuManager.getInstance().getConfigService().GETData( DasuConfiguration.DASU_SCRIPT_AME_ON_DEMAND + 
				DasuConfiguration.getInstance().getResourceScriptRequest() + "&taskid=" + taskId + "&resource=" + resource);
		if ( json != null ) {
						
			Iterator<String> iter = json.keys();
			while ( iter.hasNext() ) {					
				String key = iter.next();
				
				try {
					ret.put( key, json.getString(key) );
				}
				catch (Exception e) {
					LOGGER.info("DYNAMIC: ERROR: " + e.getMessage());
				}
			}
		}
		//ArrayList<String> l = Arrays.asList(ret.values().toArray(new String[ret.size()]));
		return ret;
		
	}
	
	public static JSONObject getJSONObjectFromServer(Integer taskId, String resource) {
		
		
		JSONObject ret = new JSONObject();
		
		JSONObject json = DasuManager.getInstance().getConfigService().GETData( DasuConfiguration.DASU_SCRIPT_AME_ON_DEMAND + 
				DasuConfiguration.getInstance().getResourceScriptRequest() + "&taskid=" + taskId + "&resource=" + resource);
		
		if ( json != null ) {
			ret = json;
		}
		
		return ret;
		
	}

	/****** CRONJOB SCHEDULER INTERFACE ********/

	public static void removeCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute) {
		removeCronJob(id, hour, day, commandToExecute, CronjobType.CUSTOM);
	}

	public static void removeCronJob(String id, String commandToExecute,
			CronjobType cronjobType) {
		removeCronJob(id, new Integer[] { -1 }, new Integer[] { -1 },
				commandToExecute, cronjobType);
	}

	private static void removeCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute, CronjobType cronjobType) {

		ArrayList<Integer> hours = new ArrayList<Integer>(Arrays.asList(hour));
		ArrayList<Integer> days = new ArrayList<Integer>(Arrays.asList(day));

		if ((hours == null) || (days == null))
			return;

		if ((hours.size() == 1) && (hours.contains(-1))) {
			hours.clear();
			for (int i = 0; i < 24; i++) {
				hours.add(i);
			}
		}

		if ((days.size() == 1) && (days.contains(-1))) {
			days.clear();
			for (int i = 0; i < 7; i++) {
				days.add(i);
			}
		}

		CronjobEntry job = new CronjobEntry(id, new HashSet<Integer>(hours),
				new HashSet<Integer>(days), commandToExecute, cronjobType);

		DasuCronScheduler.getInstance().removeCronJob(job);
	}

	public static void addCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute) {
		addCronJob(id, hour, day, commandToExecute, CronjobType.CUSTOM);
	}

	public static void addCronJob(String id, String commandToExecute,
			CronjobType cronjobType) {
		addCronJob(id, new Integer[] { -1 }, new Integer[] { -1 },
				commandToExecute, cronjobType);
	}

	private static void addCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute, CronjobType cronjobType) {

		ArrayList<Integer> hours = new ArrayList<Integer>(Arrays.asList(hour));
		ArrayList<Integer> days = new ArrayList<Integer>(Arrays.asList(day));

		if ((hours == null) || (days == null))
			return;

		if ((hours.size() == 1) && (hours.contains(-1))) {
			hours.clear();
			for (int i = 0; i < 24; i++) {
				hours.add(i);
			}
		}

		if ((days.size() == 1) && (days.contains(-1))) {
			days.clear();
			for (int i = 0; i < 7; i++) {
				days.add(i);
			}
		}

		CronjobEntry job = new CronjobEntry(id, new HashSet<Integer>(hours),
				new HashSet<Integer>(days), commandToExecute, cronjobType);

		DasuCronScheduler.getInstance().registerCronJob(job);
	}

	private static void launchHourlyProbeHistoryUpdater() {

		hourlyHistoryUpdaterTimer = Main
				.createTimer("AMCHourlyProbeHistoryUpdaterTimer");
		hourlyHistoryUpdaterTimer.addPeriodicEvent(60 * 60 * 1000,
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {

						for (TestType tt : hourlyTestTally.keySet())
							hourlyTestTally.put(tt, 0);
					}
				});

	}

	/*
	 * clean shortTerm history if necessary It only iterates over the oldest
	 * entries from the top, if no entries qualify, stop
	 */

	/*
	 * private static void updateProbeHistory() {
	 * 
	 * synchronized(shortTermProbeHistory) {
	 * 
	 * Long currTime = Util.currentTimeMillis(); Iterator<Entry<Pair<ProbeType,
	 * String>, Long>> iter = shortTermProbeHistory.entrySet().iterator(); while
	 * (iter.hasNext()) { Entry<Pair<ProbeType, String>, Long> entry =
	 * iter.next(); ProbeType probetype = entry.getKey().getKey(); if (currTime
	 * - entry.getValue() > shortTermInterval*1000*60) { iter.remove(); if
	 * (shortTermProbeTally.get(probetype) > 0)
	 * shortTermProbeTally.put(probetype, shortTermProbeTally.get(probetype)-1);
	 * } else { break; } } } }
	 */

	private static void updateTestHistory() {

		synchronized (shortTermTestHistory) {

			Long currTime = Util.currentTimeMillis();
			Iterator<Entry<Pair<TestType, String>, Long>> iter = shortTermTestHistory
					.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<Pair<TestType, String>, Long> entry = iter.next();
				TestType testType = entry.getKey().getKey();

				if (Main.DEBUG)
					LOGGER.info("DEBUG: AMCoordinatorV2: " + testType + " "
							+ (currTime - entry.getValue()) / (1000 * 1000));

				if (currTime - entry.getValue() > shortTermInterval * 1000 * 60) {
					iter.remove();
					shortTermTestTally.put(testType,
							shortTermTestTally.get(testType) - 1);
				} else {
					break;
				}
			}
		}
	}

	/*
	 * private static void appendToProbeHistory(ProbeType pt, String ipAddress,
	 * long timeStamp) {
	 * 
	 * // then add to queue synchronized(shortTermProbeHistory) {
	 * shortTermProbeHistory.put(new Pair<ProbeType, String>(pt, ipAddress),
	 * Util.currentTimeMillis()); } shortTermProbeTally.put(pt,
	 * shortTermProbeTally.get(pt)+1); hourlyProbeTally.put(pt,
	 * hourlyProbeTally.get(pt)+1);
	 * 
	 * }
	 * 
	 * private static void appendToTemporaryProbeHistory(ProbeType pt, String
	 * ipAddress, long timeStamp) {
	 * 
	 * synchronized(temporaryShortTermProbeHistory) {
	 * temporaryShortTermProbeHistory.put(new Pair<ProbeType, String>(pt,
	 * ipAddress), Util.currentTimeMillis()); }
	 * 
	 * }
	 * 
	 * private static boolean rollbackFromProbeHistory(ProbeType pt, String
	 * ipAddress) {
	 * 
	 * Long timeStamp = null; synchronized(temporaryShortTermProbeHistory) {
	 * timeStamp = temporaryShortTermProbeHistory.remove(new Pair<ProbeType,
	 * String>(pt, ipAddress));
	 * 
	 * }
	 * 
	 * return (timeStamp==null)?false:true; }
	 * 
	 * private static boolean commitToProbeHistory(ProbeType pt, String
	 * ipAddress) {
	 * 
	 * Long timeStamp = null; synchronized(temporaryShortTermProbeHistory) {
	 * timeStamp = temporaryShortTermProbeHistory.remove(new Pair<ProbeType,
	 * String>(pt, ipAddress)); }
	 * 
	 * if (timeStamp != null) { appendToProbeHistory(pt, ipAddress, timeStamp);
	 * return true; }
	 * 
	 * return false;
	 * 
	 * }
	 */

	private static void appendToTestHistory(TestType tt, String val,
			long timeSpan) {

		/* then add to queue */
		synchronized (shortTermTestHistory) {
			shortTermTestHistory.put(new Pair<TestType, String>(tt, val),
					timeSpan);
		}
		shortTermTestTally.put(tt, shortTermTestTally.get(tt) + 1);
		hourlyTestTally.put(tt, hourlyTestTally.get(tt) + 1);

	}

	private static void appendToTemporaryTestHistory(TestType tt, String val,
			long timeStamp) {

		synchronized (temporaryShortTermTestHistory) {
			temporaryShortTermTestHistory.put(tt, new Pair<String, Long>(val,
					timeStamp));
		}

	}

	private static boolean rollbackFromTestHistory(TestType tt) {

		Pair<String, Long> values = null;
		synchronized (temporaryShortTermTestHistory) {
			values = temporaryShortTermTestHistory.remove(tt);
		}

		return (values == null) ? false : true;
	}

	private static boolean commitToTestHistory(TestType tt) {

		Pair<String, Long> values = null;
		synchronized (temporaryShortTermTestHistory) {
			values = temporaryShortTermTestHistory.remove(tt);
		}

		if (values != null) {
			appendToTestHistory(tt, values.getKey(), values.getValue());
			return true;
		}

		return false;

	}

	private static boolean dispatchedSpeedTest() {

		return ((dispatchedTests.size() != 0) && (dispatchedTests
				.contains(TestType.AZUREUS_DL)
				|| dispatchedTests.contains(TestType.AZUREUS_UL)
				|| dispatchedTests.contains(TestType.AZUREUS_DL_ENCRYPTION) || dispatchedTests
					.contains(TestType.AZUREUS_UL_ENCRYPTION)));

	}

	public synchronized static void stopActiveTorrents() {

		LOGGER.info("NDT: stopping active torents...");
		for (Download d : Main.getPluginInterface().getDownloadManager()
				.getDownloads()) {

			if (d.getState() != Download.ST_STOPPED
					&& d.getState() != Download.ST_STOPPING) {

				try {
					LOGGER.info("NDT: stopping ..." + d.getName());
					d.stop();
					activeTorrents.add(d);
				} catch (DownloadException e) {
					e.printStackTrace();
				}
			}
		}

		if (activeTorrents.size() > 0)
			Main.logPlugin("Temporarily stopping all (" + activeTorrents.size()
					+ ") active torrents");

	}

	public static ArrayList<ArrayList<String>> getMyDnsServers() {

		//System.out.println("\n*******getMyDnsServers(): " + "************");
		try {
			
			DasuProbeModule dpm = probeModules.get(ProbeType.DNS);
			Object o = dpm.getClass().getMethod("getMyDnsServers", new Class<?>[]{})
					.invoke(dpm, new Object[]{});
			return (ArrayList<ArrayList<String>>) o;

		} catch (Exception e) {
			LOGGER.info("DEBUG: Cannot get localhost's DNS servers.\nStack trace: " + e.getStackTrace());
			ArrayList<String> arr = new ArrayList<String>();
			ArrayList<ArrayList<String>> ret = new ArrayList();
			ret.add(arr);
			return ret;
			
		}

	}

	public static String getMyPublicIp() {
		return Main.getPublicIpAddress();
	}

	public synchronized static void restartActiveTorrents() {

		if (Main.DEBUG)
			LOGGER.info("NDT: Restarting all " + activeTorrents.size() + " active torrents");;

		if (activeTorrents.size() > 0)
			Main.logPlugin("Restarting all (" + activeTorrents.size()
					+ ") active torrents");

		Iterator<Download> iter = activeTorrents.iterator();
		while (iter.hasNext()) {
			try {
				Download d = iter.next();
				d.restart();
				iter.remove();
			} catch (DownloadException e) {
				e.printStackTrace();
			}
		}
	}

	/******************
	 * sessionSamples /
	 ******************/

	public static Hashtable<String, SessionSignalHolder> getSessionSample() {
		return sessionSamples;
	}

	public static boolean sessionSampleExists(String name) {
		return sessionSamples.containsKey(name);
	}

	public static boolean sessionSampleCreate(String name, int size) {

		if (!sessionSampleExists(name)) {

			try {
				sessionSamples.put(name, new SessionSignalHolder(size));
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
		}

		// if (Main.DEBUG)
		// sessionSamplePrint(name);

		return true;

	}

	public static boolean sessionSampleDestroy(String name) {

		if (sessionSampleExists(name))
			sessionSamples.remove(name);
		return true;
	}

	public static boolean sessionSampleSetValue(String name, Object value) {

		if (sessionSampleExists(name)) {
			sessionSamples.get(name).set(value);
			return true;
		}
		return false;
	}

	public static void sessionSamplePrint(String name) {

		if (sessionSampleExists(name)) {
			System.out.println(sessionSamples.get(name));
		}

	}

	public static Collection<Object> sessionSampleGetValues(String name) {

		if (sessionSampleExists(name)) {
			return sessionSamples.get(name).getValues();
		}

		return null;
	}

	public static Collection<Long> sessionSampleGetTimeStamps(String name) {

		if (historicSampleExists(name)) {
			return sessionSamples.get(name).getTime();
		}

		return null;
	}

	/*******************
	 * historicSamples
	 *******************/

	public static Hashtable<String, HistoricSignalHolder> getHistoricSample() {
		return historicSamples;
	}

	public static boolean historicSampleExists(String name) {
		return historicSamples.containsKey(name);
	}

	public static boolean historicSampleExists(String name, String signalName) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).exists(signalName);
		}

		return false;
	}

	public static boolean historicSampleCreate(String name,
			TimeType timeSpanType, TimeType intervalType, int intervalLength) {

		if (!historicSampleExists(name)
				&& (historicSamples.size() < DasuConfiguration.getInstance()
						.getMaxLogEntries())) {

			try {
				historicSamples.put(name, new HistoricSignalHolder(name,
						timeSpanType, intervalType, intervalLength));
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
		}

		if (Main.DEBUG)
			historicSamplePrint(name);

		return true;

	}

	public static boolean historicSampleDestroy(String name) {

		if (historicSampleExists(name)) {
			historicSamples.remove(name);
			return true;
		}

		return false;

	}

	public static float historicSampleGetValue(String name, String signalName,
			long timeRecorded) {

		if (historicSampleExists(name))
			return historicSamples.get(name).get(signalName, timeRecorded);

		return -1;
	}

	public static HistoricSignalHolder historicSampleGetSignalHolder(String name) {

		if (historicSampleExists(name))
			return historicSamples.get(name);

		return null;

	}

	public static long historicSampleGetTimeStamp(String name, long timeRecorded) {

		if (historicSampleExists(name))
			return historicSamples.get(name).getTimeStamp(timeRecorded);

		return -1;
	}

	public static long historicSampleGetTimeStamp(String name,
			String signalName, float value) {

		if (historicSampleExists(name)) {

			LinkedHashMap<Integer, Float> vals = historicSamples.get(name)
					.getValuesHash(signalName);
			LinkedHashMap<Integer, Long> times = historicSamples.get(name)
					.getTimeStampsHash();

			Iterator<Entry<Integer, Float>> iter = vals.entrySet().iterator();
			int index = -1;
			while (iter.hasNext()) {
				Entry<Integer, Float> entry = iter.next();
				if (entry.getValue().equals(value))
					index = entry.getKey();
			}

			if (index != -1) {
				return times.get(index);
			}
		}

		return -1;

	}

	public static boolean historicSampleSetValue(String name,
			String signalName, long timeRecorded, float value, String unit) {

		if (historicSampleExists(name)) {
			historicSamples.get(name)
					.set(signalName, timeRecorded, value, unit);
			return true;
		}

		return false;
	}

	public static void historicSamplePrint(String name) {

		if (historicSampleExists(name)) {
			System.out.println(historicSamples.get(name));
		}

	}

	public static Collection<Float> historicSampleGetValues(String name,
			String signalName) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getValuesHash(signalName).values();
		}

		return null;
	}

	public static Collection<Long> historicSampleGetTimeStamps(String name) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getTimeStampsHash().values();
		}

		return null;
	}

	public static Collection<Float> historicSampleGetValuesSubsetPast(
			String name, String signalName, long timeRecorded, int numIntervals) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getValuesSubsetPast(signalName,
					timeRecorded, numIntervals);
		}

		return null;
	}

	public static Collection<Long> historicSampleGetTimeStampsSubsetPast(
			String name, String signalName, long timeRecorded, int numIntervals) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getTimeStampSubsetPast(
					timeRecorded, numIntervals);
		}

		return null;
	}

	public static void commitReportableFact(ReportableFact fpr) {

		Statistics.getInstance().addReportableFact(fpr);
		
	}

	public static void 	commitReportableFact(String type, Map<String, Object> values)
			throws Exception {
		
		ReportableFact rf = new ReportableFact("commitReportableFact");
		rf.getContentsMap().put("type", type);
		for ( String s : values.keySet() ) {
			rf.getContentsMap().put(s, values.get(s));
		}		
		Statistics.getInstance().addReportableFact(rf);

	}

	/******************/

	@Deprecated
	public static void commitStaticUPnPInfo(HashMap<String, String> vals) {

		if ( vals != null) {
			ReportableFact rf = new ReportableFact("staticUPnPInfo");
			for (String key : vals.keySet())
				rf.getContentsMap().put(key, vals.get(key));
			Statistics.getInstance().addReportableFact(rf);
		}

	}

	@Deprecated
	public static void commitInterference(String type, Long timestamp) {

		ReportableFact rf = new ReportableFact("interference");
		rf.getContentsMap().put("type", type);
		rf.getContentsMap().put("timestamp", timestamp);
		Statistics.getInstance().addReportableFact(rf);

		// Statistics.commitInterference(type, timestamp);
	}

	@Deprecated
	public static void commitFeedback(String type, String name, String email,
			String message, String dump) {

		ReportableFact rf = new ReportableFact("feedback");
		rf.getContentsMap().put("type", type);
		rf.getContentsMap().put("name", name);
		rf.getContentsMap().put("email", email);
		rf.getContentsMap().put("message", message);
		rf.getContentsMap().put("dump", dump);
		Statistics.getInstance().addReportableFact(rf);

		// Statistics.commitFeedback(type, name, email, message, dump);
	}

	@Deprecated
	public static void commitUserReports(String dids, String comment) {

		ReportableFact rf = new ReportableFact("userReports");
		rf.getContentsMap().put("dids", dids);
		rf.getContentsMap().put("comment", comment);
		Statistics.getInstance().addReportableFact(rf);

		// Statistics.commitUserComments(dids, comment);
	}

	@Deprecated
	public static void commitIspInfo(String prefix, String ispname) {

		ReportableFact rf = new ReportableFact("ispInfo");
		rf.getContentsMap().put("prefix", prefix);
		rf.getContentsMap().put("ispname", ispname);
		rf.getContentsMap().put("timeRecorded", Util.currentGMTTime());
		Statistics.getInstance().addReportableFact(rf);

		// Statistics.commitIspInfo(prefix, ispname);
	}

	@Deprecated
	public static void commitInterferenceEvent(String type, long timestamp,
			long downloadId, long droppedTo, float mean) {

		ReportableFact rf = new ReportableFact("interferenceEvent");
		rf.getContentsMap().put("type", type);
		rf.getContentsMap().put("timestamp", timestamp);
		rf.getContentsMap().put("downloadId", downloadId);
		rf.getContentsMap().put("mean", mean);
		rf.getContentsMap().put("droppedTo", droppedTo);
		Statistics.getInstance().addReportableFact(rf);

		// Statistics.commitInterferenceEvent(type,
		// (long) timestamp, (long) downloadId, (long) droppedTo, mean);
	}

	public static String getNewRandomDasuUrl() {
		return DasuManager.getInstance().getNewRandomDasuUrl();
	}

	public static float getPluginVersion() {
		return Float.parseFloat(Main.getPluginInterface().getPluginVersion());
	}

	public static boolean isActive() {
		return active;
	}

	public static boolean isPause() {
		return pause;
	}

	public static void setActive(boolean active) {
		AMCoordinatorV2.active = active;
	}

	public static void addPopup(String msg) {
		DasuMessageManager.getInstance().addMessage("NA", msg, -1, false,
				false, "");
	}

	public static void addPopup(String msg, String id, int timeout,
			boolean repeatable) {
		DasuMessageManager.getInstance().addMessage(id, msg, timeout,
				repeatable, false, "");
	}

	@Deprecated
	public static void addPopup(String msg, String id, int timeout,
			boolean repeatable, boolean audible, String details) {
		DasuMessageManager.getInstance().addMessage(id, msg, timeout,
				repeatable, audible, details);
	}

}
