/**
 * Dasu Project
 *
 * File:         DasuProbeModule.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 15, 2008 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;


import java.util.logging.Logger;

import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.dynamic.measurement.DasuDnsModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.EventTypeInterface;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.PluginInterface;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuProbeModule abstract class servers as 
 *         template for all the ProbeModules
 */

public abstract class DasuProbeModule<T> extends Thread {	

	protected final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	private float version;
	protected Boolean active = false;
	protected boolean pause = false;
	protected PluginInterface pi;
	protected int sleepInterval = 30; //seconds 
	protected int pauseInterval = 10; //seconds
	protected volatile int yieldInterval = -1; //milliseconds
	
	protected DasuProbeModule(String name, float version) {
		super(name);
		this.version = version;		
	}
	
	protected void setYieldInterval(int timeInMillis) {
		yieldInterval = timeInMillis;
	}
	
	protected void clearYieldInterval() {
		yieldInterval = -1;
	}
	
	public void setActive(boolean active) {
		this.active = active;
	}
	
	public boolean isActive() {
		return active;
	}
	
	public void setPause(boolean paused) {
		this.pause = paused;
	}
	
	public boolean isPause() {
		return pause;
	}

	public abstract void probeAll(ProbeTask pt);
	public abstract void probeWithPriority(ProbeSubTask pst);
	public abstract void probe(ProbeSubTask pst);
	protected abstract Object probeIp(ProbeSubTask pst);
	
	protected String getIp() {
		return Main.getPublicIpAddress();
	}
	
	public static boolean isValidIp(final String ip) {
        return ip.matches("^[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}$");
    }
	
	public abstract void stopModule();	
	
	public float getVersion() {
		return version;
	}
	
	protected Long generateId() {
		return System.nanoTime();
	}
	
	protected synchronized void updateResultQueue(T fact, EventTypeInterface type, Long id) {
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
					new Pair<Long, EventTypeInterface>(id, type));
			AMCoordinatorV2.events.notifyAll();
		}
		
	}
	
	public abstract Pair<String, FactTemplate> proccessResult(long eventId);
	
	protected void reportProbeResult(FactTemplate fact) {
		
		if (Main.DEBUG)
			LOGGER.info("REPORTER: DasuProbeModule: reportProbeResult: " + fact);
		
		Statistics.getInstance().addReportableFact(fact);
	}	
	
}
