/**
 * Dasu Project
 *
 * File:         Manager.java
 * RCS:          $Id: Manager.java,v 1.0 Mar 4, 2011 12:49:26 PM john Exp $
 * Description:  Manager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Mar 4, 2011 at 12:49:26 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.download
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement.monitors;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Logger;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadListener;
import org.gudy.azureus2.plugins.download.DownloadManager;
import org.gudy.azureus2.plugins.download.DownloadManagerListener;

import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactDidEvent;
import edu.northwestern.dasu.drools.FactMeasurementDsd;
import edu.northwestern.dasu.drools.FactMeasurementDss;
import edu.northwestern.dasu.drools.FactMeasurementPeerRst;
import edu.northwestern.dasu.drools.FactMeasurementRst;
import edu.northwestern.dasu.drools.RuleEngine;
import edu.northwestern.dasu.measurement.AMCoordinatorV2;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.stats.DownloadStatsDynamic;
import edu.northwestern.dasu.stats.DownloadStatsSummary;
import edu.northwestern.dasu.stats.RSTStat;
import edu.northwestern.dasu.util.PluginInterface;
import edu.northwestern.dasu.util.Util;

/**
 * @author JSO
 *
 * The Manager class ...
 * listen for changes to downloads. create/destroy DownloadObserver objects as needed.
 */
public class DownloadMonitor implements DownloadListener, DownloadManagerListener {
	
	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	
	private PluginInterface pi;
	private DownloadManager dm;
	private RuleEngine engine;
	
	private Map<String, Integer> dlmToDidMap = new Hashtable<String, Integer>();
	
	private HashSet<Download> dldListeners = new HashSet<Download>();
	private Hashtable<String, Long> dldRecentState = new Hashtable<String, Long>();
	
	private Map<Long, FactMeasurementDsd> didFactMap = new Hashtable<Long, FactMeasurementDsd>();
	private Map<Long, FactMeasurementDss> didDssFactMap = new Hashtable<Long, FactMeasurementDss>();
	private Map<Long, FactMeasurementRst> rstFactMap = new Hashtable<Long, FactMeasurementRst>();
	
	public DownloadMonitor(RuleEngine engine) {
		this.engine = engine;
		pi = Main.getPluginInterface();
		dm = pi.getDownloadManager();
		Download[] downloads = null;
		
		//if simulator is enabled don't add listeners to real BT downloads
		if (!DasuManager.USE_SIMULATION_V2) {
			dm.addListener(this);
			downloads = dm.getDownloads(); 
		}
		else {
			downloads = DasuManager.interferenceSimulator.iterGetDownloads();
		}
		
		for (Download download : downloads) {
			downloadAdded(download);
		}
		
	}
	
	public void remove() {
		for (Download download : dm.getDownloads()) {
			downloadRemoved(download);
		}
		
		dm.removeListener(this);
	}
	
	public void updateFacts() {
		
		Download[] downloads = (!DasuManager.USE_SIMULATION_V2 ? dm.getDownloads() : DasuManager.interferenceSimulator.iterGetDownloads());
		
		for (Download d: downloads) {
			
			long did = getId(d);
			
			//refreshRSTCounter(); /*why again?*/
			
			int rst = 0;
			if (rstFactMap != null) {
				rst = (rstFactMap.containsKey(did)) ? rstFactMap.get(did).getNumRst() : 0;
			}

			synchronized(didFactMap) {

				if (!didFactMap.containsKey(did)) {
					didFactMap.put(did, new FactMeasurementDsd(new DownloadStatsDynamic(d, rst)));
				} else {
					didFactMap.get(did).addEntry(new DownloadStatsDynamic(d, rst));
				}

				engine.addFact(didFactMap.get(did));
			}
			
			synchronized(didDssFactMap) {

				if (!didDssFactMap.containsKey(did)) {
					didDssFactMap.put(did, new FactMeasurementDss(new DownloadStatsSummary(d)));
				} else {
					didDssFactMap.get(did).addEntry(new DownloadStatsSummary(d));
				}

				engine.addFact(didDssFactMap.get(did));
			}
		}
		
		synchronized(rstFactMap) {
			for(FactMeasurementRst fmr: rstFactMap.values()) {
				fmr.updateEntryList();
			}				
		}
	}
	
	public void updateRstFacts(long did, RSTStat rstStat) {
		synchronized(rstFactMap) {
			if (!rstFactMap.containsKey(did)) {
				rstFactMap.put(did, new FactMeasurementRst(rstStat));
			} else {
				rstFactMap.get(did).addEntry(rstStat);
			}
			
			if (engine != null) {
				engine.addFact(rstFactMap.get(did));
				engine.addFact(new FactMeasurementPeerRst(rstStat));
			}
		}
	}
	
	/* 
	 * NOP
	 * (non-Javadoc)
	 * @see org.gudy.azureus2.plugins.download.DownloadListener#positionChanged(org.gudy.azureus2.plugins.download.Download, int, int)
	 */
	public void positionChanged(Download arg0, int arg1, int arg2) {}

	/* (non-Javadoc)
	 * @see org.gudy.azureus2.plugins.download.DownloadListener#stateChanged(org.gudy.azureus2.plugins.download.Download, int, int)
	 */
	public void stateChanged(Download download, int old_state, int new_state) {
		if ( new_state == Download.ST_STOPPED ||
				new_state == Download.ST_STOPPING ||
				new_state == Download.ST_DOWNLOADING || 
				(new_state == Download.ST_SEEDING && old_state != Download.ST_DOWNLOADING)) {
		
			LOGGER.info("state changed for download: " + download.getName());

			dldRecentState.put(download.getName(), Util.currentTimeMillis());
			
			long did = getId(download);
			
			AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent("stateChange", (long) did));
			
		}
	}

	/* (non-Javadoc)
	 * @see org.gudy.azureus2.plugins.download.DownloadManagerListener#downloadAdded(org.gudy.azureus2.plugins.download.Download)
	 */
	public void downloadAdded(Download download) {
		synchronized(dldListeners) {

			if(dldListeners.contains(download)) 
				return;

			if ( !DasuManager.USE_SIMULATION_V2 )
				dldListeners.add(download);
		}
		
		LOGGER.info("Download added!! " + download.getName());
		
		download.addListener(this);
					
		long did = getId(download);
		
		FactMeasurementDsd fmdd = new FactMeasurementDsd(new DownloadStatsDynamic(download, 0));
		synchronized(didFactMap) {		
			didFactMap.put(did, fmdd);
		}
		engine.addFact(fmdd);
		
		
		FactMeasurementDss fmds = new FactMeasurementDss(new DownloadStatsSummary(download));
		synchronized(didDssFactMap) {
			didDssFactMap.put(did, fmds);
		}
		engine.addFact(fmds);
		
		synchronized(dldRecentState) {
			dldRecentState.put(download.getName(), Util.currentTimeMillis());
		}
	}

	/* (non-Javadoc)
	 * @see org.gudy.azureus2.plugins.download.DownloadManagerListener#downloadRemoved(org.gudy.azureus2.plugins.download.Download)
	 */
	public void downloadRemoved(Download download) {
		synchronized(dldListeners) {
			if(!dldListeners.contains(download)) return;
			dldListeners.remove(download);
		}

		LOGGER.info("Download removed!! " + download.getName());
		
		download.removeListener(this);
		
		long did = getId(download);
		synchronized(dlmToDidMap) {
			dlmToDidMap.remove(download.getName());
		}
		
		synchronized(didFactMap) {
			if (didFactMap.containsKey(did))
				engine.retractFact(didFactMap.remove(did));
		}
		
		synchronized(didDssFactMap) {
			if (didDssFactMap.containsKey(did))
				engine.retractFact(didDssFactMap.remove(did));
		}

		synchronized(rstFactMap) {
			if (rstFactMap.containsKey(did))
				engine.retractFact(rstFactMap.remove(did));
		}
	}
	
	public int getId(Download d) {
		
		int id = Main.getId(d);
		dlmToDidMap.put(d.getName(), id);
		return id;

		/*
		MessageDigest md;
		try {
			md = MessageDigest.getInstance("SHA");

			md.update(d.getName().getBytes());
			md.update(Util.convertLong(d.getCreationTime()));
			byte hash[] = md.digest();
			String s = new String(hash);
			dlmToDidMap.put(d.getName(), (int) s.hashCode());
			return s.hashCode();

		} catch (NoSuchAlgorithmException e) {}

		// fallback
		dlmToDidMap.put(d.getName(), (int) (d.getName().hashCode() + d.getCreationTime()));
		return (int) (d.getName().hashCode() + d.getCreationTime());
		*/

	}
	
	public boolean hasId(Download d) {
		synchronized(dlmToDidMap) {
			return dlmToDidMap.containsKey(d.getName());
		}
	}
	
	public boolean hasId(String str) {
		synchronized(dlmToDidMap) {
			return dlmToDidMap.containsKey(str);
		}
	}
	
	public int getId(String str) {
		synchronized(dlmToDidMap) {
			if(!dlmToDidMap.containsKey(str)) {
				return 0;
			}
			return dlmToDidMap.get(str);
		}
	}
	
	public Long getLastStateChange(Download d) {
		return getLastStateChange(d.getName());
	}
	
	public Long getLastStateChange(String str) {
		synchronized(dldRecentState) {
			if(!dldRecentState.containsKey(str)) return null;
			return dldRecentState.get(str);
		}
	}
	
	public void printDebug() {
		synchronized(didFactMap) {
			
			StringBuilder sb = new StringBuilder();
			
			for(Long did: didFactMap.keySet()) {				
				
				sb.append("Did: " + did + " uploadRate: "
								+ didFactMap.get(did).getUploadRate()
								+ " avgUpload: "
								+ didFactMap.get(did)
										.getAverageUpload()
								+ " stdevUpload: "
								+ didFactMap.get(did).getStdevUpload()
								+ " downloadRate: "
								+ didFactMap.get(did).getDownloadRate()
								+ " avgDownload: "
								+ didFactMap.get(did)
										.getAverageDownload()
								+ " stdevDownload: "
								+ didFactMap.get(did)
										.getStdevDownload() + " did: "
								+ didFactMap.get(did).getDownloadId()
								+ " completed: "
								+ didFactMap.get(did).getCompleted()
								+ " numRst: "
								+ didFactMap.get(did).getNumRst() + "\n");
				   
				sb.append("Did: " + did + " numPeers: " 
						   + didFactMap.get(did).getNumPeers()
						   + " avgNumPeers: " 
						   + didFactMap.get(did).getAverageNumPeers()
						   + " connectedSeeds: "
						   + didFactMap.get(did).getConnectedSeeds()
						   + " avgConnectedSeeds: "
						   + didFactMap.get(did).getAverageConnectedSeeds()
						   + " connectedLeechers: "
						   + didFactMap.get(did).getConnectedLeechers()
						   + " avgConnectedLeechers: "
						   + didFactMap.get(did).getAverageConnectedLeechers()									   
						   + " numPendingPeers: " 
						   + didFactMap.get(did).getNumPendingPeers() + "\n");
				   
			}
		
			LOGGER.info("DEBUG: " + sb.toString());
		} //synchronized
			
	}		
}
