/**
 * Dasu Project
 *
 * File:         Scheduler.java
 * RCS:          $Id: Scheduler.java,v 1.0 Apr 17, 2012 12:22:52 PM mas939 Exp $
 * Description:  Scheduler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 17, 2012 at 12:22:52 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2012, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement.scheduler;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;

import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.util.Pair;

/**
 * @author John Otto &lt;jotto@eecs.northwestern.edu&gt;
 *
 * The Scheduler class ...
 */

public class Scheduler {
	// JSO's attempt at multi-queue scheduler.
	//
	// Individual queues can block based on quotas.
	//
	// Usage is as follows:
	// 1) jobs are submitted to the scheduler with a named priority level
	// 2) an individual queue can deny admission to the queue based on
	// quotas (LimitedQuotaQueue)
	// 3) an external thread (e.g. AMC) queries the scheduler for the next
	// job to run

	private Hashtable<String, Queue> queues = new Hashtable<String, Queue>();
	private ArrayList<String> queueOrder = new ArrayList<String>();

	private boolean useQueueOrderPriority = false; // if true, return the
													// first schedulable
													// task. else, take the
													// next schedulable one
													// in chronological
													// order
	public Scheduler() {
	}

	public Scheduler(boolean useQueueOrderPriority) {
		this.useQueueOrderPriority = useQueueOrderPriority;
	}

	public boolean addQueue(String name, Queue q) {
		if (queues.containsKey(name))
			return false;
		queueOrder.add(name);
		queues.put(name, q);
		return true;
	}

	public ArrayList<String> getQueues() {
		return new ArrayList<String>(queueOrder);
	}

	public boolean push(String qname, ProbeTask pt) {
		
		if (!queues.containsKey(qname))
			return false;
		Queue q = queues.get(qname);
		return q.push(pt);
		
	}

	public boolean admittable(String qname) {
		if (!queues.containsKey(qname))
			return false;
		Queue q = queues.get(qname);
		return q.admittable();
	}

	public ProbeSubTask peek() {
		Pair<Long, ProbeSubTask> p = null;
		for (String qname : queueOrder) {
			Queue q = queues.get(qname);
			Pair<Long, ProbeSubTask> thisP = q.peek();
			if (thisP != null) {
				if (useQueueOrderPriority) {
					return p.getValue();
				}
				if (p == null || thisP.getKey() < p.getKey()) {
					p = thisP;
				}
			}
		}
		if (p == null)
			return null;
		return p.getValue();
	}

	public ProbeSubTask poll() {
		Pair<Long, ProbeSubTask> p = null;
		String useQName = null;
		for (String qname : queueOrder) {
			Queue q = queues.get(qname);
			Pair<Long, ProbeSubTask> thisP = q.peek();
			if (thisP != null && (p == null || thisP.getKey() < p.getKey())) {
				useQName = qname;
				p = thisP;
				if (useQueueOrderPriority) {
					break;
				}
			}
		}
		if (useQName == null)
			return null;
		
		//return queues.get(useQName).poll().getValue();
		
		p = queues.get(useQName).poll();
		if ( p != null )
			return p.getValue();
		else
			return null;
		
	}

	public int size() {
		int sum = 0;
		for (String qname : queueOrder) {
			sum += queues.get(qname).size();
		}
		return sum;
	}
	
	public void pollAndPushToBack() {
		Pair<Long, ProbeSubTask> p = null;
		String useQName = null;
		for (String qname : queueOrder) {
			Queue q = queues.get(qname);
			Pair<Long, ProbeSubTask> thisP = q.peek();
			if (thisP != null && (p == null || thisP.getKey() < p.getKey())) {
				useQName = qname;
				p = thisP;
				if (useQueueOrderPriority) {
					break;
				}
			}
		}
		if (useQName != null) {
			queues.get(useQName).pollAndPushToBack();
		}
	}
	
	public void pollAndPushToSecond() {
		Pair<Long, ProbeSubTask> p = null;
		String useQName = null;
		for (String qname : queueOrder) {
			Queue q = queues.get(qname);
			Pair<Long, ProbeSubTask> thisP = q.peek();
			if (thisP != null && (p == null || thisP.getKey() < p.getKey())) {
				useQName = qname;
				p = thisP;
				if (useQueueOrderPriority) {
					break;
				}
			}
		}
		if (useQName != null) {
			queues.get(useQName).pollAndPushToSecond();
		}
	}

}