/**
 * Dasu Project
 *
 * File:         Profiler.java
 * RCS:          $Id: Profiler.java,v 1.0 Apr 18, 2011 5:07:03 PM john Exp $
 * Description:  Profiler class (see below)
 * Author:       John Otto
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 18, 2011 at 5:07:03 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.util
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.profilers;

import java.util.*;
import java.util.Map.Entry;

import org.drools.definition.rule.Rule;
import org.drools.event.rule.ActivationCancelledEvent;
import org.drools.event.rule.ActivationCreatedEvent;
import org.drools.event.rule.AfterActivationFiredEvent;
import org.drools.event.rule.AgendaEventListener;
import org.drools.event.rule.AgendaGroupPoppedEvent;
import org.drools.event.rule.AgendaGroupPushedEvent;
import org.drools.event.rule.BeforeActivationFiredEvent;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.rule.Activation;

import edu.northwestern.dasu.DasuData;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.util.ProbeEventListener;

/**
 * @author John Otto &lt;jotto@eecs.northwestern.edu&gt;
 *
 * The Profiler class ... for now, just to count rule activations and runtime distributions.
 */
public class Profiler implements AgendaEventListener, ProbeEventListener {
	
	private final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
	public class Distribution {
		public boolean log = true;
		public double resolution = 0.01;
		
		public int total = 0;
		public double sum = 0;
		public double max = 0;
		public HashMap<Integer, Integer> bins = new HashMap<Integer, Integer>();
		public int errors = 0;
		
		public void insert(double val) {
			if(val > max) max = val;
			if(val == 0.0) { val = 0.0001; }
			try {
				double initVal = val;
				if(log) val = Math.log10(val);
				val /= resolution;
				int res = new Double(Math.floor(val)).intValue();
				if(!bins.containsKey(res)) {
					bins.put(res, 0);
				}
				bins.put(res, bins.get(res) + 1);
				total += 1;
				sum += initVal;
			} catch(Exception e) {
				errors += 1;
			}
		}
		
		public ArrayList<ArrayList<Double>> cdf() {
			ArrayList<ArrayList<Double>> result = new ArrayList<ArrayList<Double>>();
			
			ArrayList<Integer> xs = new ArrayList<Integer>(bins.keySet());
			Collections.sort(xs);
			
			int soFar = 0;
			for(int bin: xs) {
				double x = new Integer(bin).doubleValue();
				x *= resolution;
				if(log) x = Math.pow(10, x);
				
				if(soFar == 0) {
					// add an initial point
					ArrayList<Double> point = new ArrayList<Double>();
					point.add(0.99 * x);
					point.add(0D);
					
					result.add(point);
				}
				
				soFar += bins.get(bin);
				double frac = ((double) soFar) / total * 100;
				
				ArrayList<Double> point = new ArrayList<Double>();
				point.add(x);
				point.add(frac);
				
				result.add(point);
			}
			
			return result;
		}
		
		public ArrayList<ArrayList<Double>> ccdf() {
			ArrayList<ArrayList<Double>> result = new ArrayList<ArrayList<Double>>();
			
			ArrayList<Integer> xs = new ArrayList<Integer>(bins.keySet());
			Collections.sort(xs);
			
			int soFar = total;
			for(int bin: xs) {
				double x = new Integer(bin).doubleValue();
				x *= resolution;
				if(log) x = Math.pow(10, x);
				
				if(soFar == total) {
					// add an initial point
					ArrayList<Double> point = new ArrayList<Double>();
					point.add(0.99 * x);
					point.add(100D);
					
					result.add(point);
				}
				
				soFar -= bins.get(bin);
				double frac = ((double) soFar) / total * 100;
				
				ArrayList<Double> point = new ArrayList<Double>();
				point.add(x);
				point.add(frac);
				
				result.add(point);
			}
			
			return result;
		}
		
		public double avg() {
			if(total == 0) return 0;
			return sum / total;
		}
		
		public double max() {
			return max;
		}
	}
	
	public class RuleProfile {
		public Rule rule;
		
		public int total = 0;
	
		public ArrayList<Long> hourly = new ArrayList<Long>();
		public ArrayList<Long> minutely = new ArrayList<Long>();
		
		public HashMap<Long, Long> active = new HashMap<Long, Long>();
		
		public Distribution durations = new Distribution();
		
		public RuleProfile(Rule rule) {
			this.rule = rule;
		}
		
		public void update() {
			Profiler.update(hourly, 1000*60*60);
			Profiler.update(minutely, 1000*60);
		}
		
		public void logActivate(long id) {
			long now = System.currentTimeMillis();
			active.put(id, now);
			total += 1;
			hourly.add(now);
			minutely.add(now);
		}
		
		public void logFinished(long id) {
			long now = System.currentTimeMillis();
			if(!active.containsKey(id)) return;
			long start = active.remove(id);
			long latency = now - start;
			durations.insert(((double)latency) / 1000); // seconds
		}
	}

	private static void update(ArrayList<Long> times, long limit) {
		long now = System.currentTimeMillis();
		Collections.sort(times);
		while(times.size() > 0 && now - times.get(0) > limit) {
			times.remove(0);
		}
	}
	
	public class ProbeProfile {
		
		public class Pair {
			long pt = 0;
			long pst = 0;
			public Pair(long pt, long pst) {
				this.pt = pt;
				this.pst = pst;
			}
			
			@Override
			public boolean equals(Object o) {
				if(!(o instanceof Pair)) return false;
				Pair p = (Pair) o;
				return p.pt == pt && p.pst == pst;
			}
			
			@Override
			public int hashCode() {
				return (int) (pt * pst);
			}
			
			@Override
			public String toString() {
				
				return "<" + pt + "," + pst + ">";
			}
		}
		
		public int queued = 0;
		public int scheduled = 0;
		public int finished = 0;
		
		public int denied = 0;
	
		// number of probe sub tasks that are queued recently
		public ArrayList<Long> qHourly = new ArrayList<Long>();
		public ArrayList<Long> qMinutely = new ArrayList<Long>(); 
		
		// number of probe sub tasks that are scheduled recently
		public ArrayList<Long> sHourly = new ArrayList<Long>();
		public ArrayList<Long> sMinutely = new ArrayList<Long>(); 
		
		// number of probe sub tasks that are finished recently
		public ArrayList<Long> fHourly = new ArrayList<Long>();
		public ArrayList<Long> fMinutely = new ArrayList<Long>(); 
		
		// number of probetasks that are denied recently
		public ArrayList<Long> dHourly = new ArrayList<Long>();
		public ArrayList<Long> dMinutely = new ArrayList<Long>(); 
		
		public HashMap<Long, Long> active = new HashMap<Long, Long>(); // probetask id -> queue time
		public HashMap<Long, Integer> remaining = new HashMap<Long, Integer>(); // probetask id -> number of subtasks left to complete
		public HashMap<Pair, Long> activeSub = new HashMap<Pair, Long>(); // (probetask id, probesubtask id) -> pending time
		
		public Distribution totalDurations = new Distribution();
		public Distribution queueDurations = new Distribution();
		public Distribution runDurations = new Distribution();
		
		public void update() {
			Profiler.update(qHourly, 1000*60*60);
			Profiler.update(qMinutely, 1000*60);
			Profiler.update(sHourly, 1000*60*60);
			Profiler.update(sMinutely, 1000*60);
			Profiler.update(fHourly, 1000*60*60);
			Profiler.update(fMinutely, 1000*60);
			Profiler.update(dHourly, 1000*60*60);
			Profiler.update(dMinutely, 1000*60);
		}
		
		public void logQueued(ProbeTask pt) {
			// log the submission of a probetask. this can contain multiple probesubtasks, which will be tracked individually.
			long now = System.currentTimeMillis();
			active.put(pt.id, now);
			remaining.put(pt.id, pt.ipsToProbe.size());
			queued += pt.ipsToProbe.size();
			for(int i = 0; i < pt.ipsToProbe.size(); i++) {
				qHourly.add(now);
				qMinutely.add(now);
			}
		}
		
		public void logScheduled(long pt, long pst) {
			
			if ( activeSub.containsKey(new Pair(pt, pst)))
				if ( Main.DEBUG)
					LOGGER.info("PROFILER: *****************\n\n\n\nALREADY IN THE LOG OVERRIDEINGG\n\n\n\n*******************!!! " + pt + " " + pst);

			
			// a probesubtask has been submitted to the module -- it is being run.
			activeSub.put(new Pair(pt, pst), System.currentTimeMillis());
			scheduled += 1;
			
			long now = System.currentTimeMillis();
			sHourly.add(now);
			sMinutely.add(now);
		}
		
		public void logFinished(long pt, long pst) {
			// a probesubtask has completed
			
			long now = System.currentTimeMillis();
			long queueTime = active.get(pt);
			long pendingTime = activeSub.remove(new Pair(pt, pst));
			totalDurations.insert(((double) (now - queueTime)) / 1000);
			runDurations.insert(((double) (now - pendingTime)) / 1000);
			queueDurations.insert(((double) (pendingTime - queueTime)) / 1000);
			
			int count = remaining.remove(pt) - 1;
			if(count == 0) {
				active.remove(pt);
			} else {
				remaining.put(pt, count);
			}
			
			finished += 1;
			fHourly.add(now);
			fMinutely.add(now);
		}
		
		public void logDenied() {
			// a probetask queue request was denied due to a time-based quota being full
			
			long now = System.currentTimeMillis();
			denied += 1;
			dHourly.add(now);
			dMinutely.add(now);
		}
	}
	
	public HashMap<Rule, RuleProfile> rules = new HashMap<Rule, RuleProfile>();
	public HashMap<String, ProbeProfile> probes = new HashMap<String, ProbeProfile>();
	
	private boolean run = true;
	private Thread updater;
	private long interval = 15*1000;
	
	public Profiler() {
		
		updater = new Thread() {
			public void run() {
				try {
					while(run) {
						synchronized(rules) {
							HashMap data = new HashMap();
							for(Entry<Rule, RuleProfile> e: rules.entrySet()) {
								e.getValue().update();
								Rule r = e.getKey();
								RuleProfile p = e.getValue();
								String key = r.getPackageName() + "/" + r.getName();
								ArrayList vals = new ArrayList();
								vals.add(p.total);
								vals.add(p.hourly.size());
								vals.add(p.minutely.size());
								vals.add(p.durations.ccdf());
								vals.add(p.durations.avg());
								vals.add(p.durations.max());
								data.put(key, vals);
							}
							
							// push to dasudata
							DasuData.getInstance().update("profiler.rules", data);
							
							data = new HashMap();
							for(Entry<String, ProbeProfile> e: probes.entrySet()) {
								e.getValue().update();
								ProbeProfile p = e.getValue();
								String key = e.getKey();
								ArrayList vals = new ArrayList();
								vals.add(p.activeSub.size());
								vals.add(p.queued);
								vals.add(p.qHourly.size());
								vals.add(p.qMinutely.size());
								vals.add(p.scheduled);
								vals.add(p.sHourly.size());
								vals.add(p.sMinutely.size());
								vals.add(p.finished);
								vals.add(p.fHourly.size());
								vals.add(p.fMinutely.size());
								vals.add(p.denied);
								vals.add(p.dHourly.size());
								vals.add(p.dMinutely.size());
								vals.add(p.totalDurations.ccdf()); //13
								vals.add(p.queueDurations.ccdf()); //14
								vals.add(p.runDurations.ccdf()); //15
								vals.add(p.queueDurations.avg());
								vals.add(p.queueDurations.max());
								vals.add(p.runDurations.avg());
								vals.add(p.runDurations.max());
								vals.add(p.totalDurations.avg());
								vals.add(p.totalDurations.max());
								data.put(key, vals);
							}
							
							// push to dasudata
							DasuData.getInstance().update("profiler.probes", data);
							
						}
						Thread.sleep(interval);
					}
				} catch (InterruptedException e) {
					//e.printStackTrace();
				}
			}
		};
		updater.setName("Dasu::Profiler");
		updater.start();
	}
	
	public void stop() {
		
		LOGGER.info("STOP: stopping Profiler...");
		run = false;
		if(updater.isAlive()) 
			updater.interrupt();
	}
	
	public void addEngine(StatefulKnowledgeSession ksession) {
		ksession.addEventListener((AgendaEventListener) this);
	}

	public void beforeActivationFired(BeforeActivationFiredEvent arg0) {
		Activation a = arg0.getActivation();
		long id = a.getPropagationContext().getPropagationNumber();
		Rule r = a.getRule();
		synchronized(rules) {
			if(!rules.containsKey(r)) {
				rules.put(r, new RuleProfile(r));
			}
			rules.get(r).logActivate(id);
		}
	}

	public void afterActivationFired(AfterActivationFiredEvent arg0) {
		Activation a = arg0.getActivation();
		long id = a.getPropagationContext().getPropagationNumber();
		Rule r = a.getRule();
		synchronized(rules) {
			if(!rules.containsKey(r)) {
				rules.put(r, new RuleProfile(r)); // shouldn't happen...just in case.
			}
			rules.get(r).logFinished(id);
		}
	}

	public void activationCancelled(ActivationCancelledEvent arg0) {}
	public void activationCreated(ActivationCreatedEvent arg0) {}
	public void agendaGroupPopped(AgendaGroupPoppedEvent arg0) {}
	public void agendaGroupPushed(AgendaGroupPushedEvent arg0) {}

	// Probe callbacks
	public void probeQueued(ProbeTask pt) {
		for(String pp: getProbeProfiles(pt)) {
			if(!probes.containsKey(pp)) {
				probes.put(pp, new ProbeProfile());
			}
			
			probes.get(pp).logQueued(pt);
		}
	}

	public void probeScheduled(ProbeSubTask pst) {
		
		for(String pp: getProbeProfiles(pst.parent)) {
			if(!probes.containsKey(pp)) {
				probes.put(pp, new ProbeProfile());
			}
			
			probes.get(pp).logScheduled(pst.parent.id, pst.id);
		}
	}

	public void probeCompleted(ProbeSubTask pst) {
		for(String pp: getProbeProfiles(pst.parent)) {
			if(!probes.containsKey(pp)) {
				probes.put(pp, new ProbeProfile());
			}
			
			probes.get(pp).logFinished(pst.parent.id, pst.id);
		}
	}
	
	public void probeDenied(ProbeTask pt) {
		for(String pp: getProbeProfiles(pt)) {
			if(!probes.containsKey(pp)) {
				probes.put(pp, new ProbeProfile());
			}
			
			probes.get(pp).logDenied();
		}
	}

	private ArrayList<String> getProbeProfiles(ProbeTask pt) {
		ArrayList<String> pps = new ArrayList<String>();
		pps.add("ALL " + pt.probeType.toString());
		pps.add(pt.sourceType.toString() + " " + pt.probeType.toString());
		return pps;
	}
}
