/**
 * Dasu Project
 *
 * File:         ResourceProfiler.java
 * RCS:          $Id: ResourceProfiler.java,v 1.0 Nov 16, 2011 10:59:51 AM mas939 Exp $
 * Description:  ResourceProfiler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Nov 16, 2011 at 10:59:51 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.profilers
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.profilers;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.logging.Logger;

import edu.northwestern.dasu.DasuData;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.ui.LinePlot;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The ResourceProfiler class ... for now, just to compute cpu usage on per-thread basis.
 */

public class ResourceProfiler implements Runnable {
	
	int MAX_NUM_TIME_ENTIES_PER_THEAD = 60*2;
	private static ResourceProfiler self = null;
	private boolean active = false;
	private static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	
	
	private Hashtable<Long, ThreadCpuInfo> threadIdsMap = new Hashtable<Long, ThreadCpuInfo>();
	
	private LinkedList<Float> vuzeCpuUsage = new LinkedList<Float>();
	private LinkedList<Float> dasuCpuUsage = new LinkedList<Float>();
	private LinkedList<Long> timeAxis = new LinkedList<Long>();
	
	private ThreadMXBean threadBean = ManagementFactory.getThreadMXBean();
	private final long profileInterval = 1000;	
	private long prevUpdate = System.currentTimeMillis();
	private long prevUIUpdate = System.currentTimeMillis();
	
	private long startTime = System.nanoTime();	
	private long prevUpTime = startTime;
	private long upTime = 0;
	
	private static boolean DEBUG = false;	
	
	//DEBUG
	private HashSet<Long> toPlot = new HashSet<Long>();

    public static ResourceProfiler getInstance() {
    	if (self == null)
    		self = new ResourceProfiler();
	    return self;
    }
    
	public void run() {
		
		if (!threadBean.isThreadCpuTimeSupported()) {
			
			if ( DEBUG ) {
				LOGGER.severe("PROFILER: ThreadCpuTime is not supportedd!!");
				ReportableFact rf = new ReportableFact("errorEvent");	
				rf.getContentsMap().put("event", "ThreadCpuTime not supported!");
				Statistics.getInstance().addReportableFact(rf);

			}
			return;
		}
			
		active = true;	
		
		try {
	    
			while (!Main.isShuttingDown() && isActive()) {
			  								
				Long time = System.currentTimeMillis();
				try {
					updateCurrentCpuUsage(time);
				}
				catch (Exception e) {
					e.printStackTrace();
				}
				
				if ( DEBUG ) {
					if ( System.currentTimeMillis() - prevUpdate > 5*1000)
						debug(time, 5f);
				}
				
				if ( System.currentTimeMillis() - prevUpdate > 5*1000) {
					cleanList();
					prevUpdate = System.currentTimeMillis();
				}
				
				
				Thread.sleep(profileInterval);
				
			}
		  
		} catch (InterruptedException iex) {
			Thread.currentThread().interrupted();
			throw new RuntimeException("Profiler interrupted");
		}
		
	}
	
	private void cleanList() {
		
		HashSet<Long> activeThreads = new HashSet<Long>();		
		for ( long t : threadBean.getAllThreadIds())
			activeThreads.add(t);
		
		threadIdsMap.keySet().retainAll(activeThreads);
		/*
		Iterator<Entry<Long, ThreadCpuInfo>> iter = threadIdsMap.entrySet().iterator();
		while ( iter.hasNext() ) {

			Entry<Long, ThreadCpuInfo> e = iter.next();
			
			if (!activeThreads.contains(e.getKey()))
				iter.remove();
			
		}
		*/
		
	}
	
	private LinkedList<Float> fillPlotGap(ThreadCpuInfo tInfo, LinkedList<Long> timeAxis) {
		
		LinkedList<Long> timeStamp = tInfo.getTimeStamp();
		LinkedList<Float> vals = tInfo.getCpuUsageList();
		LinkedList<Float> ret = new LinkedList<Float>();
		
		for ( Long l : timeAxis ) {
			
			int index = timeStamp.indexOf(l);
			if ( index != -1 ) {
				ret.add(vals.get(index));
			}
			else
				ret.add(0f);			
		}
		
		return ret;
		
	}
	
	//update the per-thread cpu usage (for all threads)
	private void updateCurrentCpuUsage(long time) {
			
		upTime = getUptime();		
		updateCpuTime(time);

    	long elapsedCpuTimeTotal = 0;
    	long elapsedCpuTimeDasu = 0;
        long elapsedTime = upTime - prevUpTime;

        if (prevUpTime > 0L && upTime > prevUpTime) {
        	
        	for ( Long threadId : threadIdsMap.keySet() ) {
        		
        		elapsedCpuTimeTotal += threadIdsMap.get(threadId).getElapsedCpuTime(time);
        		if ( threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("dasu") ||
        				threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("reaper"))
                	elapsedCpuTimeDasu += threadIdsMap.get(threadId).getElapsedCpuTime(time);
        		
        	} 	
        }
        
        if ( vuzeCpuUsage.size() > MAX_NUM_TIME_ENTIES_PER_THEAD) {
        	vuzeCpuUsage.pop();
        	dasuCpuUsage.pop();
        	timeAxis.pop();
        }
        
        float totalCpuUsage = ((float)elapsedCpuTimeTotal / ((float)elapsedTime))*100;
        vuzeCpuUsage.add(totalCpuUsage);
        
        float dasuTotalCpuUsage = ((float)elapsedCpuTimeDasu / ((float)elapsedTime))*100;
        dasuCpuUsage.add(dasuTotalCpuUsage);
        
        prevUpTime = upTime;
	}
	
	private HashMap<Long, Thread> getAllThreads() {
		
		HashMap<Long, Thread> ret = new HashMap<Long, Thread>();
		ThreadGroup rootGroup = Thread.currentThread( ).getThreadGroup( );
		ThreadGroup parentGroup;
		while ( ( parentGroup = rootGroup.getParent() ) != null ) {
		    rootGroup = parentGroup;
		}
		
		Thread[] threads = new Thread[ rootGroup.activeCount() ];
		while ( rootGroup.enumerate( threads, true ) == threads.length ) {
		    threads = new Thread[ threads.length * 2 ];
		}
		
		for ( Thread t : threads) {
			
			if ( t != null)
				ret.put(t.getId(), t);
		}
		
		return ret;		
	}
	
	//update the per-thread cpu timing (for all threads)
	private void updateCpuTime(Long time) {
				
		timeAxis.add(time);		
		HashMap<Long, Thread> threads = getAllThreads();
		
		for ( long threadId : threads.keySet() ) {
            
			if (threadId != -1) {
            	    
            	if (!threadIdsMap.containsKey(threadId)) 
            		threadIdsMap.put(threadId, new ThreadCpuInfo(threadId, threads.get(threadId)));
                
            	threadIdsMap.get(threadId).addCpuTime(threadBean.getThreadCpuTime(threadId), time);
            }
        }
        
	}
	
	public long getUptime() {
		return System.nanoTime() - startTime;		
	}	
	
	public void stop() {
		
		LOGGER.info("STOP: stopping ResourceProfiler...");
		active = false;
	}
	
	public boolean isActive() {
		return active;
	}

	
    public void debug(Long time, float pct) {
    	
    	StringBuilder sb = new StringBuilder();
    	
    	sb.append("\nPROFILER: vuze " + + vuzeCpuUsage.getLast() + "%\n");
    	sb.append("PROFILER: dasu: " + dasuCpuUsage.getLast() + "%\n");
		
		LinePlot plot5 = null;
		float reaper_cpu = 0;
		int reaper_count = 0;
		
		for ( Long threadId : threadIdsMap.keySet()) {		

			if ( threadIdsMap.get(threadId).getCpuUsage(time) > pct ) {
				toPlot.add(threadId);
			}
			
			if ( threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("dasu") &&
					threadIdsMap.get(threadId).getCpuUsageList().size() > 0) {
				
				sb.append("PROFILER: *" + threadIdsMap.get(threadId).getThread().getName() + ": " +
						threadIdsMap.get(threadId).getCpuUsageList().getLast() + "\n");

			}
			
			if ( (threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("reaper") ||
					threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("ping") ||
					threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("trace") ) && 
					threadIdsMap.get(threadId).getCpuUsageList().size() > 0) {
				
				reaper_cpu += threadIdsMap.get(threadId).getCpuUsageList().getLast();
				reaper_count++;
				//System.out.println("reaper: " + " " + threadBean.getAllThreadIds().length);
					
			}
			
			if ( toPlot.contains(threadId)) {
				
				if (vuzeCpuUsage.size() > 5) {
					
					String tname = threadIdsMap.get(threadId).getThread().getName();
					LinkedList<Float> signal = fillPlotGap(threadIdsMap.get(threadId), timeAxis);
					
					if ( plot5 == null) 
						 plot5 = new LinePlot("resource.profiler.cpu.gt5.plot", 
								 tname, timeAxis, signal, "(time)", "(utilization)", LinePlot.FORMAT_TIME);
					else
						plot5.addSeries(tname, signal);
										
				}
			}
		}
		
		sb.append("PROFILER: reaper: " + reaper_cpu + "% count: " + reaper_count + "\n");
		
		if ( System.currentTimeMillis() - prevUIUpdate > 15*1000) {
			if ( plot5 != null)
				DasuData.getInstance().update(plot5.getId(), plot5.generatePlot());
	
							
			if (vuzeCpuUsage.size() > 5) {
				LinePlot plot = new LinePlot("resource.profiler.cpu.vuze.plot", "vuze", timeAxis, vuzeCpuUsage, "(time)", "(utilization)", LinePlot.FORMAT_TIME);
				plot.addSeries("dasu", dasuCpuUsage);
				DasuData.getInstance().update(plot.getId(), plot.generatePlot());
			}
			
			prevUIUpdate = System.currentTimeMillis();
		}
				
		sb.append("PROFILER: *******\n");
		LOGGER.info("PROFILER: " + sb);


    }
    
	private class ThreadCpuInfo {
		
		Long threadId;
		LinkedList<Long> cpuTime = new LinkedList<Long>();
		LinkedList<Long> cpuTimeTimeStamp = new LinkedList<Long>();		
		
		LinkedList<Float> cpuUsage = new LinkedList<Float>();
		LinkedList<Long> cpuUsageTimeStamp = new LinkedList<Long>();
		
		//ThreadInfo threadInfo;
		Thread thread;
		
		/*
		public ThreadCpuInfo(long id, ThreadInfo info) {
			this.threadId = id;			
			this.threadInfo = info;			
		}
		*/
		
		public ThreadCpuInfo(long id, Thread t) {
			this.threadId = id;
			this.thread = t;
		}
		
		public void addCpuTime(Long ctime, Long time) {
			
        	if ( cpuTime.size() == MAX_NUM_TIME_ENTIES_PER_THEAD) {
        		cpuTime.pop();
        		cpuTimeTimeStamp.pop();
        	}

			cpuTime.add(ctime);
			cpuTimeTimeStamp.add(time);
        	
        	updateCpuUsage();

		}
		
		/*
		public ThreadInfo getThreadInfo() {
			return threadInfo;
		}
		*/
		public Thread getThread() {
			return thread;
		}
		
		
		public long getElapsedCpuTime(long timeStamp) {

			if ( cpuTimeTimeStamp.size() > 1 ) {
				
				int index = -1;
				if ( cpuTimeTimeStamp.get(cpuTimeTimeStamp.size()-1).equals(timeStamp))
					index = cpuTimeTimeStamp.size()-1;
				else 
					index = cpuTimeTimeStamp.indexOf(timeStamp);
				
				if ( index != -1)
					return cpuTime.get(index) - cpuTime.get(index-1);
			}
			
			return 0l;		
			
		}
			
		public long getCpuTime(Long timeStamp) {
			
			if ( cpuTimeTimeStamp.size() > 0 ) {
				
				int index = -1;
				if ( cpuTimeTimeStamp.get(cpuTimeTimeStamp.size()-1).equals(timeStamp) )
					index = cpuTimeTimeStamp.size()-1;
				else
					index = cpuTimeTimeStamp.indexOf(timeStamp);
				
				if ( index != -1)
					return cpuTime.get(index);
			}
			
			return 0l;
			
		}
		
		public float getCpuUsage(Long timeStamp) {

			if ( cpuUsageTimeStamp.size() > 0) {
				
				int index = -1;
				if ( cpuUsageTimeStamp.get(cpuUsageTimeStamp.size()-1).equals(timeStamp) )
					index = cpuUsageTimeStamp.size()-1;
				else
					index = cpuUsageTimeStamp.indexOf(timeStamp);
				
				if ( index != -1)
					return cpuUsage.get(index);
			}
			
			return 0l;
			
		}
		
		public LinkedList<Float> getCpuUsageList() {
			return cpuUsage;
		}
		
		public LinkedList<Long> getTimeStamp() {
			return cpuUsageTimeStamp;
		}
		
		private boolean updateCpuUsage() {
						
			long elapsedTime = upTime - prevUpTime;
			
	        if (prevUpTime > 0L && upTime > prevUpTime) {
				
	        	if ( cpuTime.size() > 1) {
					// elapsedCpu is in ns
		    		int lastIndex = cpuTime.size()-1;
		            long elapsedCpuTime = cpuTime.get(lastIndex) - cpuTime.get(lastIndex-1);
		            	
		        	if ( cpuUsage.size() == MAX_NUM_TIME_ENTIES_PER_THEAD) {		        		
		        		cpuUsage.pop();
		        		cpuUsageTimeStamp.pop();
		        		//cpuUsageTimeStampSet.remove(t);
		        	}
	
		            float usage = ((float)elapsedCpuTime / ((float)elapsedTime))*100;	        		
		    		cpuUsage.add(usage);
		    		cpuUsageTimeStamp.add(cpuTimeTimeStamp.get(lastIndex));					
		        
		        	return true;
				}
	        }
	        

			return false;
    		
		}
				
	}
}
