/**
 * Dasu Project
 *
 * File:         DasuCronScheduler.java
 * RCS:          $Id$
 * Description:  DasuCronScheduler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 4:55:53 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.timer
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.timer;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Properties;
import java.util.Map.Entry;

import org.eclipse.swt.SWT;
import org.gudy.azureus2.ui.swt.Utils;
import org.gudy.azureus2.ui.swt.shells.MessageSlideShell;

import edu.northwestern.cs.aqualab.Cattle.json.JSONObject;
import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The DasuMessageManager class ...
 */

public class DasuMessageManager implements Runnable {

	public final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
	
	public class Message {
		String text;
		int timeout; //seconds
		boolean repeatable;
		String id;
		boolean audible;
		String details;
		
		public Message(String id, String text, int timeout, boolean repeatable) {
			this.id = id;
			this.text = text;
			this.timeout = timeout;
			this.repeatable = repeatable;
			this.audible = false;
			this.details = new String();
		}
		
		public Message(String id, String text, int timeout, boolean repeatable, boolean audible, String details) {
			this(id, text, timeout, repeatable);
			this.audible = audible;
			this.details = details;
		}
		
		public boolean isRepeatable() {
			return repeatable;
		}
		
		public boolean isAudible() {
			return audible;
		}
		
		public String toString() {			
			return "id: " + id + " message: " + text + " timeout: " + timeout + " repeatable: " + repeatable + " audible: " + audible + " details: " + details; 
		}
		
		public boolean equals(Object obj) {
			
			if (!(obj instanceof Message))
				return false;		
			
			return ((Message) obj).id.equals(this.id);
			
		}
		
		public int hashCode() {
			return id.hashCode();
		}
		
	}
	
	private static DasuMessageManager self = null;
	private static int schedulerInterval = 
		DasuConfiguration.getInstance().getMessageManagerInterval();
	
	private static LinkedHashMap<String, Long> messageLog =
		DasuManager.getPersistentDataManager().getMessageLog();
	private static LinkedHashSet<Message> messages = new LinkedHashSet<Message>();
	private static HashSet<String> alreadyDisplayed = new HashSet<String>();
	
	private Boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	private static boolean DEBUG = false;

	private DasuMessageManager() {
		super();
	}
	
    public static DasuMessageManager getInstance() {
    	if (self == null)
    		self = new DasuMessageManager();
	    return self;	    
    }
    
    public boolean isActive() {
    	return isActive;
    }
    
	public void addMessage (String id, String msg, int timeout, boolean repeatable, boolean audible, String details) {
		
		Message message =new Message(id, msg, timeout, repeatable, audible, details);		
		if (DEBUG)
			LOGGER.info("MESSAGE: addMessage(): " + message);

		/* if message already displayed, don't add it to the queue */
		synchronized(messageLog) {
			if (!messageLog.containsKey(id)) {
				
				synchronized(alreadyDisplayed) {
					if ( !alreadyDisplayed.contains(id) ) {
						messages.add(message);
						synchronized(halt) {
							halt.notify();
						}
					}
				}
			}
		}
	}
	
	public void run() {
		
		isActive = true;
		LOGGER.info("MESSAGE: Starting Message Manager");
		
		while (isActive) {
					
			if (DEBUG) {
				schedulerInterval = 5;
				LOGGER.info("MESSAGE: DasuMessageManager.class: DasuMessageManager(): I'm awake!!");
			}
				
			
			checkMessages();
			
			Iterator<Message> iter = messages.iterator();
			while (iter.hasNext()) {
				
				final Message m = iter.next();
  				iter.remove();
				
  				if (DasuConfiguration.getInstance().isEnablePopups()) {
  					
  					//Logger.log(new LogAlert(m.repeatable, LogAlert.AT_INFORMATION, "Dasu Plug-in:\n" + m.text, m.timeout));
  					
  					if ( m.isAudible())
  						Utils.beep();
  					
  					Utils.execSWTThread(new Runnable(){

  						public void run() {
  							new MessageSlideShell(Utils.getDisplay(), SWT.ICON_INFORMATION, "Dasu Plug-in",
  									m.text, m.details, m.timeout);  							
  						}});
  					
  				}
  				
  				if ( !m.isRepeatable() ) {
  					
  					synchronized(messageLog) {
  						messageLog.put(m.id, Util.currentTimeMillis());
  					}  					
  				}

  				synchronized(alreadyDisplayed) {
  					alreadyDisplayed.add(m.id);
  				}
			}
			
			
			synchronized (halt) {

				try {				
					if (DEBUG)
						LOGGER.info("MESSAGE: DasuMessageManager.class: sleeping for " + schedulerInterval + " seconds...");
					
					halt.wait(schedulerInterval*1000);
					
				} catch (InterruptedException e) {					
				}	
			}
			
			if (terminate) {

				self = null;
				return;
			}

		}//while
		
	}
	
	public void checkMessages() {

		if (DEBUG) {
			LOGGER.info("MESSAGE: DasuMessageManager.class: checkMessages(): messages.size(): " + messages);
			LOGGER.info("MESSAGE: DasuMessageManager.class: checkMessages(): messageLog.size(): " + messageLog);
		}
		
		JSONObject json = DasuManager.getInstance().getConfigService().GETData(DasuConfiguration.getInstance().DASU_MESSAGE_LIST
				+ DasuConfiguration.getInstance().getResourceScriptRequest());
		if ( json != null ) {
			
			Iterator<String> iter = json.keys();
			while ( iter.hasNext() ) {
				
				String id = iter.next();
				
				try {
										
					JSONObject message = (JSONObject) json.get(id);
					int timeout = -1;
					boolean repeatable = false;
					boolean audible = false;					
					String msg = (String) message.get("message");
					String details = "";
					
					if ( message.has("timeout") )
						timeout = message.getInt("timeout");
					if ( message.has("repeatable") )
						repeatable = message.getBoolean("repeatable");
					if ( message.has("audible") )
						audible = message.getBoolean("audible");
					if ( message.has("details") )
						details = message.getString("details");
					
					addMessage(id, msg, timeout, repeatable, audible, details);
					
				}
				catch (Exception e) {					
					LOGGER.info("MESSAGE: " + e.getMessage());
				}
			}
		}
	}
		
	public void stop() {
		
		LOGGER.info("STOP: Stopping Message Manager");
		isActive = false;
		synchronized(halt) {
			terminate = true;
			halt.notify();
		}
	}
	
}
