package edu.northwestern.dasu.util;

import java.io.*;
import java.util.*;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/*
 * This file is part of FindClass.
 * FindClass is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * FindClass is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FindClass; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Find classes or other files in JARs, Zips or directories.
 * <p>See http://www.nsydenham.net/java/FindClass/FindClass.shtml for details.</p>
 *
 * @author Copyright (C) 2003-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class FindClass
{
	public List<String> files = new ArrayList();
	public List<String> locations = new ArrayList();
	public List<String> matches = new ArrayList();
	public List<String> fqn_matches = new ArrayList();
	private boolean caseSensitive = true;
	private boolean exactMatch = false;
	private boolean recurse = false;
	private boolean matchArchives = false;
	private boolean verbose = true;
	private PrintStream print = System.out;
	private final static String fs = System.getProperty("path.separator");
	public final static String version = "1.2";
	
	// constants
	private final static String JAR = "jar";
	private final static String ZIP = "zip";
		
	/** Filters for jar, zip and class files plus directories */
	class JavaFilter implements FileFilter
	{
		public boolean accept(File file)
		{
			if (file.isDirectory()) return true;
			else
			{
				String lName = file.getName().toLowerCase();
				if (lName.endsWith(JAR) || lName.endsWith(ZIP)) {
					return true;
				}
			}
			
			return false;
		}
	}
	
	/** Iterate through locations looking for files */
	public void findFiles()
	{
		// add system classpath
		if (locations.size() == 0)
		{
			addLocations(System.getProperty("java.class.path"));
		}
		
		if (locations.size() > 0)
		{
			searchFiles();
		}
		
		Collections.sort(matches);
		
		
	}
	
	/**
	 * Add a file to look for
	 * @param fileName file to look for.
	 */
	public void addFile(String fileName)
	{
		files.add(fileName);
	}
	
	/** Search for files in the path */
	private void searchFiles()
	{
		String location;
		String fileName;
		Iterator fit, lit;
		
		fit = files.iterator();
		while (fit.hasNext())
		{
			fileName = (String)fit.next();
			
			lit = locations.iterator();
			while (lit.hasNext())
			{
				location = (String)lit.next();
				
				if (location.toLowerCase().endsWith(JAR) || location.toLowerCase().endsWith(ZIP)) {
					findInArchive(fileName, location);
				} else {
					findInDirectory(fileName, location);
				}
			}
		}
	}
	
	/**
	 * Check if the file is part of the specified path based on case and exact match parameters
	 * @param fileName the name being searched for
	 * @param path the path being searched
	 */
	private boolean isMatch(String fileName, String path)
	{
		if (exactMatch &&
			(caseSensitive && path.equals(fileName)) || 
			(!caseSensitive && path.equalsIgnoreCase(fileName)))
		{
			return true;
		}
		else if (!exactMatch &&
			(caseSensitive && path.indexOf(fileName) > -1) ||
			(!caseSensitive && path.toLowerCase().indexOf(fileName.toLowerCase()) > -1))
		{
			return true;
		}
		
		return false;
	}
	
	/**
	 * See if the file exists in the specified JAR or ZIP file
	 * @param fileName name of the file to look for
	 * @param archive name of the archive to examine
	 */
	private void findInArchive(String fileName, String archive)
	{
		ZipEntry ze;
		String name;
		
		try
		{
			// use a JarFile because it can also read zips
			JarFile jf = new JarFile(archive);
			Enumeration enume = jf.entries();
			while (enume.hasMoreElements())
			{
				ze = (ZipEntry)enume.nextElement();
				name = ze.getName();
				if (isMatch(fileName, name)) {
					matches.add(archive+File.separator+name);
					name = name.replace("/", ".");
					fqn_matches.add(name);
				}
			}
		}
		catch (FileNotFoundException fnex)
		{
			print.println("    File "+archive+" doesn't exist");
		}
		catch (IOException ioex)
		{
			print.println("    Error opening file "+archive);
		}
	}
	
	/**
	 * See if the file exists in the specified directory
	 * @param fileName name of the file to look for
	 * @parm dir name of the directory to search
	 */
	private void findInDirectory(String fileName, String dir)
	{
		File dirFile = new File(dir);
		File dirEntries[];
		File dirEntry;
		String name;
		
		log("Searching "+dir);
		
		// check directory is valid
		if (!dirFile.exists())
		{
			print.println("    "+dir+" doesn't exist");
			return;
		}
		else if (!dirFile.isDirectory())
		{
			print.println("    "+dir+" is not a directory");
			return;
		}
		
		if (matchArchives) {
			dirEntries = dirFile.listFiles(new JavaFilter());
		} else {
			dirEntries = dirFile.listFiles();
		}
		
		for (int i=0; i<dirEntries.length; i++)
		{
			dirEntry = dirEntries[i];
			name = dirEntry.getName();

			if (recurse && dirEntry.isDirectory())
			{
				findInDirectory(fileName, dirEntry.getAbsolutePath());
			}
			else if (name.endsWith(JAR) || name.endsWith(ZIP))
			{
				findInArchive(fileName, dirEntry.getAbsolutePath());
			}
			else if (isMatch(fileName, name))
			{				
				matches.add(dir+File.separator+name);
			}
		}
	}
	
	/**
	 * Add entries
	 * @param filePath the path to search
	 */
	public void addLocations(String filePath)
	{
		StringTokenizer st = new StringTokenizer(filePath, fs);
		String cpEntry;
		
		while (st.hasMoreTokens())
		{
			locations.add(st.nextToken());
		}
	}
	
	/**
	 * Set whether the tool should be verbose or not
	 * @param verbose if true print progress
	 */
	public void setVerbose(boolean verbose)
	{
		this.verbose = verbose;
	}
	
	/**
	 * Uses the verbose variable to determine if the message should be displayed
	 * @param message to display
	 */
	private void log(String message)
	{
		if (verbose) print.println(message);
	}
	
	/**
	 * If set to true, recurse into directories
	 * @param recurse 
	 */
	public void setRecurse(boolean recurse)
	{
		this.recurse = recurse;
	}
	
	/**
	 * Set the OutputStream to something other than System.out
	 * @param outs the new OutputStream
	 */
	public void setOutputStream(OutputStream outs)
	{
		print = new PrintStream(outs);
	}
	
	/**
	 * Set the output to the specified file
	 * @param filename file to log to
	 * @exception FileNotFoundException if the specified file doesn't exist
	 */
	public void setOutputFile(String filename) throws FileNotFoundException
	{
		if (filename == null) return;
		setOutputStream(new FileOutputStream(filename));
	}
	
	/**
	 * Set whether the search is case sensitive
	 * @param cs if true the search will be case sensitive
	 */
	public void setCaseSensitive(boolean cs) {
		this.caseSensitive = cs;
	}
	
	/**
	 * Set whether matching will be exact to the word
	 * @param exactMatch if true match search param exactly
	 */
	public void setExactMatch(boolean exactMatch) {
		this.exactMatch = exactMatch;
	}
	
	public void setArchiveMatch(boolean archiveMatch) {
		this.matchArchives = archiveMatch;
	}
	
	/** Start the utility */
	public static void main(String[] args) {
		FindClass find = new FindClass();
		find.addLocations("/Users/mas939/Mario's Stuff/NORTHWESTERN/Dropbox/workspace/ISP Interference/testdasu/plugins/dasu/DasuUiRulesModules_4.0.jar");
		find.addFile("class");
		find.setRecurse(true);
		find.findFiles();
		System.out.println(find.matches);
		System.out.println(find.fqn_matches);
	}
}
