/**
 * Dasu Project
 *
 * File:         Updater.java
 * RCS:          $Id: Updater.java,v 1.0 Nov 17, 2011 7:42:37 PM mas939 Exp $
 * Description:  Updater class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Nov 17, 2011 at 7:42:37 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.ui
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.util;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderConfiguration;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;

import edu.northwestern.cs.aqualab.Cattle.json.JSONException;
import edu.northwestern.cs.aqualab.Cattle.json.JSONObject;
import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.security.JarVerifier;

/**
 * This should only run if DEBUG is not set.
 * Follows the same scheme that the RuleEngine uses to update the inference rules.
 * 
 * @author JSO
 *
 * The Updater class ...
 */
public class UpdaterV2 extends Thread {
	
	java.util.logging.Logger LOGGER;
	
	private final String jarPrefix = DasuConfiguration.DASU_FILENAME_PREFIX;
	private final String jarSuffix = DasuConfiguration.DASU_FILENAME_SUFFIX;
	
	private final String jarVersionFile = "VERSION";
	private final String jarRulesPrefix = "rules/";
	private final String jarRulesSuffix = ".drl";
	private final String jarRulesBaseName = "base";
	private final String jarUiPrefix = "ui/";	
	
	// XXX: This must be set to the version that is packed in the Jar!
	private final Float jarDefaultVersion = 4.0F;
	
	private Float toUse = null;
	private boolean ready = false;
	
	private String latestJarUrl = null;
	private Float latestJarVersion = null;
	
	private final String updateServerJarLocation = "dasu.inference.jar";
	private final String updateServerJarVersion = "dasu.inference.version";
	
	RulesVerifier rulesVerifier = null;
	
	public UpdaterV2(java.util.logging.Logger logger) {
		this.setName("Dasu::Updater");
		LOGGER = logger;	
		rulesVerifier = new RulesVerifier(logger);
	}
	
	public boolean canUpdate() {
		boolean checkSucceeded = checkLatestVersion();
		return checkSucceeded && (toUse == null || latestJarVersion > toUse);
	}
	
	public boolean isReady() {
		return ready;
	}
	
	public boolean haveValidFile() {
		return toUse != null;
	}
	
	public Float getVersion() {
		return toUse;
	}
	
	/*
	 * return success of check
	 */
	private boolean checkLatestVersion() {
		
		JSONObject json = DasuManager.getInstance().getConfigService().GETData(DasuConfiguration.DASU_CONFIG_DYNAMIC + 
				DasuConfiguration.getInstance().getDasuConfigDynamicResource() + DasuConfiguration.getInstance().getResourceScriptRequest());
		if ( json == null) return false; //Return failure
		
		try {
			
			Iterator<String> keys = json.keys();
			while ( keys.hasNext() ) {
				String key = keys.next();
				if (key.equals(updateServerJarLocation))
					latestJarUrl = (String) json.get(key);
				if (key.equals(updateServerJarVersion))
					latestJarVersion = new Float( (Double) json.get(key) );
			}
				
		} catch (JSONException e1) {
			e1.printStackTrace();
		}
		
		boolean result = latestJarUrl != null && latestJarVersion != null;
		LOGGER.info("UPDATER: checkLatestVersion returning " + result + " with found version " + latestJarVersion);
		return result;
	}
	
	/*
	 * Return a list of version numbers for which we have a jar file, in descending order.
	 */
	private ArrayList<Float> getLocalVersions() {
		HashMap<Float, File> jars = new HashMap<Float, File>();
		
		File dir = new File(Main.getDir());
		for(File f: dir.listFiles()) {
			String name = f.getName();
			if(!name.startsWith(jarPrefix)) continue;
			
			Float version = null;
			int toCrop = 0;
			
			// jar file
			toCrop = jarSuffix.length();
			
			try {
				version = Float.parseFloat(name.subSequence(jarPrefix.length(), name.length() - toCrop).toString());
			} catch(NumberFormatException e) {}
			
			if(version != null) {
				jars.put(version, f);
			}
			
		}
		ArrayList<Float> versions = new ArrayList<Float>(jars.keySet());
		Collections.sort(versions);
		Collections.reverse(versions);
		LOGGER.info("UPDATER: getLocalVersions found these versions: " + versions);
		return versions;
	}
	
	private String getJarName(Float version) {
		return jarPrefix + version.toString() + jarSuffix;
	}
	
	public String getJarPath(Float version) {		
		return Main.getDir() + File.separator + getJarName(version);
	}
	
	private boolean isVerified(Float version) {
		
		boolean result = true;
		try {		
			LOGGER.info("UPDATER: verify " + getJarPath(version));
			JarVerifier.verify(getJarPath(version), DasuConfiguration.getInstance().getKeyStore());
		}
		catch (Exception e) {
			e.printStackTrace();
			result = false;
		}
		
		LOGGER.info("UPDATER: checking if signature is valid for version " + version + " -- " + result + "!");
		return result;
	}
	
	@SuppressWarnings("unchecked")
	private boolean versionMatches(Float version) {
		ZipFile zf;
		boolean result = false;
		try {
			zf = new ZipFile(getJarPath(version));
			Enumeration e = zf.entries();
			ZipEntry ze = null;
			while(e.hasMoreElements()) {
				ZipEntry obj = (ZipEntry) e.nextElement();
				if(obj.getName().equals(jarVersionFile)) {
					ze = obj;
					break;
				}
			}
			if(ze == null) {
				LOGGER.info("UPDATER: no " + jarVersionFile + " in " + getJarPath(version));
				zf.close();
			} else {
				BufferedReader br = new BufferedReader(new InputStreamReader(zf.getInputStream(ze)));
				String line = br.readLine().trim();
				br.close();
				zf.close();
				Float thisVersion = Float.parseFloat(line);
				result = thisVersion.equals(version);
			}
		} catch (IOException e) {
			//e.printStackTrace();
		} catch (NumberFormatException e) {
			//e.printStackTrace();
		}
		LOGGER.info("UPDATER: checking if version matches for version " + version + " -- " + result + "!");
		return result;
	}
	
	
	@SuppressWarnings("unchecked")
	private boolean rulesBuild(Float version) {
		
		String fileName = getJarPath(version);
		boolean result = false;
		
		try {
			result = rulesVerifier.sendVerificationRequest(fileName);
		}
		catch (Exception e) {
			//LOGGER.severe("UPDATER: rulesVerifier is not starting properly... all verifications will fail");
			LOGGER.severe("UPDATER: rulesVerifier request timed out");
		}
		
		return result;
	}
	
	
	@SuppressWarnings("unchecked")
	private boolean rulesBuildv1(Float version) {
		boolean result = false;
		try {
			
			LOGGER.info("UPDATER: rulesBuild: " + version + " " + getJarPath(version));
			ZipFile zf = new ZipFile(getJarPath(version));
			Properties props = new Properties();
	        props.setProperty("drools.dialect.java.compiler", "JANINO");
	        
	        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, 
	        		new ClassLoader[]{this.getClass().getClassLoader()});
			KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
			
			// load the base rule file
			boolean found = false;
			Enumeration e = zf.entries();
			while(e.hasMoreElements()) {
				ZipEntry obj = (ZipEntry) e.nextElement();
				if(obj.getName().equals(jarRulesPrefix + jarRulesBaseName + jarRulesSuffix)) {
					found = true;
					
					// try to build this file
					try {
						InputStream is = zf.getInputStream(obj);
						kbuilder.add(ResourceFactory.newInputStreamResource(is), ResourceType.DRL);
						is.close();
						if (kbuilder.getErrors().size() > 0) {
							throw new IllegalArgumentException("Could not parse knowledge.");
						}
						result = true;
					} catch(Exception e2) {
						LOGGER.info("UPDATER: exception while building " + jarRulesBaseName + " in " + getJarPath(version));
						e2.printStackTrace();
					}
					break;
				}
			}
			
			if(!found) {
				LOGGER.info("UPDATER: did not find base rule file " + jarRulesPrefix + jarRulesBaseName + jarRulesSuffix + " in " + getJarPath(version));
			}
			if (found && result) {
				// Make sure all other files build, too.
				boolean okay = true;
				
				e = zf.entries();
				while(e.hasMoreElements()) {
					ZipEntry obj = (ZipEntry) e.nextElement();
					if(obj.getName().startsWith(jarRulesPrefix) && obj.getName().endsWith(jarRulesSuffix) && !obj.getName().equals(jarRulesPrefix + jarRulesBaseName + jarRulesSuffix)) {
						
						// try to build this file
						try {
							InputStream is = zf.getInputStream(obj);
							kbuilder.add(ResourceFactory.newInputStreamResource(is), ResourceType.DRL);
							is.close();
							if (kbuilder.getErrors().size() > 0) {
								throw new IllegalArgumentException("Could not parse knowledge.");
							}
						} catch(Exception e2) {
							okay = false;
							LOGGER.info("UPDATER: exception while building " + obj.getName() + " in " + getJarPath(version));
							break;
						}
					}
				}
				
				if(!okay) result = false;
			}
			zf.close();
		} catch (IOException e) {
			//e.printStackTrace();
		} catch (NumberFormatException e) {
			//e.printStackTrace();
		}
		LOGGER.info("UPDATER: checking if rules build for version " + version + " -- " + result + "!");
		return result;
	}
	
	private void delete(Float version) {
		File zf = new File(getJarPath(version));
		if(zf.exists() && zf.canWrite()) zf.delete(); 
	}
	
	private boolean extractFromJar(String fileToExtract) throws IOException {
		
		ClassLoader classLoader = ClassLoader.getSystemClassLoader();
		if ( Main.getPluginInterface() != null)
			classLoader = Main.getPluginInterface().getPluginClassLoader();

		if (!(classLoader instanceof URLClassLoader)) {
			LOGGER.info("UPDATER: classloader not instanceof URL classloader");
			return false;
		}
		
		URLClassLoader current = (URLClassLoader) classLoader;

		URL url = current.findResource(fileToExtract);
		if(url == null) {
			LOGGER.info("UPDATER: can't find file " + fileToExtract);
			return false;
		}
		
		DataInputStream in = new DataInputStream(url.openStream());	        
	    byte[] bytes = new byte[in.available()];
	  
	    int offset = 0;
	    int numRead = 0;
	    while (offset < bytes.length
	             && (numRead=in.read(bytes, offset, bytes.length-offset)) >= 0) {
	          offset += numRead;
	    }
        
        String filename = Main.getDir() + java.io.File.separator + fileToExtract;

	    File file=new File(filename);
	    DataOutputStream outs = new DataOutputStream(new FileOutputStream(file,false));
	    outs.write(bytes);
	    outs.close();
	    
		return true;
	}
	
	private Float getBestWorkingLocalVersion() {
		Float bestWorkingLocalVersion = null;
		for(Float version: getLocalVersions()) {
			if(isVerified(version) && versionMatches(version) && rulesBuild(version) && bestWorkingLocalVersion == null) {
				bestWorkingLocalVersion = version;
				LOGGER.info("UPDATER: testing version " + version + " with path " + getJarPath(version) + " -- OK!");
			} else {
				// Delete versions that fail or when we already have a newer working version.
				LOGGER.info("UPDATER: testing version " + version + " with path " + getJarPath(version) + " -- failed or old. Deleting!");
				delete(version);
			}
		}
		return bestWorkingLocalVersion;
	}
	
	@SuppressWarnings("unchecked")
	public HashMap<String, ByteArrayOutputStream> getUiData() {
		HashMap<String, ByteArrayOutputStream> data = new HashMap<String, ByteArrayOutputStream>();
		
		if(toUse != null) {
			try {
				ZipFile zf = new ZipFile(getJarPath(toUse));
				Enumeration e = zf.entries();
				while(e.hasMoreElements()) {
					ZipEntry ze = (ZipEntry) e.nextElement();
					if(!ze.getName().startsWith(jarUiPrefix)) continue;
					
					// get the input stream, load the byte array
					InputStream is = zf.getInputStream(ze);
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					int nRead;
					byte[] arr = new byte[4096];
					while((nRead = is.read(arr, 0, arr.length)) != -1) {
						baos.write(arr, 0, nRead);
					}
					is.close();
					data.put(ze.getName().substring(jarUiPrefix.length()), baos);
				}
				zf.close();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (NumberFormatException e) {
				e.printStackTrace();
			}
		}
		return data;
	}
	
	@SuppressWarnings("unchecked")
	public Vector<ByteArrayOutputStream> getRulesData() {
		Vector<ByteArrayOutputStream> fileData = new Vector<ByteArrayOutputStream>();
		if(toUse != null) {
			try {
				
				LOGGER.info("UPDATER: loading file " + getJarPath(toUse));
				JarFile jf = new JarFile(getJarPath(toUse));
				
				Enumeration e = jf.entries();
				while(e.hasMoreElements()) {
					ZipEntry obj = (ZipEntry) e.nextElement();
					if(obj.getName().equals(jarRulesPrefix + jarRulesBaseName + jarRulesSuffix)) {
						
						LOGGER.info("UPDATER: reading file: " + jarRulesPrefix + jarRulesBaseName + jarRulesSuffix);
						// base rule file -- load first.
						InputStream is = jf.getInputStream(obj);
						ByteArrayOutputStream baos = new ByteArrayOutputStream();
						int nRead;
						byte[] data = new byte[16384];
						while((nRead = is.read(data, 0, data.length)) != -1) {
							baos.write(data, 0, nRead);
						}
						
						is.close();
						fileData.add(baos);
						break;
					}
				}
				
				e = jf.entries();
				while(e.hasMoreElements()) {
					ZipEntry obj = (ZipEntry) e.nextElement();
					if(obj.getName().startsWith(jarRulesPrefix) && obj.getName().endsWith(jarRulesSuffix) && !obj.getName().equals(jarRulesPrefix + jarRulesBaseName + jarRulesSuffix)) {
						// additional rule files -- load in any order.
						LOGGER.info("UPDATER: reading file: " + obj.getName());
						InputStream is = jf.getInputStream(obj);
						ByteArrayOutputStream baos = new ByteArrayOutputStream();
						int nRead;
						byte[] data = new byte[16384];
						while((nRead = is.read(data, 0, data.length)) != -1) {
							baos.write(data, 0, nRead);
						}
						
						is.close();
						fileData.add(baos);
					}
				}					
				
				jf.close();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (NumberFormatException e) {
				e.printStackTrace();
			}
		}
		return fileData;
	}
	
	public void run() {
		
		LOGGER.info("UPDATER: Updater started");
		if ( rulesVerifier != null )
			rulesVerifier.initRulesVerifier();
		
		// Get the latest local version that works.
		Float bestWorkingLocalVersion = getBestWorkingLocalVersion();
		LOGGER.info("UPDATER: at start, bestWorkingVersion = " + bestWorkingLocalVersion);		

		// What is the latest version, as reported by the server?
		boolean checkSucceeded = checkLatestVersion();
		
		if(checkSucceeded && (bestWorkingLocalVersion == null || latestJarVersion > bestWorkingLocalVersion)) {
			LOGGER.info("UPDATER: trying to update");
			
			// Is the newest version (remote) newer than the best available local version?
			// Or, do we not have a local working version?
			// Then...download the newest files
			boolean jarSuccess = DasuManager.getInstance().getConfigService().GETFile(latestJarUrl, getJarPath(latestJarVersion));
			
			if(!jarSuccess) {
				LOGGER.info("UPDATER: update download failed");
				// update failed. delete one or more of the files.
				delete(latestJarVersion);
			} else {
				// Re-update to see what's working
				bestWorkingLocalVersion = getBestWorkingLocalVersion();
				LOGGER.info("UPDATER: update download succeeded. new bestWorkingVersion = " + bestWorkingLocalVersion);
				
			}
		}	
		
		if(bestWorkingLocalVersion == null) {
			LOGGER.info("UPDATER: after update, bestWorkingVersion is null. trying to extract from jar.");
			// after update, still don't have a working version.
			// extract from the jar
			try {
				
				boolean zipSuccess = extractFromJar(getJarName(jarDefaultVersion));
				
				if(!zipSuccess) {
					LOGGER.severe("UPDATER: extract from jar failed.");
					LOGGER.severe("UPDATER: using jarName " + getJarName(jarDefaultVersion));
					delete(jarDefaultVersion);
				} else {
					LOGGER.info("UPDATER: extract from jar succeeded.");
				}
				
				/*
				// Get the best working version, again.
				//bestWorkingLocalVersion = getBestWorkingLocalVersion();
				//LOGGER.info("UPDATER: after extract from jar, bestWorkingLocalVersion = " + bestWorkingLocalVersion);
				*/
				
				//don't check the rules if the jar comes from inside the plugin jar and it's properly signed, assume it's correct
				if(isVerified(jarDefaultVersion) && versionMatches(jarDefaultVersion) && bestWorkingLocalVersion == null) {
					bestWorkingLocalVersion = jarDefaultVersion;
					LOGGER.info("UPDATER: testing version " + jarDefaultVersion + " with path " + getJarPath(jarDefaultVersion) + " -- OK!");
				} else {
					// Delete versions that fail or when we already have a newer working version.
					LOGGER.info("UPDATER: testing version " + jarDefaultVersion + " with path " + getJarPath(jarDefaultVersion) + " -- failed or old. Deleting!");
					delete(jarDefaultVersion);
				}

				
			} catch(IOException e) {
				LOGGER.severe("UPDATER: extract from jar failed");
				e.printStackTrace();
				delete(jarDefaultVersion);
			}
		}
		
		if(bestWorkingLocalVersion == null) {
			// Even version from the jar must be busted. Whoa.
			LOGGER.severe("UPDATER: all attempts to load a UI/rules version failed");
		} else {
			toUse = bestWorkingLocalVersion;
			LOGGER.info("UPDATER: using UI version " + toUse);
		}
		
		if ( toUse != null ) { 
							
			try {
				 LOGGER.info("UPDATER: " + getJarPath(toUse));
		         JarFile jf = new JarFile(getJarPath(toUse));
		         ZipEntry ze = jf.getEntry("probe.modules.properties");
		         Properties prop = new Properties();
		         prop.load(jf.getInputStream(ze));

		         DasuConfiguration.getInstance().setProbeModulesProperties(prop);
		         
			}
			catch (Exception e) {
				e.printStackTrace();
			}

			// Set the version being used...
			DasuConfiguration.getInstance().setInferenceVersion(toUse);
			DasuConfiguration.getInstance().updatePropertiesJson();
			
			ready = true;
			LOGGER.info("UPDATER: UI update finished; ready");
		}
		
		if ( rulesVerifier != null )
			rulesVerifier.disconnect();
		
	}
	
	@SuppressWarnings("unchecked")
	public boolean rulesBuild(String fileName) {
		boolean result = false;
		try {
						
			ZipFile zf = new ZipFile(fileName);
			Properties props = new Properties();
	        props.setProperty("drools.dialect.java.compiler", "JANINO");
	        
	        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, 
	        		new ClassLoader[]{this.getClass().getClassLoader()});
			KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
			
			// load the base rule file
			boolean found = false;
			Enumeration e = zf.entries();
			while(e.hasMoreElements()) {
				ZipEntry obj = (ZipEntry) e.nextElement();
				if(obj.getName().equals(jarRulesPrefix + jarRulesBaseName + jarRulesSuffix)) {
					found = true;
					
					// try to build this file
					try {
						InputStream is = zf.getInputStream(obj);
						kbuilder.add(ResourceFactory.newInputStreamResource(is), ResourceType.DRL);
						is.close();
						if (kbuilder.getErrors().size() > 0) {
							throw new IllegalArgumentException("Could not parse knowledge.");
						}
						result = true;
					} catch(Exception e2) {
						LOGGER.info("UPDATER: exception while building " + jarRulesBaseName + " in " + fileName);
						e2.printStackTrace();
					}
					break;
				}
			}
			
			if(!found) {
				LOGGER.info("UPDATER: did not find base rule file " + jarRulesPrefix + jarRulesBaseName + jarRulesSuffix + " in " + fileName);
			}
			if (found && result) {
				// Make sure all other files build, too.
				boolean okay = true;
				
				e = zf.entries();
				while(e.hasMoreElements()) {
					ZipEntry obj = (ZipEntry) e.nextElement();
					if(obj.getName().startsWith(jarRulesPrefix) && obj.getName().endsWith(jarRulesSuffix) && !obj.getName().equals(jarRulesPrefix + jarRulesBaseName + jarRulesSuffix)) {
						
						// try to build this file
						try {
							InputStream is = zf.getInputStream(obj);
							kbuilder.add(ResourceFactory.newInputStreamResource(is), ResourceType.DRL);
							is.close();
							if (kbuilder.getErrors().size() > 0) {
								throw new IllegalArgumentException("Could not parse knowledge.");
							}
						} catch(Exception e2) {
							okay = false;
							LOGGER.info("UPDATER: exception while building " + obj.getName() + " in " + fileName);
							break;
						}
					}
				}
				
				if(!okay) result = false;
			}
			zf.close();
		} catch (IOException e) {
			//e.printStackTrace();
		} catch (NumberFormatException e) {
			//e.printStackTrace();
		}
		LOGGER.info("UPDATER: checking if rules build for version " + fileName + " -- " + result + "!");
		return result;
	}
	
	public static void main(String args[]) throws Exception {
		
		
		ServerSocket serverSocket = null;
		final java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
		UpdaterV2 up = new UpdaterV2(LOGGER);
	
	    try { 
	         serverSocket = new ServerSocket(1234); 	
	         Socket clientSocket = null; 	
	         clientSocket = serverSocket.accept(); 
	         PrintWriter out = new PrintWriter(clientSocket.getOutputStream(), 
		                                      true); 
	         BufferedReader in = new BufferedReader( 
		            new InputStreamReader( clientSocket.getInputStream())); 
		
	         String inputLine; 
		
	         while ((inputLine = in.readLine()) != null) { 
		         System.out.println ("Server: " + inputLine);
		         
		         Boolean result = up.rulesBuild(inputLine);	         
		         out.println(result); 
		
		         if (inputLine.equals("Bye.")) 
		             break; 
		    } 
		
		    out.close(); 
		    in.close(); 
		    clientSocket.close(); 
		    serverSocket.close(); 
	   }
	   catch (Exception e) {
	    	e.printStackTrace();
	   }
	}
		
}

